/*
** Module   : DatabaseManager.java
** Abstract : Registers and configures databases and manages session factories
**
** Copyright (c) 2004-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050901   @22473 Created initial version. Extracted and
**                           enhanced database configuration/management
**                           functionality from Persistence class.
** 002 ECF 20050915   @23252 Store Hibernate configurations during
**                           initialization. Provide lookups of ORM
**                           metadata at runtime.
** 003 ECF 20051111   @23378 Added notNullProperties method. Allows query
**                           of all non-nullable properties for a specific
**                           DMO class. Also replaced many checked
**                           exceptions with unchecked exceptions.
** 004 ECF 20051204   @23668 Added permanentDatabases method. Returns an
**                           iterator over auto-connect, permanent
**                           databases.
** 005 ECF 20060112   @23942 Partial integration with P2J server. Some
**                           directory service access points implemenated.
** 006 ECF 20060117   @23958 Remaining directory service access points
**                           implemented. Hibernate properties are now
**                           read from the directory.
** 007 ECF 20060121   @24040 Integrated with P2J server. Replaced
**                           ThreadLocal with ContextLocal.
** 008 ECF 20060130   @24353 Added temp table support. Introduced new
**                           methods and inner classes to support runtime
**                           management of temp tables.
** 009 ECF 20060309   @24993 Fixed NPE in notNullProperties method. Added
**                           a safety check to test the iterator for
**                           content before accessing it.
** 010 ECF 20060317   @25231 Added HQL where clause preprocessing support.
**                           Configuration data about lists of composite
**                           properties is stored for runtime lookup.
** 011 ECF 20060329   @25257 Changed directory entry name 'permanent' to
**                           'load_at_startup'. The former name was not
**                           intuitive in that the _temp (temporary table
**                           database) is actually a permanent database
**                           (i.e., loaded at server startup).
** 012 ECF 20060504   @26006 Added runtime support for the creation and
**                           dropping of secondary temp tables. These are
**                           tables created to support the extent fields
**                           of a primary temp table.
** 013 ECF 20060620   @27504 Added facility to store and report database
**                           dialects. Dialects are stored during startup
**                           and are queried using the getDialect method.
** 014 ECF 20060718   @28090 Made class public. Also exposed a number of
**                           generally useful methods as public.
** 015 ECF 20060831   @29084 Enable schema lookup by PDB or by DMO class.
** 016 NVS 20070126   @32007 "database" node of the directory is searched
**                           as a server level standard search item.
**                           HIBERNATE properties are read in natural
**                           data types and converted to strings.
** 017 ECF 20070212   @32159 Fixed lookups by DMO implementation class.
**                           Need to scan up chain of superclasses in case
**                           the caller of these lookup methods obtained
**                           the DMO class by invoking getClass() on a DMO
**                           instance. This is necessary because such an
**                           instance may actually be a proxied subclass
**                           of the actual DMO implementation class.
** 018 ECF 20070412   @32977 intern() table name strings. This will use
**                           slightly more permanent generation heap, but
**                           will avoid the creation of many duplicate
**                           instances, since these are used heavily for
**                           record buffers and record identifiers.
** 019 ECF 20070416   @33020 Integrated user interrupt handling.
** 020 ECF 20070508   @33455 Added getColumnToPropertyMap(). Maps column
**                           names to their corresponding Java property
**                           names from information read from Hibernate
**                           mapping documents.
** 021 ECF 20070517   @33675 Modified TempTableHelper's index generation.
**                           Added a unique index on a temp table's
**                           primary key.
** 022 ECF 20070629   @34903 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 023 ECF 20070906   @34995 Optimized temp table indices. Ensure leading
**                           component of every index is multiplex ID. If
**                           no other indices exist, create an index on
**                           multiplex ID only.
** 024 ECF 20070911   @35284 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections.
** 025 ECF 20071002   @35330 Exposed certain members as public. Expanded
**                           use of generics.
** 026 ECF 20071018   @35601 Intermediate set of changes to support remote
**                           database connections more appropriately. This
**                           feature set is not yet fully implemented, but
**                           no existing functionality has been removed.
** 027 ECF 20071120   @35908 Completed work to support remote database
**                           connections. Corrected transient database
**                           registration/deregistration. Implemented code
**                           to fetch and filter DatabaseConfig properties
**                           from a remote server. Updated initialize() to
**                           initialize LockManagerMultiplexer and
**                           DirtyReadMultiplexer network service
**                           providers.
** 028 ECF 20071128   @36120 Modified initialization. Temp table database
**                           is not included in set of locally managed
**                           databases.
** 029 ECF 20080111   @36797 Modified initialization. Exposed remote API
**                           for IdentityManagerMultiplexer to allow
**                           remote servers to request this server to
**                           generate unique primary keys for a database
**                           managed by the local server.
** 030 ECF 20080118   @36859 Added logging to registerDatabase().
** 031 ECF 20080310   @37474 Extracted TempTableHelper into a separate,
**                           top level class. Moved TableMapping as a
**                           secondary class into TempTableHelper.java.
** 032 ECF 20080331   @37730 Code cleanup. Renamed certain methods and
**                           added more generics.
** 033 SVL 20080421   @38096 If foreign keys are disabled, check whether
**                           all columns are mapped is disabled too.
** 034 CA  20080425   @38112 Modified initialization - a dirty database
**                           will be created for each non-temp database.
**                           On registration, this dirty database will
**                           have all its foreign-key relations dropped,
**                           except the relations regarding secondary 
**                           tables (which keep the "extent" data). This
**                           database is an in-memory, H2 database.
** 035 ECF 20080529   @38484 Fixed initialization. Secondary tables for
**                           extent fields must be mapped for permanent
**                           tables as well as for temporary tables.
** 036 ECF 20080605   @38602 Initialize DirtyShareMultiplexerImpl.
**                           Replaces DirtyReadMultiplexerImpl.
** 037 ECF 20080606   @38640 Added getSchemaByInterface() and supporting
**                           code.
** 038 ECF 20080610   @38679 Changed getColumnToPropertyMap() signature.
**                           Optionally include primary key 'id' mapping
**                           in resulting map.
** 039 ECF 20080611   @38693 Improved support for computed columns.
**                           Information for computed columns needed by
**                           dirty databases is now collected during
**                           server initialization, so it is available to
**                           the HQLPreprocessor when generating HQL
**                           optimized to the dirty databases' dialect.
**                           This required making JDBC connections to
**                           primary databases available during server
**                           initialization, before the Hibernate
**                           infrastructure is fully ready.
** 040 ECF 20080624   @38917 Optimized early connection management. These
**                           connections are now recycled within a (very)
**                           crude connection pool. No error handling was
**                           added, since a JDBC error during server
**                           initialization is generally a fatal error.
** 041 SVL 20080725   @39176 UnpooledConnectionProvider is specified as 
**                           the connection provider for dirty databases.
** 042 SVL 20080806   @39312 DirtyShareContext and DirtyShareManager
**                           unregistered into deregisterDatabase().
** 043 CA  20080815   @39453 Support API change in DBUtils.
** 044 GES 20090421   @41828 Matched utility class package change.
** 045 ECF 20090617   @42765 Minor startup optimization. Avoid unnecessary
**                           work in mapComputedCharacterColumns().
** 046 ECF 20090801   @43497 Added support for record lock administration. New
**                           getAllRecordLockInfo() method returns a snapshot
**                           of record lock information for all managed
**                           databases.
** 047 SVL 20091126   @44443 Added getDatabaseTables() and
**                           getEntityNameByTableName() functions. 
** 048 CA  20101029          Added getBackingTableName(dmoClass), which reads 
**                           the DMO's backing table name directly from its
**                           corresponding .hbm.xml file. Also,set any dialect
**                           specific VM properties before the database driver
**                           is loaded.
** 049 CA  20101202          The remote connection URL is determined using the
**                           dialect.
** 050 SVL 20130331          Upgraded to Hibernate 4.
** 051 VMN 20130711          Fixed registerDatabase, which must store Hibernate configs before
**                           computed columns are mapped.
** 052 SVL 20130716          Added support for dynamic tables.
** 053 SVL 20130822          Added support for secondary dynamic tables.
** 054 CA  20131015          Added support for registration/deregistration of legacy table/field 
**                           names.
** 055 ECF 20130806          Added metadata support; moved logging to J2SE logging framework.
** 056 ECF 20131028          Ensure metadata databases are pre-loaded, even if not auto-connect
**                           (TODO: this may not be optimal for databases which are never or
**                           seldom connected, since this will consume memory for no good reason).
**                           Added EntityResolver parameter to getBackingTableName variant which
**                           reads hbm files, to prevent error finding DTD when launching server
**                           offline. Added lock metadata support.
** 057 OM  20131029          Added support for dirty database registration for non
**                           load-at-startup databases.
**                           Fixed cleanup of localTransients context-local variable.
** 058 SVL 20131204          Do not remove schema mapping from TableMapper if there are other
**                           databases sharing the same schema.
** 059 ECF 20140116          Change to LockTableUpdater constructor signature.
** 060 SVL 20140205          Added dynamicComposites.
** 061 OM  20140417          Granted mapComputedCharacterColumns() package access.
** 062 CA  20140531          Added a weight for the context-local vars, to ensure predetermined 
**                           order during context reset.
** 063 ECF 20140611          Rolled back 061, refactored mapComputedCharacterColumns, extracting
**                           a worker method which also is called from registerDynamicTable.
**                           Combined multiple, context-local variables into inner class Context,
**                           stored under a single variable. Added context-local mapping for
**                           computed character column info, for dynamically generated tables.
** 064 ECF 20141204          Only populate a metaschema database once per schema, instead of once
**                           per physical database; allow qualified PDB names as directory IDs;
**                           reduce log noise when reconnecting a transient database. Don't
**                           initialize metaschema databases if MetadataManager did not
**                           initialize.
** 065 ECF 20141220          Allow early connection use during transient database registration to
**                           avoid premature dependency on Persistence instance.
** 066 OM  20150114          Fixed collection of processed meta databases in initialization.
** 067 ECF 20150123          Call ConnectionManager during initialization to initialize connection
**                           metadata support.
** 068 ECF 20150414          Log warning instead of throwing IllegalArgumentException if _temp
**                           database is configured in directory but there are no static DMOs
**                           defined for the application (in dmo_index.xml).
** 069 ECF 20150504          Optimized normalizeDMOClass to avoid hitting context local variable
**                           if possible.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.cfg.*;
import org.hibernate.dialect.*;
import org.hibernate.internal.util.xml.*;
import org.hibernate.mapping.*;
import org.hibernate.mapping.Collection;
import org.hibernate.service.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import com.goldencode.p2j.admin.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.persist.embed.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.persist.meta.*;
import com.goldencode.p2j.persist.remote.*;
import com.goldencode.p2j.security.*;
import com.goldencode.util.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.LogHelper;

/**
 * Manages configuration for all databases accessed by the current server,
 * and provides access to Hibernate session factories to create new sessions.
 * This class is a helper for the {@link Persistence} class, but since it
 * contains generally useful methods, it has been made public.
 * <p>
 * During server bootstrap, this class reads configuration data from the
 * directory service.  This information includes the names of databases which
 * must be configured and be permanently accessible, as well as those which
 * are considered "transient".  A transient database is one which can be
 * connected to and disconnected from (in the Progress sense) on demand, from
 * application code.
 * <p>
 * When a database is initialized for use via Hibernate, whether permanent
 * or transient, additional configuration data, including Hibernate's
 * configuration properties, is read from the directory service.  This is
 * assembled into a <code>Properties</code> object and passed to Hibernate's
 * configuration subsystem.
 * <p>
 * Each database is represented by a Hibernate <code>SessionFactory</code>
 * object, which is the object which the <code>Persistence</code> class uses
 * to create database <code>Session</code>s lazily.  In addition to the
 * configuration options read from the directory, each session factory
 * requires the list of ORM mapping classes which will be used to represent
 * database tables in any session established with that database.  These are
 * retrieved from a {@link DMOIndex}, one of which is created for each
 * physical database to which we must connect.
 * <p>
 * <code>SessionFactory</code> objects for transient databases are created on
 * demand, which can be an expensive process.  A session factory for a
 * transient database is required when at least one client session connects
 * dynamically to that database.  A reference count is maintained across
 * client sessions for each transient database, such that a session factory
 * need not be created for one client context, if another is already using it.
 * Session factory objects for transient databases thus are not discarded
 * while any context is connected to the associated database.
 * <p>
 * This class can both register remote databases for local use, as well as
 * serve requests from remote servers to provide access to local databases.
 * Whether a database is remote or local is transparent to users of this
 * class.  Both are registered using the {@link #registerDatabase(Database)}
 * method.
 * <p>
 * Implementation Note:  several methods of this class synchronizes on two
 * objects:
 * <ul>
 *   <li><code>configLock</code>
 *   <li><code>sessionFactories</code>
 * </ul>
 * <p>
 * <code>configLock</code> is the less granular of these. It is important
 * that any code path which requires access to both locks simultaneously
 * always acquire <code>configLock</code> first, then
 * <code>sessionFactories</code>, to avoid deadlock.
 */
public final class DatabaseManager
{
   /** Reserved name of temp table schema */
   public static final String TEMP_TABLE_SCHEMA = "_temp";
   
   /** Temp table database information object */
   public static final Database TEMP_TABLE_DB =
      new Database(TEMP_TABLE_SCHEMA, Database.Type.PRIMARY, true);
   
   /** Reserved name of primary key identifier column/property */
   public static final String PRIMARY_KEY = "id";
   
   /** Port for mixed mode, remote, SSL connections to embedded databases */
   public static final int MIXED_MODE_SSL_PORT = 7691;
   
   /** Logger */
   private static final Logger log = LogHelper.getLogger(DatabaseManager.class.getName());
   
   /** Map of local DatabaseConfig objects, keyed by database info objects */
   private static final Map<Database, DatabaseConfig> dbConfigs = new HashMap<>();
   
   /** Map of Hibernate session factories, keyed by database. */
   private static final Map<Database, SessionFactory> sessionFactories = new HashMap<>();
   
   /** Map of Hibernate configuration, keyed by database. */
   private static final Map<Database, Configuration> hibernateConfigs = new HashMap<>();
   
   /** Map of Hibernate configuration, keyed by schema. */
   private static final Map<String, Configuration> configBySchema = new HashMap<>();
   
   /** Map of Hibernate mappings, keyed by database schema. */
   private static final Map<String, Set<XmlDocument>> hibernateMappings = new HashMap<>();
   
   /** Map of dirty Hibernate mappings, keyed by database schema */
   private static final Map<String, Set<XmlDocument>> dirtyHibernateMappings = new HashMap<>();
   
   /** Map of schema names, keyed by database */
   private static final Map<Database, String> schemaByDatabase = new HashMap<>();
   
   /** Map of schema names, keyed by DMO implementation class */
   private static final Map<Class<?>, String> schemaByDMOClass = new HashMap<>();
   
   /** Map of schema names, keyed by DMO interface */
   private static final Map<Class<?>, String> schemaByDMOIface = new HashMap<>();
   
   /** Set of databases auto-connected at bootstrap */
   private static final Set<Database> permanentConnections = new HashSet<>();
   
   /** Map of databases to Hibernate <code>Dialect</code>s */
   private static final Map<Database, Dialect> dialects = new HashMap<>();
   
   /** Map of transient databases to reference counts of registrations */
   private static final Map<Database, Integer> transientDatabases = new HashMap<>();
   
   /** Set of all databases managed by the local P2J server instance */
   private static final Set<Database> managed = new LinkedHashSet<>();
   
   /** Map of DMO classes to lists of non-nullable properties in those DMOs */
   private static final Map<Class<?>, ArrayList<String>> notNullProperties = new HashMap<>();
   
   /** Map of DMO classes to maps of property names to composite names */
   private static final Map<Class<?>, Map<String, String>> composites = new HashMap<>();
   
   /**
    * Map of databases to maps of primary table names to lists of Hibernate <code>Table</code>
    * objects (whew!)
    */
   private static final Map<Database, Map<String, List<Table>>> secondaryTables = new HashMap<>();
   
   /** Context local data */
   private static final ContextLocal<Context> context = new ContextLocal<Context>()
   {
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_6;
      }
      
      protected Context initialValue()
      {
         return new Context();
      }
      
      @Override
      protected void cleanup(Context value)
      {
         value.cleanup();
      }
   };
   
   /** Map of computed column base property names to ignore-case booleans */
   private static final Map<Class<?>, Map<String, Boolean>> compCharColInfo =
      new IdentityHashMap<>();
   
   /** Object upon which configuration routines synchronize */
   private static final Object configLock = new Object();
   
   /** Map of databases to maps of table names to entity names for these databases */
   private static final Map<Database, Map<String, String>> tablesToEntities =
      new HashMap<Database, Map<String, String>>();
   
   /** Map of LockTableUpdater instances, keyed by database */
   private static Map<Database, LockTableUpdater> lockTableUpdaters = null;
   
   /** Deques of JDBC connections in use during initialization, by database */
   private static Map<Database, Deque<Connection>> earlyConnections = new HashMap<>();
   
   /**
    * This class cannot be instantiated;  all methods are static.
    */
   private DatabaseManager()
   {
   }
   
   /**
    * Retrieve the session factory for the given database.
    *
    * @param   database
    *          Database information object.
    *
    * @return  Session factory, or <code>null</code> if there is no factory
    *          registered for <code>database</code>.
    */
   public static SessionFactory getSessionFactory(Database database)
   {
      synchronized (sessionFactories)
      {
         return sessionFactories.get(database);
      }
   }
   
   /**
    * Get Hibernate <code>Configuration</code> object for the given database.
    *
    * @param  database
    *         Database which <code>Configuration</code> we want to get.
    *
    * @return Hibernate <code>Configuration</code> object.
    */
   public static Configuration getHibernateConfiguration(Database database)
   {
      synchronized (sessionFactories)
      {
         return hibernateConfigs.get(database);
      }
   }

   /**
    * Get Hibernate <code>Configuration</code> object for the given schema (however note that
    * from Hibernate 4 we do not share the same <code>Configuration</code> for multiple
    * databases).
    *
    * @param  schema
    *         Schema which <code>Configuration</code> we want to get.
    *
    * @return Hibernate <code>Configuration</code> object.
    */
   public static Configuration getConfigBySchema(String schema)
   {
      return configBySchema.get(schema);
   }
   
   /**
    * Report the database dialect in use for the given database.
    *
    * @param   database
    *          Database information object.
    *
    * @return  Dialect in use for <code>database</code>.
    */
   public static Dialect getDialect(Database database)
   {
      synchronized (configLock)
      {
         return dialects.get(database);
      }
   }
   
   /**
    * Lookup the schema name associated with the given physical database
    * name.
    *
    * @param   database
    *          Database information object.  May be <code>null</code> if
    *          <code>dmoClass</code> represents a temporary table.
    *
    * @return  Schema name for the given database.
    */
   public static String getSchema(Database database)
   {
      synchronized (configLock)
      {
         return schemaByDatabase.get(database);
      }
   }
   
   /**
    * Lookup the schema name associated with the given DMO implementation
    * class.
    *
    * @param   dmoClass
    *          A DMO implementation class;  if a temporary DMO, the temp
    *          table schema name is returned.
    *
    * @return  Schema name for the given class.
    */
   public static String getSchemaByClass(Class<?> dmoClass)
   {
      synchronized (configLock)
      {
         String schema = null;
         Class<?> nextClass = dmoClass;
         Class<?> rootClass = Object.class;
         
         // We have to loop our way up the superclass chain in case dmoClass
         // is actually a Hibernate CGLIB proxy class injected as a proxy
         // subclass of a "primary" DMO class.  This could have happened if
         // the caller tried to divine dmoClass by invoking getClass() on a
         // DMO instance created by Hibernate.  Only the primary DMO classes
         // are mapped to schema names.
         Set<Class<?>> dynamic = context.get().dynamicDMOClasses;
         
         while (schema == null && !rootClass.equals(nextClass))
         {
            schema = schemaByDMOClass.get(nextClass);
            if (schema == null && dynamic.contains(nextClass))
            {
               schema = DatabaseManager.TEMP_TABLE_SCHEMA;
            }
            nextClass = nextClass.getSuperclass();
         }
         
         return schema;
      }
   }
   
   /**
    * Lookup the schema name associated with the given DMO interface.
    *
    * @param   dmoIface
    *          A DMO interface;  if a temporary DMO, the temp table schema
    *          name is returned.
    *
    * @return  Schema name for the given interface.
    */
   public static String getSchemaByInterface(Class<?> dmoIface)
   {
      synchronized (configLock)
      {
         String res = schemaByDMOIface.get(dmoIface);
         if (res == null && context.get().dynamicDMOIfaces.contains(dmoIface))
         {
            res = DatabaseManager.TEMP_TABLE_SCHEMA;
         }
         return res;
      }
   }
   
   /**
    * Retrieve the database table name for a given DMO implementation class,
    * by reading it directly from the .hbm.xml document.  
    *
    * @param   dmoClass
    *          DMO implementation class to which the table is mapped via ORM,
    *          or <code>null</code> if couldn't be determined.
    *
    * @return  Name of the database table as determined from the .hbm.xml 
    *          document.
    */
   public static String getBackingTableName(Class<?> dmoClass, EntityResolver resolver)
   throws PersistenceException
   {
      String clazzName = dmoClass.getSimpleName();
      String fileName = clazzName + ".hbm.xml";
      String table = null;
      try
      {
         InputStream in = dmoClass.getResourceAsStream(fileName);
         Document doc = XmlHelper.parse(in, true, resolver, null);
         
         NodeList nodes = doc.getElementsByTagName("class");
         for (int i = 0; i < nodes.getLength(); i++)
         {
            Node n = nodes.item(i);
            String clazz = n.getAttributes().getNamedItem("name").getTextContent();
            
            if (!clazzName.equals(clazz))
            {
               continue;
            }
            
            table = n.getAttributes().getNamedItem("table").getTextContent();
            break;
         }
      }
      catch (Exception e)
      {
         // couldn't read the .hbm.xml file, stop server initialization
         throw new PersistenceException("Couldn't read " + fileName + " document ", e);
      }
      
      return table;
   }
   
   /**
    * Retrieve the database table name for a given database and DMO
    * implementation class.  This is necessary for use with the lock
    * manager.
    *
    * @param   database
    *          Database in which the table resides.
    * @param   dmoClass
    *          DMO implementation class to which the table is mapped via ORM.
    *
    * @return  Name of the database table as specified in the ORM
    *          configuration, or <code>null</code> if no match is found.
    */
   public static String getBackingTableName(Database database,
                                            Class<?> dmoClass)
   {
      Table table = getBackingTable(database, dmoClass);
      
      return (table != null ? table.getName().intern() : null);
   }
   
   /**
    * Get an iterator on all configured, permanent databases.
    * <p>
    * Although this method returns a technically thread-unsafe iterator, the
    * set of permanent databases over which it iterates is established during
    * server bootstrap and is not changed thereafter.  Thus, so long as this
    * method is invoked after bootstrap, the iterator effectively will be
    * thread-safe.
    *
    * @return  Iterator over permanent database information objects.
    */
   public static Iterator<Database> permanentConnections()
   {
      synchronized (configLock)
      {
         return permanentConnections.iterator();
      }
   }
   
   /**
    * Get a list of all configured databases managed by this P2J server
    * instance.
    * 
    * @return  List of database information objects managed by this server
    *          instance.
    */
   public static List<Database> getManagedDatabases()
   {
      synchronized (configLock)
      {
         return (new ArrayList<Database>(managed));
      }
   }
   
   /**
    * Gather snapshots of all record locks for all, non-temporary, permanently
    * connected databases which are authoritatively managed by this P2J server
    * instance.  Note that this data is for informational, reporting purposes
    * only.  It is guaranteed to be consistent (per database) at the moment it
    * is collected, but it is possible (likely) for this information to become
    * stale quickly.
    * 
    * @return  Record lock information for all non-temporary databases managed
    *          by this server.
    */
   public static RecordLockInfo[] getAllRecordLockInfo()
   {
      List<RecordLockInfo> lockInfo = new ArrayList<RecordLockInfo>();
      
      for (Database database : getManagedDatabases())
      {
         Persistence persistence = PersistenceFactory.getInstance(database);
         LockManager manager = persistence.getLockManager();
         LockAdministrator admin = manager.getLockAdministrator();
         if (admin != null)
         {
            lockInfo.addAll(admin.collectRecordLockInfo());
         }
      }
      
      return lockInfo.toArray(new RecordLockInfo[lockInfo.size()]);
   }

   /**
    * Get the entity name which corresponds the given table of the given
    * database.
    *
    * @param   database
    *          Database which contains the target table.
    * @param   table
    *          Target table
    *
    * @return  Entity name which corresponds the given table.
    */
   public static String getEntityNameByTableName(Database database, String table)
   {
      synchronized (tablesToEntities)
      {
         Map<String, String> databaseTables = tablesToEntities.get(database);

         String entityName = null;
         if (databaseTables != null)
         {
            entityName = databaseTables.get(table);
         }

         if (entityName != null)
         {
            return entityName;
         }

         Configuration cfg =
                          DatabaseManager.getHibernateConfiguration(database);
         Iterator iter = cfg.getClassMappings();
         while (iter.hasNext())
         {
            Object next = iter.next();
            if (next instanceof RootClass)
            {
               RootClass rc = (RootClass) next;
               if (table.equals(rc.getTable().getName()))
               {
                  entityName = rc.getEntityName();
                  if (databaseTables == null)
                  {
                     databaseTables = new HashMap<String, String>();
                     tablesToEntities.put(database, databaseTables);
                  }

                  databaseTables.put(table, entityName);
                  break;
               }
            }
         }

         return entityName;
      }
   }

   /**
    * Get the tables of the given database.
    *
    * @param   database
    *          Target database.
    *
    * @return  the tables of the given database.
    */
   public static List<String> getDatabaseTables(Database database)
   {
      Configuration cfg = DatabaseManager.getHibernateConfiguration(database);
      if (cfg == null)
      {
         return null;
      }

      List<String> tableNames = new ArrayList<String>();
      Iterator iter = cfg.getClassMappings();
      while (iter.hasNext())
      {
         Object obj = iter.next();
         if (obj instanceof RootClass)
         {
            RootClass rc = (RootClass) obj;
            tableNames.add(rc.getTable().getName());
         }
      }
      
      return tableNames;
   }
   
   /**
    * Given a DMO implementation class, return the associated class which is
    * registered with the <code>DatabaseManager</code>.  In most cases, the
    * class returned will be the same as the <code>dmoClass</code> argument.
    * However, there is a possibility that this argument is actually a
    * Hibernate CBLIB proxy class injected as a proxy subclass of the
    * "primary", registered class.  This could have happened if the caller
    * tried to divine dmoClass by invoking getClass() on a DMO instance
    * created by Hibernate.  In this case, we need to walk up the superclass
    * chain until we find the primary class.
    *
    * @param   dmoClass
    *          DMO implementation class to be normalized.
    *
    * @return  The registered, "primary" DMO implementation class associated
    *          with <code>dmoClass</code>.  This may be <code>dmoClass</code>
    *          or its superclass.
    *
    * @throws  IllegalArgumentException
    *          if no class associated with <code>dmoClass</code> is
    *          registered.
    */
   public static Class<?> normalizeDMOClass(Class<?> dmoClass)
   {
      // We have to loop our way up the superclass chain in case dmoClass is
      // actually a Hibernate CGLIB proxy class injected as a proxy subclass
      // of a "primary" DMO class.  This could have happened if the caller
      // tried to divine dmoClass by invoking getClass() on a DMO instance
      // created by Hibernate.  Only the primary DMO classes are mapped to
      // values in the schemaByDMOClass map.
      Set<Class<?>> dynamic = null;
      Class<?> rootClass = Object.class;
      Class<?> nextClass = dmoClass;
      
      synchronized (configLock)
      {
         while (!rootClass.equals(nextClass))
         {
            if (schemaByDMOClass.containsKey(nextClass))
            {
               return nextClass;
            }
            
            if (dynamic == null)
            {
               dynamic = context.get().dynamicDMOClasses;
            }
            
            if (dynamic.contains(nextClass))
            {
               return nextClass;
            }
            
            nextClass = nextClass.getSuperclass();
         }
      }
      
      throw new IllegalArgumentException("Unknown DMO implementation class:  " + dmoClass);
   }
   
   /**
    * Get a mapping of column names to property names for the given DMO in the
    * given database.
    * 
    * @param   database
    *          Target database.
    * @param   dmoClass
    *          Target DMO implementation class.
    * @param   columnNames
    *          Set of column names to be mapped to properties.  If
    *          <code>null</code>, all of the table's columns will be mapped to
    *          DMO property names.
    * @param   includePK
    *          <code>true</code> to include the reserved mapping of primary
    *          "id" to reserved property name "id";  <code>false</code> to
    *          omit this mapping.
    *          
    * @return  A map whose keys are the given set of column names, and whose
    *          values are the corresponding DMO property names.
    */
   public static Map<String, String> getColumnToPropertyMap(Database database,
                                                            Class<?> dmoClass,
                                                            Set<String> columnNames,
                                                            boolean includePK)
   {
      PersistentClass pc = getClassMapping(database, dmoClass);
      if (pc == null)
      {
         return null;
      }
      
      Map<String, String> map = new HashMap<String, String>();
      
      Iterator<Property> iter = pc.getPropertyIterator();
      while (iter.hasNext())
      {
         Property prop = iter.next();
         Iterator<Column> cols = prop.getColumnIterator();
         
         if (cols.hasNext())
         {
            // Assume property is mapped to a single column.
            Column column = cols.next();
            String colName = column.getName();
            if (columnNames == null || columnNames.contains(colName))
            {
               map.put(colName, prop.getName());
            }
         }
      }
      
      if (includePK)
      {
         map.put(PRIMARY_KEY, PRIMARY_KEY);
      }
      
      // Have we mapped all requested column names to property names?
      // Check is disabled if foreign keys are disabled, because we may have
      // a false alarm for __fk columns.
      Set<String> mappedCols = map.keySet();
      if (columnNames != null &&
          Persistence.isForeignKeysEnabled() &&
          !mappedCols.containsAll(columnNames))
      {
         Set<String> missing = new HashSet<String>(columnNames);
         missing.removeAll(mappedCols);
         
         throw new IllegalArgumentException(
            "Column(s) " + missing + " not mapped for " + dmoClass.getName());
      }
      
      return map;
   }
   
   /**
    * Get a JDBC connection, either from the appropriate {@link Persistence}
    * instance, or if during persistence initialization, directly from the
    * JDBC driver.
    * <p>
    * Note:  since early connections are recycled using a crude connection
    * pool, the caller must not call <code>close()</code> on the returned
    * connection;  rather, {@link #checkInConnection(Connection, Database)}
    * must be used.  All open connections are closed when the persistence
    * runtime is finished initializing.
    * 
    * @param   database
    *          Database to which a connection is needed.
    * 
    * @return  JDBC connection for the given database.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    */
   public static Connection checkOutConnection(Database database)
   throws PersistenceException
   {
      boolean needEarlyConnection = isInitializing();
      if (needEarlyConnection)
      {
         Deque<Connection> deque = earlyConnections.get(database);
         if (deque == null)
         {
            deque = new ArrayDeque<Connection>();
            earlyConnections.put(database, deque);
         }
         
         if (!deque.isEmpty())
         {
            Connection conn = deque.pop();
            
            if (log.isLoggable(Level.FINE))
            {
               log.log(Level.FINE,
                       "Early connection checked out for " +
                        database +
                        ";  pool count:  " +
                        deque.size());
            }
            
            return conn;
         }
         
         return getEarlyConnection(database);
      }
      else
      {
         synchronized (sessionFactories)
         {
            needEarlyConnection = !sessionFactories.containsKey(database);
         }
      }
      
      if (needEarlyConnection)
      {
         return getEarlyConnection(database);
      }
      
      Persistence persistence = PersistenceFactory.getInstance(database);
      
      return persistence.checkOutConnection();
   }
   
   /**
    * Check in a JDBC connection previously obtained with {@link
    * #checkOutConnection(Database)}.  This will close the connection if it is
    * not otherwise in use.
    * 
    * @param   connection
    *          JDBC connection which is no longer needed by the caller.
    * @param   database
    *          Database associated with the returned connection.
    *          
    * @throws  PersistenceException
    *          if there is an error closing the session associated with the
    *          connection.
    */
   public static void checkInConnection(Connection connection, Database database)
   throws PersistenceException
   {
      boolean needEarlyConnection = isInitializing();
      if (needEarlyConnection)
      {
         Deque<Connection> deque = earlyConnections.get(database);
         deque.push(connection);
         
         if (log.isLoggable(Level.FINE))
         {
            log.log(Level.FINE,
                    "Early connection checked in for " +
                    database +
                    ";  pool count:  " +
                    deque.size());
         }
         
         return;
      }
      else
      {
         synchronized (sessionFactories)
         {
            if (!sessionFactories.containsKey(database))
            {
               return;
            }
         }
      }
      
      Persistence persistence = PersistenceFactory.getInstance(database);
      persistence.checkInConnection(connection);
   }
   
   /**
    * If the given combination of DMO class and property indicate a property
    * of a composite element of a collection associated with the DMO class,
    * return the name of the collection as defined in the Hibernate mapping
    * for the DMO class.
    *
    * @param   dmoClass
    *          A DMO implementation class.
    * @param   property
    *          A property of <code>dmoClass</code> or of the component stored
    *          in the associated collection.
    *
    * @return  Name of the associated collection of composite-elements.
    */
   static String getComposite(Class<?> dmoClass, String property)
   {
      Map<String, String> map = composites.get(dmoClass);
      
      if (map == null)
      {
         map = context.get().dynamicComposites.get(dmoClass);
      }
      
      return (map != null ? map.get(property) : null);
   }
   
   /**
    * Helper method to determine whether the specified class or interface
    * represents the DMO front-end for temp table records.
    *
    * @param   dmoClass
    *          A DMO interface or implementation class.
    *
    * @return  <code>true</code> if <code>dmoClass</code> is assignable from
    *          the {@link Temporary} marker interface;  else
    *          <code>false</code>.
    */
   static boolean isTempTable(Class<?> dmoClass)
   {
      return Temporary.class.isAssignableFrom(dmoClass);
   }
   
   /**
    * Check if the specified database is a permanent database. This is <code>true</code> only if
    * the database is neither a dirty database nor the temporary database.
    * 
    * @param    database
    *           The database to be checked.
    *           
    * @return   <code>true</code> if this database object is configured for a 
    *           permanent DB.
    */
   static boolean isPermamentDB(Database database)
   {
      return (!database.isDirty() && !database.equals(DatabaseManager.TEMP_TABLE_DB));
   }
   
   /**
    * Get an iterator on the names of all properties for a given DMO class,
    * for which the mapped column is not nullable.
    *
    * @param   database
    *          Database in which the table mapped to the DMO represented by
    *          <code>dmoClass</code> resides.
    * @param   dmoClass
    *          DMO implementation class.
    * 
    * @return  Iterator on property names.
    */
   static Iterator<String> notNullProperties(Database database, Class<?> dmoClass)
   {
      ArrayList<String> props;
      
      synchronized (notNullProperties)
      {
         props = notNullProperties.get(dmoClass);
         if (props == null)
         {
            PersistentClass pc = getClassMapping(database, dmoClass);
            if (pc == null)
            {
               return new EmptyIterator<String>();
            }
            
            // Create a list of property names for those properties mapped to
            // a non-nullable column.
            props = new ArrayList<String>();
            notNullProperties.put(dmoClass, props);
            Iterator<Property> iter = pc.getPropertyIterator();
            while (iter.hasNext())
            {
               Property prop = iter.next();
               Iterator<Column> cols = prop.getColumnIterator();
               
               if (cols.hasNext())
               {
                  // Assume property is mapped to a single column.
                  Column column = cols.next();
                  if (!column.isNullable())
                  {
                     props.add(prop.getName());
                  }
               }
            }
            props.trimToSize();
         }
      }
      
      return props.iterator();
   }
   
   /**
    * Initialize the persistence framework. This entails configuring and
    * storing a Hibernate session factory for each database to be supported.
    * This may be done immediately (for databases flagged as "auto-connect"),
    * or may be deferred for databases which are dynamically connected and
    * disconnected at runtime. This information will be retrieved from the
    * P2J directory.
    * <p>
    * Remote database service provider exports are registered at this time as
    * well.  This includes services to fetch database configurations for
    * remote requesters, record locking services, and the tracking of dirty
    * records within uncommitted transactions.
    * <p>
    * Databases and the DMO classes associated with them are determined by
    * reading the DMO index configuration file. The location of this file is
    * read from the P2J directory. Hibernate configuration properties for
    * each session factory are read from the directory as well.
    * <p>
    * The logic flow is:
    * <ol>
    *   <li>read from the directory the location of the DMO index document;
    *   <li>parse the DMO index document and compile a list of fully
    *       qualified class names for each schema node in the document;
    *       store these in a map, keyed by schema name.
    *   <li>read from the directory the list of physical database names of
    *       those databases supported by the current server;
    *   <li>for each database:
    *   <ul>
    *     <li>get the name of its schema, which should match one of the
    *         names mapped to a DMOIndex instance;
    *     <li>determine whether it must be auto-connected.
    *   </ul>
    *   <li>if a database is to be auto-connected:
    *   <ul>
    *     <li>create a Hibernate <code>Configuration</code>;
    *     <li>retrieve all Hibernate properties for that database from the
    *         directory and add them to the configuration;
    *     <li>add to the configuration each DMO class which corresponds to
    *         that database's schema name, as read from the DMO index;
    *     <li>build a Hibernate <code>SessionFactory</code> from the
    *         configuration and store it in a map, keyed by the database
    *         name (not the schema name, which may be different)
    *     <li>add the database name to the set of permanent databases which
    *         have been configured.
    *   </ul>
    *   <li>if a database is <i>not</i> to be auto-connected:
    *   <ul>
    *     <li>defer session factory configuration until such time as a
    *         dynamic configuration request is processed (i.e., do nothing
    *         at this time).
    *   </ul>
    * </ol>
    *
    * @throws  IllegalStateException
    *          if invoked more than once.
    * @throws  PersistenceException
    *          if any error is encountered initializing the persistence
    *          framework.
    */
   static void initialize()
   throws PersistenceException
   {
      synchronized (configLock)
      {
         try
         {
            synchronized (sessionFactories)
            {
               if (!sessionFactories.isEmpty())
               {
                  throw new IllegalStateException("Persistence is already initialized");
               }
            }
            
            DirectoryService ds = DirectoryService.getInstance();
            if (!ds.bind())
            {
               throw new RuntimeException("Directory bind failed");
            }
            
            try
            {
               boolean useMeta = false;
               
               String dbID = Utils.findDirectoryNodePath(ds, "database", "container", false);
               
               // Read the names of all permanently connected databases.
               String[] dbNodes = (dbID == null ? null : ds.enumerateNodes(dbID));
               if (dbNodes == null || dbNodes.length == 0)
               {
                  if (log.isLoggable(Level.WARNING))
                  {
                     log.log(Level.WARNING, "No databases are configured");
                  }
               }
               else
               {
                  if (log.isLoggable(Level.INFO))
                  {
                     log.log(Level.INFO, "Using H2 database version " + H2Helper.getH2Version());
                  }
                  
                  useMeta = MetadataManager.initialize();
                  
                  if (useMeta)
                  {
                     // determine whether we need to create LockTableUpdaters
                     try
                     {
                        // if the MetaLock interface is found, we will need to register listeners
                        // with our lock managers
                        String metaIface = String.format("%s._meta.MetaLock",
                                                         DMOIndex.getBasePackage());
                        Class.forName(metaIface);
                        lockTableUpdaters = new HashMap<Database, LockTableUpdater>();
                     }
                     catch (ClassNotFoundException exc)
                     {
                        // this is OK, just means the current application does not need to update
                        // the _lock metadata table with lock change information
                     }
                  }
                  
                  HashSet<String> metaDBs = new HashSet<>();
                  
                  // Collect name and permanence data.
                  for (String nodeID : dbNodes)
                  {
                     String pdb = nodeID.replace('.', File.separatorChar);
                     Database primaryDB = new Database(pdb, Database.Type.PRIMARY, true);
                     Database metaDB = null;
                     Database[] dbs = null;
                     
                     if (useMeta && !TEMP_TABLE_DB.equals(primaryDB))
                     {
                        // create and prepare the physical meta database
                        metaDB = new Database(pdb, Database.Type.META, true);
                        MetadataManager.prepareDatabase(metaDB);
                        dbs = new Database[] { metaDB, primaryDB };
                     }
                     else
                     {
                        dbs = new Database[] { primaryDB };
                     }
                     
                     // determine auto-connect status
                     Boolean perm = ds.getNodeBoolean(
                        dbID + "/" + nodeID + "/p2j/load_at_startup",
                        "value");
                     
                     for (Database database : dbs)
                     {
                        // read and cache the database configuration info
                        DatabaseConfig cfg = getLocalConfiguration(database);
                        
                        if (!TEMP_TABLE_DB.equals(database))
                        {
                           // remember that this server is authoritative for this database (not
                           // including temp table database)
                           managed.add(database);
                           
                           if (useMeta && database.isPrimary() && lockTableUpdaters != null)
                           {
                              // create and store lock table updater for primary database, if
                              // needed; will be registered with lock manager when that object is
                              // created later; creating these now ensures the proper server
                              // context is used for the associated worker thread created by the
                              // updater
                              LockTableUpdater ltu = new LockTableUpdater(metaDB);
                              lockTableUpdaters.put(primaryDB, ltu);
                           }
                        }
                        
                        // if database must be auto-connected or is a meta database, register it
                        // now
                        if ((perm != null && perm) || database.isMeta())
                        {
                           // register database
                           registerDatabase(cfg, database, true);
                           
                           if (useMeta && !TEMP_TABLE_DB.equals(primaryDB))
                           {
                              // initialize infrastructure for _connect metadata
                              ConnectionManager.initializeMeta(primaryDB, metaDB);
                           }
                           
                           // preload Persistence object for database
                           PersistenceFactory.getInstance(database);
                           
                           // register a dirty database, only if this is a non-temp database
                           if (database.isPrimary() && !TEMP_TABLE_DB.equals(database))
                           {
                              registerDirtyDatabase(cfg, database, Database.Type.DIRTY, true);
                           }
                        }
                     }
                     
                     if (useMeta && metaDB != null)
                     {
                        DatabaseConfig cfg = getLocalConfiguration(primaryDB);
                        String schema = cfg.getSchema();
                        if (!metaDBs.contains(schema))
                        {
                           metaDBs.add(schema);
                           
                           // populate the meta database
                           MetadataManager.populateDatabase(metaDB, schema);
                           
                           // initialize the meta persistence's lock table updater to be that of
                           // the primary database
                           if (lockTableUpdaters != null)
                           {
                              Persistence mPrim = PersistenceFactory.getInstance(metaDB);
                              mPrim.shareLockTableUpdater(primaryDB);
                           }
                        }
                     }
                  }
               }
            }
            finally
            {
               // Done with directory.
               ds.unbind();
            }
         }
         finally
         {
            try
            {
               for (Deque<Connection> deque : earlyConnections.values())
               {
                  for (Connection conn : deque)
                  {
                     try
                     {
                        if (!conn.isClosed())
                        {
                           conn.close();
                        }
                     }
                     catch (SQLException exc)
                     {
                        log.log(Level.SEVERE,
                                "Error cleaning up early database connections",
                                exc);
                     }
                  }
               }
            }
            finally
            {
               earlyConnections = null;
            }
         }
      }
      
      // Initialize remote database service provider exports.
      
      // Database configuration fetcher.
      DatabaseConfigFetcher dcf = new DatabaseConfigFetcherImpl();
      RemoteObject.registerNetworkServer(DatabaseConfigFetcher.class, dcf);
      
      // Lock manager multiplexer.
      LockManagerMultiplexerImpl.initialize();
      
      // Dirty share multiplexer.
      DirtyShareMultiplexerImpl.initialize();
      
      // Identity manager multiplexer.
      IdentityManagerMultiplexerImpl.initialize();
   }
   
   /**
    * Indicate whether the database manager is currently initializing.  This
    * will only ever return <code>true</code> if called on the same thread as
    * the thread on which the initialization occurs, since it is synchronized
    * to the same lock which the {@link #initialize()} method uses.
    * 
    * @return  <code>true</code> if database manager is initializing, else
    *          <code>false</code> if initialization is complete.
    */
   static boolean isInitializing()
   {
      synchronized (configLock)
      {
         return (earlyConnections != null);
      }
   }
   
   /**
    * Obtain a JDBC connection for the given database during persistence initialization.
    * 
    * @param   database
    *          Database to which a connection is needed.
    * 
    * @return  JDBC connection for the given database.
    * 
    * @throws  PersistenceException
    *          if there is any database error.
    * @throws  IllegalStateException
    *          if this method is invoked after the persistence layer or specific database is no
    *          longer initializing.
    */
   static Connection getEarlyConnection(Database database)
   throws PersistenceException
   {
      if (!isInitializing())
      {
         synchronized (sessionFactories)
         {
            if (sessionFactories.containsKey(database))
            {
               throw new IllegalStateException("Early connection no longer available");
            }
         }
      }
      
      Connection conn = null;
      
      try
      {
         DatabaseConfig cfg = getLocalConfiguration(database);
         Properties props = cfg.getHibernateProperties();
         String url = props.getProperty(Environment.URL);
         String user = props.getProperty(Environment.USER);
         String pass = props.getProperty(Environment.PASS);
         conn = DriverManager.getConnection(url, user, pass);
      }
      catch (SQLException exc)
      {
         throw new PersistenceException(
            "Error obtaining early JDBC connection for database " +
            database,
            exc);
      }
      
      if (log.isLoggable(Level.FINE))
      {
         log.log(Level.FINE, "Early connection created for " + database);
      }
      
      return conn;
   }
   
   /**
    * Register an individual database with the persistence framework. This
    * configures the database to allow the framework to access it.
    * Registration involves the following:
    * <ul>
    *   <li>create a Hibernate <code>Configuration</code>;
    *   <li>retrieve all Hibernate properties for that database from the
    *       directory and add them to the configuration;
    *   <li>add to the configuration each DMO class which corresponds to
    *       that database's schema name, as read from the DMO index;
    *   <li>build a Hibernate <code>SessionFactory</code> from the
    *       configuration and store it in a map, keyed by the database
    *       name (not the schema name, which may be different)
    *   <li>if the database is transient (i.e., not configured immediately
    *       during bootstrap), update its configuration reference count;
    *       otherwise, add the database name to the set of permanent
    *       databases which have been configured.
    * </ul>
    * <p>
    *
    * @param   database
    *          Database to be configured.
    *
    * @return  <code>true</code> if the database was newly registered;
    *          <code>false</code> if the database was already registered,
    *          and only the reference count was incremented, or if this
    *          registration attempt was redundant in the local context.
    *
    * @throws  PersistenceException
    *          if <code>database</code> is not recognized, based upon
    *          configuration information in the directory.
    */
   static boolean registerDatabase(Database database)
   throws PersistenceException
   {
      DatabaseConfig cfg = (database.isLocal()
                            ? getLocalConfiguration(database)
                            : getRemoteConfiguration(database));
      
      return registerDatabase(cfg, database, false);
   }

   /**
    * Given a normal database, this method will register its associated dirty database with the
    * persistence framework. This configures the database to allow the framework to access it.
    * The created dirty database will have the same name as the original database.
    * Registration involves the following:
    * <ul>
    *   <li>create a Hibernate <code>Configuration</code>;
    *   <li>retrieve all Hibernate properties for that database from the
    *       directory and add them to the configuration;
    *   <li>add to the configuration each DMO class which corresponds to
    *       that database's schema name, as read from the DMO index;
    *   <li>build a Hibernate <code>SessionFactory</code> from the
    *       configuration and store it in a map, keyed by the database
    *       name (not the schema name, which may be different)
    *   <li>if the database is transient (i.e., not configured immediately
    *       during bootstrap), update its configuration reference count;
    *       otherwise, add the database name to the set of permanent
    *       databases which have been configured.
    * </ul>
    * <p>
    *
    * @param   database
    *          Database to be configured.
    *
    * @return  <code>true</code> if the database was newly registered;
    *          <code>false</code> if the database was already registered,
    *          and only the reference count was incremented, or if this
    *          registration attempt was redundant in the local context.
    *
    * @throws  PersistenceException
    *          if <code>database</code> is not recognized, based upon
    *          configuration information in the directory.
    */
   static boolean registerDirtyDatabase(Database database)
   throws PersistenceException
   {
      DatabaseConfig cfg = (database.isLocal()
                            ? getLocalConfiguration(database)
                            : getRemoteConfiguration(database));
      
      return registerDirtyDatabase(cfg, database, Database.Type.DIRTY, false);
   }
   
   /**
    * Register dynamic table represented by the given DMO interface and persistent class.
    *
    * @param dmoIface
    *        DMO interface class.
    * @param persistentClass
    *        Persistent class.
    */
   static void registerDynamicTable(Class<?> dmoIface, PersistentClass persistentClass)
   throws PersistenceException
   {
      Context ctx = context.get();
      ctx.dynamicDMOIfaces.add(dmoIface);
      ctx.dynamicDMOClasses.add(persistentClass.getMappedClass());
      
      // map secondary tables
      List<Table> secondaries = null;
      Iterator<Property> iter = persistentClass.getPropertyIterator();
      while (iter.hasNext())
      {
         Property property = iter.next();
         if (property.getType().isCollectionType())
         {
            Collection collection = (Collection) property.getValue();
            
            Value element = collection.getElement();
            if (element instanceof Component)
            {
               mapCompositeProperties(collection, (Component) element, true);
            }
            
            if (secondaries == null)
            {
               secondaries = new ArrayList<>();
            }
            secondaries.add(collection.getCollectionTable());
         }
      }
      
      if (secondaries != null)
      {
         ctx.dynamicSecondaryTables.put(persistentClass.getTable().getName(), secondaries);
      }
      
      mapComputedCharacterColumnsImpl(TEMP_TABLE_SCHEMA,
                                      TEMP_TABLE_DB,
                                      persistentClass,
                                      ctx.compCharColInfo);
   }
   
   /**
    * Deregister dynamic table represented by the given DMO interface and persistent class.
    *
    * @param dmoIface
    *        DMO interface class.
    * @param persistentClass
    *        Persistent class.
    */
   static void deregisterDynamicTable(Class<?> dmoIface, PersistentClass persistentClass)
   {
      Class<?> dmoClass = persistentClass.getMappedClass();
      Context ctx = context.get();
      ctx.dynamicDMOIfaces.remove(dmoIface);
      ctx.dynamicDMOClasses.remove(dmoClass);
      ctx.dynamicSecondaryTables.remove(persistentClass.getTable().getName());
      ctx.dynamicComposites.remove(dmoClass);
      ctx.compCharColInfo.remove(dmoClass);
   }
   
   /**
    * Return the lock table updater associated with the given database, if any.
    * 
    * @param   database
    *          A primary database for which the lock table updater is needed.
    * 
    * @return  Lock table updater, or <code>null</code> if lock metadata is not used by the
    *          current application.
    * 
    * @throws  IllegalArgumentException
    *          if lock metadata is used by the current application, but no lock table updater is
    *          found for the given database.
    */
   static LockTableUpdater getLockTableUpdater(Database database)
   {
      if (lockTableUpdaters == null)
      {
         return null;
      }
      
      LockTableUpdater ltu = lockTableUpdaters.get(database);
      if (ltu == null)
      {
         throw new IllegalArgumentException("No lock table updater found for database " +
                                            database);
      }
      
      return ltu;
   }
   
   /**
    * Retrieve the Hibernate configuration for the given database schema.
    *
    * @param   schema
    *          Database schema name.
    *
    * @return  Configuration, or <code>null</code> if there is none
    *          registered for <code>database</code>.
    */
   private static Set<XmlDocument> getHibernateMappings(String schema)
   {
      synchronized (sessionFactories)
      {
         return hibernateMappings.get(schema);
      }
   }
   
   /**
    * Retrieve the dirty Hibernate configuration for the given database schema.
    *
    * @param   schema
    *          Database schema name.
    *
    * @return  Configuration, or <code>null</code> if there is none
    *          registered for <code>database</code>.
    */
   private static Set<XmlDocument> getDirtyHibernateMappings(String schema)
   {
      synchronized (sessionFactories)
      {
         return dirtyHibernateMappings.get(schema);
      }
   }
   
   /**
    * Get all data necessary to configure a database managed by the current
    * P2J server instance.  This information is read from the directory the
    * first time it is needed and then cached for future accesses.
    * 
    * @param   database
    *          A locally managed database.
    *          
    * @return  A configuration object which provides all information needed to
    *          connect to a physical database.
    *          
    * @throws  PersistenceException
    *          if an error occurs accessing configuration data.
    */
   private static DatabaseConfig getLocalConfiguration(Database database)
   throws PersistenceException
   {
      synchronized (configLock)
      {
         DatabaseConfig cfg = dbConfigs.get(database);
         if (cfg == null)
         {
            cfg = readDatabaseConfig(database);
            dbConfigs.put(database, cfg);
         }
         
         return cfg;
      }
   }
   
   /**
    * Get all data necessary to configure a database managed by a remote P2J server instance.
    * 
    * @param   database
    *          A remotely managed database.
    *          
    * @return  A configuration object which provides all information needed to connect to a
    *          physical database.
    * 
    * @throws  PersistenceException
    *          if the specified database is not managed by the remote server.
    */
   private static DatabaseConfig getRemoteConfiguration(Database database)
   throws PersistenceException
   {
      // Obtain database configuration fetcher proxy for local use.
      Session session = ConnectionManager.getSession(database);
      DatabaseConfigFetcher dcf = (DatabaseConfigFetcher)
         RemoteObject.obtainNetworkInstance(DatabaseConfigFetcher.class, session);
      DatabaseConfig config = dcf.fetch(database);
      
      if (config == null)
      {
         throw new PersistenceException(
            "Remote database configuration for '" + database + "' was not found");
      }
      
      // Validate and possibly modify connection URL.
      Properties remoteProps = config.getHibernateProperties();
      checkRemoteConnectURL(database, remoteProps);
      
      // Override connection pool settings.
      // TODO:  this will create a simple connection pool of up to 20
      // connections;  make this more robust/configurable.
      remoteProps.setProperty(
         Environment.CONNECTION_PROVIDER,
         "org.hibernate.service.jdbc.connections.internal.DriverManagerConnectionProviderImpl");
      
      return config;
   }
   
   /**
    * Validate and possibly modify the remote connection URL.  Validation
    * consists of confirming the existence of the appropriate property in the
    * <code>props</code> object and confirming it begins with the necessary
    * <code>jdbc:</code> prefix.
    * <p>
    * It is modified in the event it specifies a localhost/loopback address.
    * Such an address specification is relative to the remote server and is
    * most likely invalid for the requesting server.  The host in such a case
    * is replaced the host name specified in <code>database</code>, and the
    * modified URL is set back into <code>props</code> under the appropriate
    * key.
    * <p>
    * NOTE:  this approach assumes the local and remote servers share a DNS
    * and will resolve the given connection URL to the same physical address.
    *   
    * @param   database
    *          Object containing address information for the remotely managed
    *          database.
    * @param   props
    *          Hibernate configuration properties as reported by the remote
    *          server.
    * 
    * @throws  PersistenceException
    *          if the connection URL specified in <code>props</code> is
    *          missing or invalid.
    */
   private static void checkRemoteConnectURL(Database database, Properties props)
   throws PersistenceException
   {
      String url = props.getProperty(Environment.URL);
      if (url == null)
      {
         throw new PersistenceException(
            "Remote database configuration does not specify connection URL");
      }
      
      if (!url.startsWith("jdbc:"))
      {
         throw new PersistenceException("Invalid database connection URL [" + url + "]");
      }
      
      P2JDialect dialect = (P2JDialect) Dialect.getDialect(props);
      url = dialect.buildRemoteURL(url, database);
      
      // Update hibernate properties.
      props.setProperty(Environment.URL, url);
   }
   
   /**
    * Read from the directory all data necessary to configure a database
    * managed by the current P2J server instance.
    * 
    * @param   database
    *          A locally managed database.
    *          
    * @return  A configuration object which provides all information needed to
    *          connect to a physical database.
    *          
    * @throws  PersistenceException
    *          if an error occurs accessing configuration data.
    */
   private static DatabaseConfig readDatabaseConfig(Database database)
   throws PersistenceException
   {
      Properties props = null;
      String schema = null;
      
      if (database.isMeta())
      {
         props = createEmbeddedProperties(database);
         schema = MetadataManager.META_SCHEMA;
      }
      else
      {
         DirectoryService ds = DirectoryService.getInstance();
         if (!ds.bind())
         {
            throw new PersistenceException("Directory bind failed");
         }
         
         // get schema name used by this database, which may not be same as physical database name
         String nodeID = Utils.findDirectoryNodePath(ds, "database", "container", false);
         if (nodeID != null)
         {
            String path = String.format("%s/%s/p2j/schema", nodeID, database.getId());
            schema = ds.getNodeString(path, "value");
         }
         
         ds.unbind();
         
         // If schema name is not explicitly mapped in the directory, assume
         // it is the same as the physical database name.
         if (schema == null)
         {
            schema = database.getName();
         }
         
         props = loadConfigProperties(database);
      }
      
      return (new DatabaseConfig(props, schema));
   }
   
   /**
    * Deregister the specified, transient database, if and only if all other
    * contexts using the database have deregistered it. After deregistration,
    * sessions can no longer be opened for <code>database</code>, without
    * first building a new session factory (an expensive process).
    *
    * @param   database
    *          Database to be deregistered.
    *
    * @return  <code>true</code> if <code>database</code> was deregistered;
    *          else <code>false</code>. Note that the latter does not
    *          represent an error condition, but rather indicates that the
    *          database could not be deregistered because other contexts
    *          still referenced it, or because it is a permanent database.
    *
    * @throws  IllegalArgumentException
    *          if <code>database</code> is not currently registered.
    */
   static boolean deregisterDatabase(Database database)
   {
      synchronized (configLock)
      {
         // Permanent databases do not get deregistered.
         if (permanentConnections.contains(database))
         {
            return false;
         }
         
         if (!transientDatabases.containsKey(database))
         {
            throw new IllegalArgumentException("Database '" + database + "' not registered");
         }
      }
      
      boolean result = false;
      
      DirtyShareFactory.unregisterContextInstance(database);
      
      // If no one is using the database any longer, clean up most resources associated with it.
      if (updateTransientRefCount(database, false) == 0)
      {
         // clean up local context before closing session factory
         PersistenceFactory.getInstance(database).cleanup();
         
         synchronized (sessionFactories)
         {
            hibernateConfigs.remove(database);
            SessionFactory factory = sessionFactories.remove(database);
            PersistenceFactory.remove(database);
            
            if (factory != null)
            {
               try
               {
                  factory.close();
               }
               catch (HibernateException exc)
               {
                  if (log.isLoggable(Level.WARNING))
                  {
                     String msg = String.format(
                       "Unable to close Hibernate SessionFactory [%s]",
                       database);
                     log.log(Level.WARNING, msg, exc);
                  }
               }
               result = true;
            }

            // deregister the legacy name mappings for all tables in this permanent DB
            String schema = DatabaseManager.getSchema(database);

            synchronized (configLock)
            {
               schemaByDatabase.remove(database);

               // multiple databases can share the same schema
               if (!schemaByDatabase.containsValue(schema))
               {
                  TableMapper.removePermanentSchema(schema);
               }
            }
         }
         
         // Unregister dirty share manager for this database (and close the session held by it).
         DirtyShareFactory.unregisterManagerForDatabase(database);
      }
      
      return result;
   }
   
   /**
    * Indicate whether this P2J server instance is authoritative for the given
    * database.  Remote P2J servers must connect to the authoritative server
    * instance for a given database, in order to access that database and its
    * supporting resources, such as a lock manager, dirty read manager, etc.
    * 
    * @param   database
    *          Database information object.
    *          
    * @return  <code>true</code> if this server instance is authoritative for
    *          the specified database.
    */
   static boolean isAuthoritative(Database database)
   {
      synchronized (configLock)
      {
         return managed.contains(database);
      }
   }
   
   /**
    * Look up the Hibernate configuration helper object for a table.
    *
    * @param   database
    *          Database in which table resides.
    * @param   dmoClass
    *          DMO implementation class to which the table is mapped via ORM.
    *
    * @return  Hibernate table helper, or <code>null</code> if it cannot be
    *          found.
    */
   static Table getBackingTable(Database database, Class<?> dmoClass)
   {
      PersistentClass pc = getClassMapping(database, dmoClass);
      
      return (pc != null ? pc.getTable() : null);
   }
   
   /**
    * Get a list of secondary (associated) temporary tables for the given
    * primary, temporary table.  Secondary tables are those used to normalize
    * array (extent) fields from the original, Progress schema.
    * 
    * @param   database
    *          Database containing primary table.
    * @param   primaryTable
    *          Name of the primary temporary table.
    * 
    * @return  List of secondary tables, or <code>null</code> if none exist.
    */
   static List<Table> getSecondaryTables(Database database, String primaryTable)
   {
      Map<String, List<Table>> name2Tables = secondaryTables.get(database);
      List<Table> list = name2Tables.get(primaryTable);
      
      if (DatabaseManager.TEMP_TABLE_DB.equals(database) && list == null)
      {
         list = context.get().dynamicSecondaryTables.get(primaryTable);
      }
      
      return (list != null ? Collections.unmodifiableList(list) : null);
   }
   
   /**
    * Retrieve the Hibernate configuration object associated with the given
    * database and DMO implementation class.
    *
    * @param   database
    *          Database information object.
    * @param   dmoClass
    *          DMO implementation class which the object represents in the
    *          ORM configuration.
    *
    * @return  Mapped DMO class configuration object, or <code>null</code> if
    *          no match was found.
    */
   static PersistentClass getClassMapping(Database database, Class<?> dmoClass)
   {
      Configuration config = getHibernateConfiguration(database);
      if (config == null)
      {
         return null;
      }
      
      // We have to loop our way up the superclass chain in case dmoClass is
      // actually a Hibernate CGLIB proxy class injected as a proxy subclass
      // of a "primary" DMO class.  This could have happened if the caller
      // tried to divine dmoClass by invoking getClass() on a DMO instance
      // created by Hibernate.  Only the primary DMO classes are mapped to
      // PersistentClass instances.
      PersistentClass pc = null;
      Class<?> rootClass = Object.class;
      Class<?> nextClass = dmoClass;
      while (pc == null && !rootClass.equals(nextClass))
      {
         pc = config.getClassMapping(nextClass.getName());
         nextClass = nextClass.getSuperclass();
      }
      
      return pc;
   }
   
   /**
    * For a given DMO property which may represent an indexed text column, indicate whether that
    * column is set to ignore case.
    * 
    * @param   dmoClass
    *          DMO implementation class.
    * @param   property
    *          DMO property name.
    * 
    * @return  <code>true</code> to ignore case, <code>false</code> to preserve case sensitivity,
    *          <code>null</code> if the property does not represent an indexed text column.
    */
   static Boolean getIgnoreCase(Class<?> dmoClass, String property)
   {
      Map<String, Boolean> inner = compCharColInfo.get(dmoClass);
      if (inner == null)
      {
         inner = context.get().compCharColInfo.get(dmoClass);
         if (inner == null)
         {
            return null;
         }
      }
      
      return inner.get(property);
   }
   
   /**
    * Given a normal database, this method will register its associated 
    * dirty database; the created dirty database will have the same name as
    * the original database.
    * <p>
    * The dirty database is hard-coded to be an in-memory H2 database. See
    * {@link #createEmbeddedProperties(Database)} method for the Hibernate settings
    * used.
    *
    * @param   dbConfig
    *          Configuration properties for the given database.
    * @param   database
    *          Database to be configured.
    * @param   bootstrap
    *          <code>true</code> if this method is invoked during bootstrap
    *          initialization at server start-up;  <code>false</code> if
    *          invoked otherwise. If <code>false</code>, it is assumed that
    *          <code>database</code> represents a transient database.
    *
    * @return  <code>true</code> if the dirty database was newly registered;
    *          <code>false</code> if the database was already registered, and
    *          only the reference count was incremented (transient only), or
    *          if this registration attempt was redundant in the local
    *          context.
    *
    * @throws  PersistenceException
    *          If there is an error configuring the Hibernate framework.
    * @throws  IllegalStateException
    *          if <code>database</code> is permanent and already has been
    *          registered.
    * @throws  IllegalArgumentException
    *          if <code>database</code> is not recognized, based upon
    *          configuration information in the directory.
    */
   private static boolean registerDirtyDatabase(DatabaseConfig dbConfig,
                                                Database       database,
                                                Database.Type  type,
                                                boolean        bootstrap)
   throws PersistenceException
   {
      String schema = dbConfig.getSchema();
      Properties props = createEmbeddedProperties(database);
      
      DatabaseConfig secConfig = new DatabaseConfig(props, schema);
      Database secDB = new Database(database.getName(), type, database.isLocal());
      
      boolean newReg = registerDatabase(secConfig, secDB, bootstrap);
      
      if (newReg)
      {
         H2Helper.prepareDatabase(secDB);
      }
      
      return newReg;
   }
   
   /**
    * Register an individual database with the persistence framework. This
    * configures the database to allow the framework to access it. This
    * method ensures that a permanent database is never registered more
    * than once. Registration involves the following:
    * <ul>
    *   <li>create a Hibernate <code>Configuration</code>;
    *   <li>retrieve all Hibernate properties for that database from the
    *       directory and add them to the configuration;
    *   <li>add to the configuration each DMO class which corresponds to
    *       that database's schema name, as read from the DMO index;
    *   <li>build a Hibernate <code>SessionFactory</code> from the
    *       configuration and store it in a map, keyed by the database
    *       name (not the schema name, which may be different)
    *   <li>if the database is transient (i.e., not configured immediately
    *       during bootstrap), update its configuration reference count;
    *       otherwise, add the database name to the set of permanent
    *       databases which have been configured.
    * </ul>
    * <p>
    * This method may be invoked during bootstrap initialization, to register
    * a permanent database, or after initialization, to dynamically register
    * a transient database.
    *
    * @param   dbConfig
    *          Configuration properties for the given database.
    * @param   database
    *          Database to be configured.
    * @param   bootstrap
    *          <code>true</code> if this method is invoked during bootstrap
    *          initialization at server start-up;  <code>false</code> if
    *          invoked otherwise. If <code>false</code>, it is assumed that
    *          <code>database</code> represents a transient database.
    *
    * @return  <code>true</code> if the (permanent or transient) database was
    *          newly registered;
    *          <code>false</code> if the database was already registered, and
    *          only the reference count was incremented (transient only), or
    *          if this registration attempt was redundant in the local
    *          context.
    *
    * @throws  PersistenceException
    *          If there is an error configuring the Hibernate framework.
    * @throws  IllegalStateException
    *          if <code>database</code> is permanent and already has been
    *          registered.
    */
   private static boolean registerDatabase(DatabaseConfig dbConfig,
                                           Database database,
                                           boolean bootstrap)
   throws PersistenceException
   {
      boolean ret = false;
      synchronized (configLock)
      {
         // Can't configure a permanent database more than once.
         if (permanentConnections.contains(database))
         {
            if (log.isLoggable(Level.WARNING))
            {
               String msg = String.format(
                  "{%s} Attempt to register permanent database more than once has failed [%s]",
                  Utils.describeContext(),
                  database.toString());
               log.log(Level.WARNING, msg);
            }
            
            return false;
         }
         
         // Determine global reference count for this database.
         Integer refCountObj = transientDatabases.get(database);
         int refCount = (refCountObj != null) ? refCountObj : 0;
         
         if (schemaByDatabase.get(database) != null)
         {
            if (bootstrap)
            {
               log.log(Level.SEVERE,
                       "Auto-connect database " + database + " was already registered.");
            }
            
            refCount = -1;
         }
         
         // Has transient database already been configured?  If so, we don't
         // need to re-configure it, but we do need to update the global reference count.
         if (bootstrap || refCount == 0)
         {
            ret = true;
            String schema = dbConfig.getSchema().intern();
            
            // Remember the database to schema name mapping.
            schemaByDatabase.put(database, schema);
            
            // Get list of classes for the given schema.
            Iterator<Class<?>> iter = DMOIndex.classes(schema);
            if (iter == null && log.isLoggable(Level.WARNING))
            {
               String msg = "No schema for '" + database +
                            "' found in the DMO index (schema name:  " + schema + ")";
               log.log(Level.WARNING, msg);
            }
            
            try
            {
               // Remember dialect for this database.
               Properties props = dbConfig.getHibernateProperties();
               Dialect dialect = Dialect.getDialect(props);
               dialects.put(database, dialect);
               
               // Note that parsed Hibernate mappings are stored by schema, for each database
               // a new Configuration is created basing on these mappings. This is safe and
               // saves some time rather than parsing mappings each time.
               Configuration config = new Configuration();
               config.addProperties(props);
               
               // Use mutable session factory for temporary database
               if (TEMP_TABLE_SCHEMA.equals(schema))
               {
                  config.setProperty("session_factory.class",
                                     "org.hibernate.internal.MutableSessionFactoryImpl");
               }
               
               Set<XmlDocument> mappings = database.isDirty() ?
                                           getDirtyHibernateMappings(schema) :
                                           getHibernateMappings(schema);
               
               P2JDialect p2jDial = (P2JDialect) dialect;
               p2jDial.beforeDriverLoad();
               if (mappings == null)
               {
                  // Create configuration documents for the database.
                  ORMHandler handler = new ORMHandler(p2jDial, database);
                  
                  // Load all classes associated with database's schema.
                  if (iter != null)
                  {
                     while (iter.hasNext())
                     {
                        Class<?> dmoClass = iter.next();
                        handler.mapClass(schema, dmoClass);
                        
                        // Remember the DMO class to schema name mapping.
                        schemaByDMOClass.put(dmoClass, schema);
                        
                        // Remember the DMO interface to schema name mapping.
                        Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
                        schemaByDMOIface.put(dmoIface, schema);
                     }
                  }
                  
                  mappings = handler.getHibernateMappings();
                  
                  synchronized (sessionFactories)
                  {
                     if (database.isDirty())
                     {
                        dirtyHibernateMappings.put(schema, mappings);
                     }
                     else
                     {
                        hibernateMappings.put(schema, mappings);
                     }
                  }
               }
               
               ORMHandler.applyHibernateMappings(mappings, config);
               
               // Transient databases must not share the second-level cache,
               // or we'll get unpredictable behavior.  Also, we don't want
               // transient SessionFactory instances to tear down the cache
               // when they're closed.
               if (!bootstrap)
               {
                  config.setProperty(Environment.USE_SECOND_LEVEL_CACHE, "false");
                  config.setProperty(Environment.USE_QUERY_CACHE, "false");
               }
               
               // Create and store the session factory.  Session factories are
               // stored by physical database, so we are sure each physical
               // database has its own session factory.
               ServiceRegistry serviceRegistry = new ServiceRegistryBuilder()
                                                   .applySettings(config.getProperties())
                                                   .buildServiceRegistry();
               SessionFactory factory = config.buildSessionFactory(serviceRegistry);
               synchronized (sessionFactories)
               {
                  sessionFactories.put(database, factory);
                  
                  mapCompositeProperties(config);
                  
                  if (!database.isPrimary() || TEMP_TABLE_SCHEMA.equals(schema))
                  {
                     // For dirty databases and the temp database, map all secondary tables used
                     // for extent field normalization.
                     mapSecondaryTables(database, config);
                  }
                  
                  hibernateConfigs.put(database, config);
                  configBySchema.put(schema, config);
                  if (p2jDial.needsComputedColumns())
                  {
                     // Map all indexed, computed columns which hold text data. We do this using
                     // the primary database, since the dirty database indexes will not have been
                     // created during this stage of initialization.
                     Database db = (database.isDirty()
                                    ? new Database(database, Database.Type.PRIMARY)
                                    : database);
                     mapComputedCharacterColumns(schema, db, config);
                  }
               }
            }
            catch (HibernateException exc)
            {
               DBUtils.handleException(database, exc);
               
               throw new PersistenceException(
                  "Error configuring persistence framework for database: " + database,
                  exc);
            }
         }
         
         // Increment reference count for a transient database, or add
         // database to configured, permanent databases if database is not transient.
         if (database.isPrimary())
         {
            if (bootstrap)
            {
               permanentConnections.add(database);
            }
            else
            {
               int newCount = updateTransientRefCount(database, true);
               
               return (newCount > refCount);
            }
         }
         
         return ret;
      }
   }
   
   /**
    * For each character column which must be represented within a database index as a computed
    * column (certain dialects only), map the name of the associated DMO property to a
    * <code>Boolean</code> indicating whether that computed column is designed to ignore case in
    * its ordering.  This allows runtime code (such as {@link SortCriterion} and the {@link
    * HQLPreprocessor} to quickly find this information later.
    * <p>
    * A multi-level mapping is created:  the first level maps each DMO implementation class to
    * another map of property names to ignore-case values.
    *
    * @param   schema
    *          Database schema.
    * @param   database
    *          Database information object.
    * @param   config
    *          Hibernate ORM configuration.
    *
    * @throws  PersistenceException
    *          if there is an error querying index information from the JDBC driver.
    */
   private static void mapComputedCharacterColumns(String schema,
                                                   Database database,
                                                   Configuration config)
   throws PersistenceException
   {
      Iterator pcIter = config.getClassMappings();
      while (pcIter.hasNext())
      {
         PersistentClass pc = (PersistentClass) pcIter.next();
         mapComputedCharacterColumnsImpl(schema, database, pc, compCharColInfo);
      }
   }
   
   /**
    * For each character column which must be represented within a database index as a computed
    * column (certain dialects only), map the name of the associated DMO property to a
    * <code>Boolean</code> indicating whether that computed column is designed to ignore case in
    * its ordering.  This allows runtime code (such as {@link SortCriterion} and the {@link
    * HQLPreprocessor} to quickly find this information later.
    * <p>
    * A multi-level mapping is created: the first level maps each DMO implementation class to
    * another map of property names to ignore-case values.
    *
    * @param   schema
    *          Database schema.
    * @param   database
    *          Database information object.
    * @param   persistentClass
    *          Hibernate persistent class which contains the ORM mapping for a particular DMO.
    * @param   cccInfo
    *          Computed character column map into which the derived information is stored.
    *
    * @throws  PersistenceException
    *          if there is an error querying index information from the JDBC driver.
    */
   private static void
   mapComputedCharacterColumnsImpl(String schema,
                                   Database database,
                                   PersistentClass persistentClass,
                                   Map<Class<?>, Map<String, Boolean>> cccInfo)
   throws PersistenceException
   {
      Class<?> dmoClass = persistentClass.getMappedClass();
      Class<?> dmoIface = DMOIndex.getDMOInterface(schema, dmoClass);
      Map<String, P2JIndexComponent> map =
         DMOIndex.getIndexedCharacterColumns(schema, database, dmoIface);
      
      if (map.isEmpty())
      {
         return;
      }
      
      Map<String, Boolean> inner = null;
      Iterator<Property> propIter = persistentClass.getUnjoinedPropertyIterator();
      while (propIter.hasNext())
      {
         Property prop = propIter.next();
         
         Iterator<Column> colIter = prop.getColumnIterator();
         if (!colIter.hasNext())
         {
            // This can happen for collection associations with element entries.
            continue;
         }
         
         // Assume no more than one column per unjoined property.
         Column col = colIter.next();
         String colName = col.getName();
         
         P2JIndexComponent idxComp = map.get(colName);
         if (idxComp != null)
         {
            String propName = prop.getName();
            boolean ignoreCase = idxComp.isIgnoreCase();
            if (inner == null)
            {
               inner = cccInfo.get(dmoClass);
               if (inner == null)
               {
                  inner = new HashMap<>();
                  cccInfo.put(dmoClass, inner);
               }
            }
            inner.put(propName, ignoreCase);
         }
      }
   }
   
   /**
    * Examine all collection mappings in a Hibernate configuration and
    * organize those which represent collections of
    * <code>composite-element</code>s for fast access later by the {@link
    * HQLPreprocessor} when rewriting where clauses.
    *
    * @param   config
    *          Hibernate configuration for all DMO classes loaded for a
    *          particular physical database.
    */
   private static void mapCompositeProperties(Configuration config)
   {
      Iterator<Collection> iter = config.getCollectionMappings();
      while (iter.hasNext())
      {
         Collection next = iter.next();
         Value element = next.getElement();
         if (element instanceof Component)
         {
            mapCompositeProperties(next, (Component) element, false);
         }
      }
   }
   
   /**
    * Organize the information in a <code>composite-element</code> collection
    * mapping for fast access later by the {@link HQLPreprocessor}.  The
    * preprocessor needs to look up the name of a composite element
    * collection, given a DMO implementation class and the name of a property
    * within the composite-element.
    * <p>
    * The data structure we create to enable this is a map within a map.  The
    * inner map contains composite element collection names as values, and is
    * indexed by property name keys.  These inner maps are the values stored
    * within the outer map, whose keys are the DMO classes which contain these
    * collections.  A single outer map is shared across all schemae, since
    * DMO implementation classes are unique within this space.
    *
    * @param   collection
    *          The Hibernate mapping object which defines the collection of
    *          composite elements.
    * @param   component
    *          The Hibernate mapping object which defines the composite
    *          elements stored in the collection.
    * @param   dynamic
    *          <code>true</code> if DMO class is dynamic, <code>false</code> if
    *          it is static.
    */
   private static void mapCompositeProperties(Collection collection,
                                              Component component,
                                              boolean dynamic)
   {
      // Get the name of the composite collection variable.
      String compName = collection.getRole();
      compName = compName.substring(compName.lastIndexOf(".") + 1);
      
      // Get DMO implementation class.  This is the key to the outer map.
      Class<?> dmoClass = component.getOwner().getMappedClass();
      
      Context ctx = (dynamic ? context.get() : null);
      
      // Get or create the inner map.
      Map<String, String> inner = dynamic ? ctx.dynamicComposites.get(dmoClass) :
                                            composites.get(dmoClass);
      if (inner == null)
      {
         inner = new HashMap<>();
         if (dynamic)
         {
            ctx.dynamicComposites.put(dmoClass, inner);
         }
         else
         {
            composites.put(dmoClass, inner);
         }
      }
      
      // Iterate the properties of the component.  In the inner map, the
      // property name is the key and the component (i.e., composite) name is the value.
      Iterator<Property> iter = component.getPropertyIterator();
      while (iter.hasNext())
      {
         Property next = iter.next();
         String propName = next.getName();
         inner.put(propName, compName);
      }
   }
   
   /**
    * Examine all collection mappings in a Hibernate configuration and group
    * them into lists of secondary temp tables, mapped by the names of primary
    * temp tables.  For the purposes of this mapping, a "primary" table is
    * the main table which backs a temp table DMO.  A "secondary" table is
    * one which supports one or more extent fields of the main table, by
    * creating a foreign key relation back to the primary table.
    * <p>
    * It is expected that this method is only called for the special database
    * which backs temp tables.  The mapping is created at P2J server
    * initialization, and is used when creating and dropping temp tables at
    * runtime.
    *
    * @param   config
    *          Hibernate configuration for all DMO classes loaded for a
    *          particular physical database.
    */
   private static void mapSecondaryTables(Database database, Configuration config)
   {
      Map<String, List<Table>> name2Tables = secondaryTables.get(database);
      if (name2Tables == null)
      {
         name2Tables = new HashMap<String, List<Table>>();
         secondaryTables.put(database, name2Tables);
      }
      
      Iterator<Collection> iter = config.getCollectionMappings();
      while (iter.hasNext())
      {
         Collection next = iter.next();
         
         // Get name of primary table.
         String name = next.getOwner().getTable().getName();
         
         // Get list for this primary table, creating it if necessary.
         List<Table> secondaries = name2Tables.get(name);
         if (secondaries == null)
         {
            secondaries = new ArrayList<Table>();
            name2Tables.put(name, secondaries);
         }
         
         // Add secondary table for this collection to the list.
         secondaries.add(next.getCollectionTable());
      }
      
      // Reclaim a little memory.
      Iterator<List<Table>> iter2 = name2Tables.values().iterator();
      while (iter.hasNext())
      {
         ((ArrayList<Table>) iter2.next()).trimToSize();
      }
   }
   
   /**
    * Update the reference count of registrations of the specified database
    * across all contexts. The reference count for each database is global
    * across the system. Its value determines when a database configuration
    * may be safely dropped by the server.
    * <p>
    * The reference count is either incremented by 1, decremented by 1, or
    * left unchanged, based upon the following rules:
    * <ul>
    *   <li>if <code>increment</code> is <code>true</code>:
    *   <ul>
    *     <li>if <code>database</code> is already registered in the local
    *         context, no change is made;
    *     <li>otherwise, the count is incremented by 1.
    *   </ul>
    *   <li>if <code>increment</code> is <code>false</code>:
    *   <ul>
    *     <li>if <code>database</code> is <i>not</i> registered in the local
    *         context, no change is made;
    *     <li>otherwise, the count is decremented by 1.
    *   </ul>
    * </ul>
    * <p>
    * This logic ensures that attempts to register and deregister the same
    * database, if not properly bracketed, will not corrupt the global
    * reference count.
    *
    * @param   database
    *          A transient database.
    * @param   increment
    *          <code>true</code> to increment reference count by 1;
    *          <code>false</code> to decrement reference count by 1;
    *          See above for the implications of this.
    *
    * @return  A non-negative number indicating the reference count for
    *          <code>database</code> <i>after</i> the update; or -1 if
    *          <code>database</code> was not recognized.
    */
   private static int updateTransientRefCount(Database database, boolean increment)
   {
      // Amount to increment or decrement the reference count.
      int delta = (increment ? 1 : -1);
      
      // The context local set of registered databases must be managed, to
      // make sure that multiple register/deregister requests made within
      // the same context do not skew the global reference count.
      Context ctx = context.get();
      if (ctx.localTransients == null)
      {
         ctx.localTransients = new HashSet<Database>();
      }
      
      boolean locallyRegistered = ctx.localTransients.contains(database);
      if (increment)
      {
         // We should not increment the reference count if this database is
         // already registered in the local context.
         if (locallyRegistered)
         {
            delta = 0;
         }
         else
         {
            ctx.localTransients.add(database);
         }
      }
      else
      {
         // We should not decrement the reference count if this database is
         // not registered in the local context.
         if (!locallyRegistered)
         {
            delta = 0;
         }
         else
         {
            ctx.localTransients.remove(database);
            if (ctx.localTransients.isEmpty())
            {
               ctx.localTransients = null;
            }
         }
      }
      
      return updateTransientRefCountWorker(database, delta);
   }
   
   /**
    * Update the reference count of registrations of the specified database. This method must
    * be called from a point where the <code>delta</code> have been correctly computed,
    * respectively <code>updateTransientRefCount()</code> where the current context is inspected
    * to see if incrementation is needed, or from context clean when all databases from context
    * should be de-ref counted (and <code>delta</code> is -1).
    *
    * @param   database
    *          A transient database.
    * @param   delta
    *          Specifies the direction of change (1 = increment, -1 decrement, 0 = no change);
    *          See updateTransientRefCount() for details.
    *
    * @return  A non-negative number indicating the reference count for
    *          <code>database</code> <i>after</i> the update; or -1 if
    *          <code>database</code> was not recognized.
    */
   private static int updateTransientRefCountWorker(Database database, int delta)
   {
      int result = -1;
      synchronized (configLock)
      {
         // Update the global reference count for this transient database.
         Integer refCount = transientDatabases.get(database);
         result = (refCount != null) ? Math.max(refCount + delta, 0) : delta;
         transientDatabases.put(database, result);
      }
      
      return result;
   }
   
   /**
    * Load configuration properties from the P2J directory service and
    * return them as a <code>Properties</code> object. The contents of this
    * object must match the properties expected by a Hibernate
    * <code>Configuration</code> object.
    *
    * @param   database
    *          Database for which properties are to be loaded.  This
    *          differentiates the directory key by database name.
    *
    * @return  Hibernate configuration properties.
    *
    * @throws  PersistenceException
    *          if no configuration properties for the specified database
    *          were found;  if the directory service is unavailable.
    */
   private static Properties loadConfigProperties(Database database)
   throws PersistenceException
   {
      DirectoryService ds = DirectoryService.getInstance();
      if (!ds.bind())
      {
         throw new PersistenceException("Directory bind failed");
      }
      
      try
      {
         String nodeId = Utils.findDirectoryNodePath(ds, "database", "container", false);
         String baseID = (nodeId == null
                          ? null
                          : nodeId + "/" + database.getId() + "/hibernate");
         Properties props = new Properties();
         
         // TODO:  handle fetching props from remote server if database is
         // remote and directory lookup fails.
         readHibernateProperties(ds, baseID, props);
         
         return props;
      }
      finally
      {
         ds.unbind();
      }
   }
   
   /**
    * This method returns the Hibernate properties which should be used when connecting to a
    * secondary (e.g., dirty or metadata) database associated with a normal database.
    * <p>
    * These used properties are:
    * <table style="text-align: left; width: 100%;" border="1" cellpadding="2" 
    *        cellspacing="2">
    *    <tr>
    *       <td style="vertical-align: top; text-align: center; 
    *           font-weight: bold;">
    *           Property
    *       </td>
    *       <td style="vertical-align: top; text-align: center; 
    *           font-weight: bold;">
    *           Value
    *        </td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.username</td>
    *       <td>{randomly generated}</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.password</td>
    *       <td>{randomly generated}</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.url</td>
    *       <td>jdbc:h2:mem:&lt;database name&gt;;DB_CLOSE_DELAY=-1;MULTI_THREADED=1</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.cache.use_second_level_cache</td>
    *       <td>false</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.release_mode</td>
    *       <td>on_close</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.provider_class</td>
    *       <td>
    *          {@link com.goldencode.p2j.persist.UnpooledConnectionProvider}
    *       </td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.connection.driver_class</td>
    *       <td>org.h2.Driver</td>
    *    </tr>
    *    <tr>
    *       <td>hibernate.dialect</td>
    *       <td>{@link com.goldencode.p2j.persist.dialect.P2JH2Dialect}</td>
    *    </tr>
    * </table>
    * 
    * @param   database
    *          The database information object for the embedded database.
    * 
    * @return  The created Hibernate properties.
    */
   private static Properties createEmbeddedProperties(Database database)
   {
      Properties props = new Properties();
      
      StringBuilder url = new StringBuilder();
      url.append("jdbc:h2:mem:");
      if (database.isMeta())
      {
         url.append(MetadataManager.META_SCHEMA);
         url.append("_");
      }
      url.append(database.getName());
      url.append(";DB_CLOSE_DELAY=-1;MULTI_THREADED=1");
      
      props.put(Environment.URL, url.toString());
      
      props.put(Environment.USER, RandomWordGenerator.create(128, 16, 16));
      props.put(Environment.PASS, RandomWordGenerator.create(128, 16, 16));
      
      props.put(Environment.USE_SECOND_LEVEL_CACHE, "false");
      props.put(Environment.RELEASE_CONNECTIONS, "on_close");
      
      props.put(Environment.CONNECTION_PROVIDER,
         "com.goldencode.p2j.persist.UnpooledConnectionProvider");
      
      props.put(Environment.DRIVER, "org.h2.Driver");
      
      props.put(Environment.DIALECT, "com.goldencode.p2j.persist.dialect.P2JH2Dialect");
      
      return props;
   }
   
   /**
    * Read hibernate configuration properties from the directory and populate
    * a properties file which will be passed to Hibernate's configuration API.
    * Properties are all strings, organized hierarchically in the directory.
    * This method recursively walks the tree and adds a new property entry
    * when it reaches any <code>/meta/class/string</code> directory entry
    * terminus.
    * <p>
    * This method does not know or care about the names of individual
    * properties;  it simply performs a depth-first enumeration of a
    * directory subtree and composes the property names from the directory
    * entry names it encounters.  Error checking of property names and values
    * is left to the Hibernate configuration subsystem.
    *
    * @param   ds
    *          Directory service instance.
    * @param   id
    *          Directory entry key for the node currently being visited.
    * @param   props
    *          Object to which properties are added.
    */
   private static void readHibernateProperties(DirectoryService ds, String id, Properties props)
   {
      String nodeClass = ds.getNodeClass(id);
      
      if (nodeClass.endsWith("/container"))
      {
         // We are at a node containing nested nodes;  enumerate and call
         // ourselves recursively to walk down to the next level.
         String[] names = ds.enumerateNodes(id);
         
         if (names == null || names.length == 0)
         {
            throw new RuntimeException("Invalid directory entry:  " + id);
         }
         
         for (String name : names)
         {
            readHibernateProperties(ds, id + "/" + name, props);
         }
      }
      else
      {
         String name = id.substring(id.indexOf("/hibernate/") + 1).replaceAll("/", ".");
         
         if (nodeClass.endsWith("/string"))
         {
            // We are at a string terminus;  record the property.
            String value = ds.getNodeString(id, "value");
            props.setProperty(name, value);
         }
         else if (nodeClass.endsWith("/integer"))
         {
            // We are at an integer terminus;  record the property.
            Integer value = ds.getNodeInteger(id, "value");
            props.setProperty(name, value.toString());
         }
         else if (nodeClass.endsWith("/boolean"))
         {
            // We are at a boolean terminus;  record the property.
            Boolean value = ds.getNodeBoolean(id, "value");
            props.setProperty(name, value.toString());
         }
         else
         {
            throw new RuntimeException("Directory entry is incorrect type:  " + id);
         }
      }
   }
   
   /**
    * Interface for a remote object whose purpose is to fetch a local {@link DatabaseConfig}
    * object and deliver it to the remote caller. It is the interface used as a proxy for a
    * remote object.
    */
   public static interface DatabaseConfigFetcher
   {
      /**
       * Fetch and return a local database configuration, given the physical
       * name of the database.
       * 
       * @param   database
       *          Database information object.
       *          
       * @return  Database configuration for the given database.
       */
      public DatabaseConfig fetch(Database database);
   }
   
   /**
    * Concrete implementation of the <code>DatabaseConfigFetcher</code> interface. Instances of
    * this class run as remote objects, serving requests for database configuration information
    * from other servers.
    */
   public static class DatabaseConfigFetcherImpl
   implements DatabaseConfigFetcher
   {
      /**
       * Default constructor.
       */
      private DatabaseConfigFetcherImpl()
      {
      }
      
      /**
       * Fetch and return a local database configuration, given the database info object.
       * 
       * @param   database
       *          Database information object.
       *          
       * @return  Database configuration for the given database.
       */
      public DatabaseConfig fetch(Database database)
      {
         synchronized (configLock)
         {
            // since this is running on the remote server (not the requesting server), we have to
            // treat the database object as though it represents a local database;  we do this by
            // creating a new Database object without the host address information
            Database localDB = new Database(database.getName(), database.getType());
            
            return dbConfigs.get(localDB);
         }
      }
   }
   
   /**
    * A class containing context-specific data.
    */
   private static class Context
   {
      /** Map of dynamic table names to lists of Hibernate <code>Table</code> objects */
      Map<String, List<Table>> dynamicSecondaryTables = new HashMap<>();
      
      /** Map of dynamic DMO classes to maps of property names to composite names */
      Map<Class<?>, Map<String, String>> dynamicComposites = new HashMap<>();
      
      /** Set of DMO interface classes for dynamic tables */
      Set<Class<?>> dynamicDMOIfaces = new HashSet<>();
      
      /** Set of DMO implementation classes for dynamic tables */
      Set<Class<?>> dynamicDMOClasses = new HashSet<>();
      
      /** Map of computed column base property names to ignore-case booleans */
      Map<Class<?>, Map<String, Boolean>> compCharColInfo = new IdentityHashMap<>();
      
      /** Transient databases registered in the current context */
      Set<Database> localTransients = null;
      
      /**
       * Deregister all transient databases registered for use in the current context, when that
       * context is closing. This method is invoked automatically when the context is being
       * cleaned up.
       */
      void cleanup()
      {
         if (localTransients != null)
         {
            for (Database database : localTransients)
            {
               // if a database is contained by the value set, it was used in the context that
               // is cleaned up so the counter must be decremented.
               updateTransientRefCountWorker(database, -1);
            }
         }
      }
   }
}
