/*
** Module   : TemporaryBuffer.java
** Abstract : Specialized record buffer which supports temp table semantics
**
** Copyright (c) 2004-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20060125   @24356 Created initial version. Extends RecordBuffer
**                           to provide temp table semantics.
** 002 ECF 20060414   @25545 Added clear() method. Removes all records
**                           from a virtual temp table.
** 003 ECF 20060504   @26008 Added runtime support for the creation and
**                           dropping of secondary temp tables. These are
**                           tables created to support the extent fields
**                           of a primary temp table.
** 004 GES 20060517   @26201 Match new Finalizable interface requirement.
** 005 ECF 20060608   @27002 Use updated signature for Persistence.list().
** 006 ECF 20060621   @27508 Fixed temp table cleanup. Ensure that
**                           dropping a table or removing records as part
**                           of multiplex scope closure is performed
**                           within a committed database transaction. Also
**                           added a per-context reference count facility
**                           for temp tables.
** 007 ECF 20060627   @27594 Fixed temp table creation and buffer scope
**                           opening. Temp table creation must occur
**                           within a transaction. The superclass'
**                           implementation of openScope() is invoked from
**                           the overriding method.
** 008 ECF 20060713   @28047 Use new DBUtils class.
** 009 ECF 20060719   @28103 Use updated signature for Persistence.list().
** 010 ECF 20060914   @29596 Implemented master/slave buffers. These are
**                           linked to one another by a common multiplex
**                           ID.
** 011 ECF 20060919   @29699 Override the parent's implementations of
**                           processChange(), incrementDMOUseCount(),
**                           decrementDMOUseCount(), commit(), and
**                           rollback(). The first three disable the
**                           parent class' aggressive record eviction
**                           policy for temp tables, and the latter two
**                           disable subtransaction-level undo support for
**                           temp tables. NOTE: some level of undo support
**                           may need to be implemented eventually.
** 012 ECF 20061010   @30312 Added support for undoable temp tables. By
**                           default, temp tables are not undoable,
**                           however, this setting can be overridden by
**                           using a new variant of the static define()
**                           method.
** 013 ECF 20061019   @30545 Fixed NPE in copyAllRows(). Must not iterate
**                           if no source records are returned by the list
**                           query.
** 014 ECF 20061116   @31238 Fixed shared buffers and slave buffers.
**                           Tables were not being cleaned up properly in
**                           certain situations.
** 015 ECF 20061121   @31346 Fixed commit problem. Force ChangeBroker to
**                           report a pending flush after persisting or
**                           deleting records.
** 016 ECF 20061211   @31657 Make temp-tables undoable by default. The
**                           no-undo variant requires a separate option to
**                           be declared in Progress and will convert to
**                           special variant of the define() method, so
**                           undoability is the more appropriate default.
** 017 ECF 20070130   @32043 Various cleanup. Introduced methods for table
**                           counting by context. Fixed invocations of
**                           superclass methods to use new signatures.
** 018 ECF 20070202   @32072 Fixed regression in copyAllRows(). Recent
**                           changes to bulk copy implementation had
**                           broken this implementation.
** 019 ECF 20070326   @32612 Fixed associate() method and improved
**                           removeRecords() implementation. The former
**                           required a conditional flush before copying
**                           input records. The latter now uses a bulk
**                           delete instead of a looping delete.
** 020 ECF 20070329   @32649 Fixed shared temp-table support. Added static
**                           method useShared() to access a previously
**                           established, shared temp-table definition.
** 021 ECF 20070412   @32922 Override parent's setPinnedLockType() method.
**                           Does nothing, since locking is a no-op for
**                           temp tables.
** 022 ECF 20070415   @33012 Fixed define() variant which creates a slave
**                           buffer. In some cases, master buffer was not
**                           being determined correctly, resulting in an
**                           NPE.
** 023 ECF 20070416   @33030 Integrated user interrupt handling.
** 024 ECF 20070508   @33463 Fixed associate() to properly handle a
**                           validation exception. New requirement based
**                           on a change to RecordBuffer.flush().
** 025 ECF 20070601   @33919 Fixed memory leak. Buffers were not being
**                           removed from context's domains map when the
**                           buffer was no longer needed, which pinned the
**                           instances in context local memory. Changed
**                           implementation of domains map to be a weak
**                           hash map.
** 026 ECF 20070602   @33927 Implemented deferred table drop optimization.
**                           Temp tables created for the current database
**                           session are not dropped until the session is
**                           being closed. This prevents a condition where
**                           the database repeatedly creates and drops the
**                           same tables as a secondary program which
**                           creates TemporaryBuffers and allows them to
**                           go out of scope is invoked within a loop of
**                           another program.
** 027 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 028 ECF 20070706   @34405 Added deregisteredSessionListener() method.
**                           Required by change to SessionListener
**                           interface. Also replaced StringBuffer with
**                           StringBuilder.
** 029 ECF 20070715   @34856 Fixed removeRecords(). Release current buffer
**                           record after bulk delete.
** 030 ECF 20070827   @34994 Added bulk delete and optimized multiplexing.
**                           Override parent's implementation of delete()
**                           variant which accepts a where clause and
**                           query substitution parms to perform a bulk
**                           delete. Omit multiplex ID from where clauses
**                           when possible.
** 031 ECF 20070914   @35071 Fixed regression introduced by H030 (@34994).
**                           When removing records during the closure of a
**                           multiplex scope, we have to ensure only those
**                           records associated with the multiplex ID of
**                           the closing scope are removed from the temp
**                           table.
** 032 ECF 20070914   @35294 Fixed ConnectionManager notifications. We
**                           override the parent's implementation of two
**                           protected methods which notify the connection
**                           manager of buffer open and close events to do
**                           nothing, because we don't want to trace these
**                           for temp tables. Replaced database name
**                           string with database information object.
** 033 ECF 20071002   @35332 Exposed certain methods as public. Expanded
**                           use of generics.
** 034 ECF 20071018   @35592 Integrated PersistenceFactory.
** 035 ECF 20071108   @35909 Made nextMPID access threadsafe.
** 036 ECF 20071130   @36126 Integrated generics.
** 037 ECF 20071220   @36548 Modified Context.sessionClosing(). We now do
**                           more state cleanup when the session is
**                           closing due to an error.
** 038 SVL 20071231   @36628 Next primary key generation is carried out
**                           to the separate function which overrides the
**                           one in RecordBuffer.
** 039 ECF 20080102   @36638 Replaced use of deprecated nextPrimaryKey()
**                           method in Persistence. Primary key generation
**                           is now managed within Context inner class.
** 040 ECF 20080103   @36696 Fixed regression caused by #039 (@36638).
**                           Used wrong buffer to call nextPrimaryKey()
**                           from copyAllRows().
** 041 ECF 20080108   @36732 Fixed bulk delete. Changed call to delete()
**                           method in Persistence.
** 042 ECF 20080310   @37487 Fixed potential memory leak. copyAllRows()
**                           could leave DMOs in the Hibernate session if
**                           they were never loaded into a buffer. Also
**                           integrated DDL to drop indexes and to sweep
**                           index names in create and drop statements.
** 043 CA  20080430   @38161 Changes due to the TempTableHelper.sql*Drop
**                           /Create method signature change: instead of
**                           a Persistence object, the first parameter is
**                           a Database object.
** 044 SVL 20080516   @38292 Some superclass methods were overridden to
**                           no-ops to reduce memory consumption.
** 045 CA  20080530   @38458 Temporary tables passed as parameters to
**                           external procedures must match using each
**                           field's position and type, not its name;
**                           PROGRESS will not care if field names are
**                           different. Ref: #38457
** 046 ECF 20080602   @38615 Override getIndexedProperties(boolean) to do
**                           nothing. This method supports dirty checking,
**                           which is not needed for temp tables.
** 047 ECF 20080609   @38654 Backed out #046 (@38615) which regressed
**                           unique constraint validation. Removed
**                           getIndexProperties(boolean) method and added
**                           postValidateDirtyShare(), which overrides the
**                           parent's method to do nothing.
** 048 CA  20080605   @38570 The multiplex ID must start from 1.
**                           Implemented reversible bulk delete action
**                           for temporary buffers.
** 049 ECF 20080708   @39075 Fixed copyAllRows(). When copying from one
**                           table type to another, properties which are
**                           not available via a DMO's API must be
**                           skipped (i.e., computed columns).
** 050 ECF 20080723   @39172 Fixed flushing behavior for temp-tables with
**                           no indexes. Updates to such tables must be
**                           flushed to the database immediately.
** 051 ECF 20080806   @39329 Added rowid support. New primary keys are
**                           64-bit integers.
** 052 CA  20080808   @39360 For no-undo tables, the bulk delete will 
**                           always force removal of the records. Also,
**                           a reversible is registered with so the delete
**                           will be re-executed if db tx is rolled back.
** 053 ECF 20080815   @39558 Support global DMO change events.
** 054 CA  20080815   @39468 Support API change in DBUtils.
** 055 SIY 20080904   @39720 Inferred generics in buffer creation methods.
** 056 ECF 20080925   @39950 Refactored Reversible concrete classes.
** 057 ECF 20080926   @39954 Fixed ReversibleBulkDelete.setDMO(). Should
**                           not throw UnsupportedOperationException.
** 058 ECF 20081009   @40083 Integrated HQLPreprocessor API change.
** 059 ECF 20081022   @40212 Modified to adjust for Reversible API change.
** 060 ECF 20081127   @40710 Implement reclamation of temp table primary
**                           keys. On delete, a record's primary key is
**                           collected for possible recycling. On full
**                           transaction commit/rollback, pending keys are
**                           made available for re-use, or discarded,
**                           respectively. Currently, keys surrendered by
**                           bulk deletes and by rolled back record
**                           creations are not reclaimed.
** 061 ECF 20081128   @40718 Fixed regression introduced by #060 (@40710).
**                           Data structures for recycling primary keys
**                           were not being managed by table properly.
** 062 ECF 20081201   @40734 Modified primary key reclamation strategy.
**                           We now piggyback off the implementation in
**                           the superclass, so that surrendered keys are
**                           collected during the buffer's commit() and
**                           rollback() operations. This fixes yet another
**                           regression caused by the previous two mods.
** 063 ECF 20081202   @40744 Fixed regression in key reclamation. Must
**                           clear key reclamation data structures when
**                           the session is closed to avoid unique
**                           constraint violations in following session.
** 064 ECF 20081211   @40871 Improved logging in temp table creation code.
** 065 ECF 20090107   @41025 Fixed record release during bulk delete. A
**                           bulk delete event is now broadcast, which
**                           allows listening buffers to release their
**                           current records if they were deleted.
** 066 ECF 20090112   @41164 Fixed undo-related locking behavior. Locks
**                           were being downgraded rather than released in
**                           certain transaction-boundary cases. The
**                           change necessary in this class was to
**                           override certain lock-related methods from
**                           the superclass to provide no-op behavior.
** 067 ECF 20090218   @41334 Added safety code for dealing with maps of
**                           Reversibles. BufferManager methods which
**                           return these maps can now return null.
** 068 ECF 20090227   @41446 Attempted to address memory leak. Remove all
**                           Reversibles from BufferManager when the
**                           associated virtual table is no longer in use.
** 069 GES 20090424   @41943 Import change.
** 070 ECF 20090508   @42128 Added inner class TempValidationHelper. Extends
**                           RecordBuffer's implementation to override certain
**                           methods.
** 071 ECF 20090610   @42634 Optimized copyAllRows(). Call BufferManager's
**                           evictDMOIfUnused() method instead of incrementing
**                           then decrementing DMO use counts.
** 072 ECF 20090701   @43223 Various fixes and optimizations. Changed
**                           copyAllRows() method to use scrollable results
**                           instead of a results list. Added table empty
**                           tracking. Removed immediate flush concept.
**                           Refactored initialization to be lazy, upon open
**                           scope event.
** 073 ECF 20090717   @43230 Fixed regressions in #072 (@43223). Opening a
**                           multiplex scope was unconditionally setting the
**                           emptyTempTables flag for that multiplex ID, but
**                           it must only do so if the reference count is 1.
**                           When closing the last scope using a physical temp
**                           table, we must remove all the rows before the
**                           table is dropped, otherwise we can get into
**                           trouble flushing deferred rollbacks.
** 074 ECF 20090724   @43423 Added getMasterBuffer() method. This is needed to
**                           support proper rollback handling of shared temp
**                           tables.
** 075 ECF 20090729   @43450 Fixed copyAllRows(). Call to ChangeBroker to
**                           force a pending flush must come before the micro
**                           transaction commit, otherwise the copied records
**                           are lost. Also implemented no-op validate() call
**                           for Context, which implements Commitable.
** 076 ECF 20090810   @43591 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 077 CA  20090916   @43916 Fixed bulk delete for no-undo temp tables. On
**                           commit, do not flush the bulk deletes for no-undo 
**                           temp tables, as this are relevant only for
**                           undoable tables.
** 078 CA  20091001   @44061 Fixed leak caused by bulk delete duplication: do 
**                           not duplicate it if the record was deleted 
**                           explicitly (via a DELETE operation) or implicitly
**                           (via a bulk delete).
** 079 CA  20091202   @44469 Reworked the rollback mechanism, to maintain 
**                           proper reversible order and thus to solve some
**                           unique-constraint violations during rollback of
**                           some bulk delete cases and mixed record operations.
**                           The ReversibleChain is no longer used, instead a 
**                           linked set with all operations which affect a 
**                           certain table is used.
** 080 ECF 20130119          Changed the public API to this class; a number of
**                           static methods have been removed and replaced
**                           by the Buffer interface API. This enables more
**                           natural, direct method calls on DMO proxy objects
**                           for standard services. This entailed a change to
**                           the DMO proxy implementation from the dynamic
**                           proxy support in java.lang.reflect to that in
**                           com.goldencode.proxy (we needed to inherit our
**                           proxies from a functional base class).
** 081 SVL 20130223          Added createDynamicTable and associate(TableParameter ...).
** 082 VMN 20130407          Method copyAllRows() changes related with runtime support for
**                           BUFFER-COPY (KW_BUF_COPY)
** 083 CA  20130610          Added APIs to populate or access temp table data using a result set.
** 084 SVL 20130822          Added openFreeScope, dropTable and DYNAMIC attribute.
** 085 CA  20131015          Added registration/deregistration of legacy names for a temp-table.
** 086 ECF 20131029          Import change and some methods made public for lock metadata support.
** 087 SVL 20131028          Added openScopeAt(), added legacyName and blockDepth parameter to
**                           constructors and define(), added tableHandle(), refactored multiplex
**                           scope closing.
** 088 CA  20131209          Added no-op deleted() method, required by the changes in Finalizable
**                           interface.
** 089 SVL 20140106          Changes related to copyAllRows.
** 090 SVL 20140123          Added support for extents in copyAllRows.
** 091 SVL 20140124          getResource is used for unwrapping handles.
** 092 SVL 20140205          copyAllRows supports looseCopy and skipUniqueConflicts modes.
** 093 SVL 20140210          Use HQLExpression instead of HQL string.
** 094 CA  20140214          Fixed readAllRows - use the actual DMO, not the DMO proxy.
** 095 SVL 20140221          OutputTableHandleCopier was made external; changes in
**                           createDynamicTable.
** 096 ECF 20140305          Removed explicit transaction processing from dropAllTables. Replaced
**                           Apache commons logging with J2SE logging. Adapted for API changes in
**                           RecordBuffer for copy operations.
** 097 OM  20140417          Dynamic temp-table buffers are directly available at the moment they 
**                           are declared instead of waiting for next block to open.
** 098 SVL 20140414          Use outputParameter instead of outputDMO + outputAppend in order to
**                           support remote parameters.
** 099 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 100 ECF 20140613          Fixed tableHandle to force the assignment regardless of pending error
**                           status; simplified tempTableResource.
** 101 SVL 20140709          insertAllRows was properly implemented. Other changes related to
**                           table parameters.
** 102 SVL 20140725          Fixed depth at which a buffer was opened in
**                           createDynamicBufferForTempTable. Added an overloaded version of
**                           TemporaryBuffer.define function.
** 103 ECF 20140814          Replaced RecordBuffer.flushAll with flush in copyAllRows method.
**                           Added extra-transaction validation for no-undo temp-tables.
** 104 ECF 20140920          Changed to accommodate memory fixes in DynamicTablesHelper. Adjusted
**                           openScope, openScopeAt, createTable to properly map table information
**                           for DMOs shared across dynamic temp-tables.
** 105 OM  20141003          COPY and COMPARE buffer operations use legacy names for pairing
**                           fields to be copied.
** 106 ECF 20141118          Fixed NPE in openMultiplexScope.
** 107 OM  20150107          Avoid double call of super.stateChanged(event) for INSERT events in
**                           stateChanged() method. Added stack trace when detected database synch
**                           is lost and documented event.
** 108 ECF 20150302          Registration of NoUndoValidator objects was too aggressive; only
**                           register them when buffer scopes explicitly are opened. Prevent commit
**                           or rollback when StopConditionException is thrown during copyAllRows.
** 109 OM  20150303          Added Dynamic Temp Buffers to HandeChain.
** 110 ECF 20150324          API change in TableMapper.
** 111 ECF 20150326          Added optional logging when dropping tables.
** 112 OM  20150504          Make sure the buffer is released when calling removeRecords().
**                           Removed warning about incorrect caching.
** 113 ECF 20150516          Changed readAllRows to sort by primary index (defaulting to primary
**                           key if no primary index is defined). Fixed dynamic resource cleanup.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.InstantiationException;
import java.lang.reflect.*;
import java.util.*;
import java.util.logging.*;
import org.hibernate.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.util.*;

/**
 * A specialized record buffer implementation which uses a temporary table
 * as backing storage and supports Progress' temp/work-table semantics.  A
 * new instance is created by invoking the static {@link #define} method,
 * which, like the superclass' method of the same name, creates a proxy
 * object which client code uses to interact with the database.
 * <p>
 * <b>Implementation Notes</b><br>
 * <b>Multiplexing.</b>  Because of limitations imposed by Hibernate's lack
 * of direct support for temp tables, a single temp table instance in the
 * backing database is multiplexed if multiple buffers of the same type of
 * DMO class are opened simultaneously in the same context.  This can happen
 * either when calling code explicitly defines multiple temp table buffers
 * using the same DMO type, or when method calls pass a temp table DMO proxy
 * as a parameter.  Multiplexing is used to enable Progress' semantic of
 * copying data between temp tables when they are passed as parameters to
 * nested procedure calls.
 * <p>
 * Multiplexing is achieved by adding a multiplex ID column to every temp
 * table schema.  Each buffer instance is assigned a unique multiplex ID the
 * first time a scope is opened on it.  Every access to the backing temp
 * table (i.e., queries, updates, inserts, deletes) take the multiplex ID
 * into account, such that the backing table is managed as multiple, virtual
 * table instances, one per distinct multiplex ID.
 * <p>
 * A temp table is created (along with its indexes, if any) the first time a
 * particular DMO type is accessed.  It is only dropped when no buffers
 * reference it.  As buffers go in and out of scope, a context local "in-use"
 * map is updated to reflect which buffers reference which physical temp
 * tables.  When a buffer goes out of scope, cleanup processing checks
 * whether it is the last buffer to reference its backing table.  If so, the
 * table is dropped.  If not, the table is not dropped, but all records with
 * that buffer's multiplex ID are removed from the table.
 * <p>
 * Temporary buffers may be scoped globally, or to the top-level (i.e.,
 * method) scope.  If global, cleanup is performed only when the context
 * itself is about to go out of scope.  If not global, cleanup is performed
 * when the method scope is popped by the transaction manager.  The latter
 * coincides with the external procedure scope in Progress.
 * <p>
 * There is a notion of a <i>master</i> buffer and <i>slave</i> buffers.  The
 * former is created as the result of converting a DEFINE TEMP-TABLE Progress
 * statement.  The latter is created as the result of converting a DEFINE
 * BUFFER FOR <i>temp-table</i> Progress statement.  A master buffer and all
 * of the associated slave buffers share the same multiplex ID, since these
 * buffers must all point to the same backing, virtual temp table for their
 * backing data.  The first time any buffer in such a group has a multiplex
 * ID assigned, all buffers in the group have their multiplex IDs assigned to
 * the same value.
 */
public final class TemporaryBuffer
extends RecordBuffer
{
   /** Name of the multiplex ID field in temp table DMO classes */
   static final String MULTIPLEX_FIELD_NAME = "_multiplex";
   
   /** Error text for error 12324. */
   public static final String ERROR_MISMATCHED_FIELDS = "Remote mismatched fields or " +
              "mismatched BEFORE-TABLE attribute for temp-table parameter or dataset " +
              "member parameter";
   
   /** Error text template for error 8030. */
   public static final String ERROR_TABLE_MISMATCH_TEMPLATE = "Parameter number %d (table %s) " +
                                        "mismatch. Has %d fields - client schema has %d fields";
   
   /** Error text template for error 8029. */
   public static final String ERROR_EXTENT_MISMATCH_TEMPLATE = "Table %s mismatch: The extent " +
                                        "number of field %d does not match the local definition";
   
   /** Error text template for error 8028. */
   public static final String ERROR_FIELD_TYPE_MISMATCH_TEMPLATE = "Table %s  mismatch: The " +
                "client data type of field %d does not match the definition on the AppServer";
   
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(TemporaryBuffer.class.getName());
   
   /** Multiplex ID sets currently in use for this context by DMO interface */
   private static final ContextLocal<Context> context =
   new ContextLocal<Context>()
   {
      @Override
      public WeightFactor getWeight() { return WeightFactor.LEVEL_2; }
      protected Context initialValue() { return (new Context()); }
   };
   
   /** Hibernate parameter types passed to copyAllRows() query */
   private static final Type[] COPY_ALL_ROWS_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { 0 });
   
   /** Hibernate parameter types passed to bulk delete state change query */
   private static final Type[] BULK_DELETE_EVENT_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { 0L });
   
   /** Hibernate parameter types passed to bulk delete check query */
   private static final Type[] BULK_DELETE_CHECK_QUERY_PARM_TYPES =
      DBUtils.makeTypeArray(new Object[] { 0L, 0 });
   
   /** Does this buffer represent a global temp table? */
   private final boolean global;
   
   /** Can property changes on this buffer be undone? */
   private final boolean undoable;
   
   /** Constructor used to create new instances of the backing DMO */
   private Constructor<?> dmoCtor = null;
   
   /** HQL used to check whether a record was deleted by a bulk delete */
   private String bulkDeleteEventHQL = null;
   
   /** HQL used to check whether a record was deleted from a virtual table */
   private String bulkDeleteCheckHQL = null;
   
   /** Destination table parameter for output copy if necessary before scope closes */
   private TableParameter outputParameter = null;
   
   /** Multiplex ID assigned to this buffer instance */
   private Integer multiplexID = null;
   
   /** Globally-scoped multipler for dynamic temp-tables (<code>null</code> for others) */
   private Multiplexer dynamicMultiplexer = null;
   
   /** Master buffer for this buffer;  null indicates this is the master */
   private TemporaryBuffer master = null;
   
   /** Reference to parent temp table; if master buffer exists this field is <code>null</code> */
   private TempTable tempTableRef = null;
   
   /** Order by clause associated with primary index of backing temp-table, if any */
   private String primaryOrderBy = null;
   
   /**
    * Constructor. This class is only instantiated via the {@link #define}
    * factory method.  The implementation class associated with the given
    * <code>dmoIface</code> must have a constructor which accepts a multiplex
    * ID and must contain a <code>_multiplex</code> instance variable of type
    * <code>Integer</code>.
    *
    * @param   dmoBufIface
    *          Interface which defines DMO's API.  This interface MUST extend
    *          {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer is associated
    *          in the transaction manager whenever it is registered.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer is defined or <code>-1</code> for the current scope.
    */
   private TemporaryBuffer(Class<?> dmoBufIface,
                           String variable,
                           boolean global,
                           boolean undoable,
                           int blockDepth)
   {
      super(dmoBufIface, DatabaseManager.TEMP_TABLE_DB.getName(), variable, blockDepth);
      
      this.global = global;
      this.undoable = undoable;
   }
   
   /**
    * Defer error that has occurred during {@link #copyAllRows} processing.
    */
   public static void deferCopyError()
   {
      TransactionManager.deferError("copying",
                                    new ErrorConditionException("Copy all rows error"),
                                    true);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoBufIface,
                                                String variable,
                                                boolean global)
   {
      return define(dmoBufIface, variable, null, global);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   legacyName
    *          Legacy name of the buffer.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoBufIface,
                                                String variable,
                                                String legacyName,
                                                boolean global)
   {
      TemporaryBuffer buffer = makeBuffer(dmoBufIface, variable, global, true);
      
      Map<TemporaryBuffer, Object> domain = new WeakHashMap<>();
      domain.put(buffer, null);
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains = context.get().domains;
      domains.put(buffer, domain);
      
      return createProxy(dmoBufIface, buffer, legacyName);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoBufIface,
                                                String variable,
                                                boolean global,
                                                boolean undoable)
   {
      return define(dmoBufIface, variable, null, global, undoable);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    * 
    * @param   <T> 
    *          Type of constructed instance.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   legacyName
    *          Legacy name of the buffer.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoBufIface,
                                                String variable,
                                                String legacyName,
                                                boolean global,
                                                boolean undoable)
   {
      return define(dmoBufIface, variable, legacyName, global, undoable, -1);
   }
   
   /**
    * Define a new buffer which is based upon the specified interface. The
    * buffer is permanently associated with the temp table database. It will
    * be registered with the transaction manager under the specified variable
    * name.
    *
    * @param   <T>
    *          Type of constructed instance.
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          sub-interface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   legacyName
    *          Legacy name of the buffer.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer is defined or <code>-1</code> for the current scope.
    *
    * @return  An object which implements the specified interface, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>dmoIface</code> is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Class<T> dmoBufIface,
                                                String variable,
                                                String legacyName,
                                                boolean global,
                                                boolean undoable,
                                                int blockDepth)
   {
      TemporaryBuffer buffer = makeBuffer(dmoBufIface, variable, global, undoable, blockDepth);
      
      Map<TemporaryBuffer, Object> domain = new WeakHashMap<>();
      domain.put(buffer, null);
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains = context.get().domains;
      domains.put(buffer, domain);
      
      return createProxy(dmoBufIface, buffer, legacyName);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer.
    * This is used as an analog to the DEFINE BUFFER statement when defining
    * a buffer for a temp-table or work-table.  The DMO interface and global
    * setting of the master buffer are used for the new buffer.  This buffer
    * will have the same multiplex ID as the master buffer, once it is
    * assigned to either buffer.
    *
    * @param   <T> 
    *          Type of constructed instance.
    * @param   template
    *          An instance of <code>Temporary</code> which is associated with
    *          a temporary buffer that serves as a template for the new
    *          buffer.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer, but which initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Temporary template,
                                                String variable)
   {
      return define(template, variable, null);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer.
    * This is used as an analog to the DEFINE BUFFER statement when defining
    * a buffer for a temp-table or work-table.  The DMO interface and global
    * setting of the master buffer are used for the new buffer.  This buffer
    * will have the same multiplex ID as the master buffer, once it is
    * assigned to either buffer.
    *
    * @param   <T>
    *          Type of constructed instance.
    * @param   template
    *          An instance of <code>Temporary</code> which is associated with
    *          a temporary buffer that serves as a template for the new
    *          buffer.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   legacyName
    *          Legacy name of the buffer.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer, but which initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Temporary template,
                                                String variable,
                                                String legacyName)
   {
      return define(template, variable, legacyName, -1);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer.
    * This is used as an analog to the DEFINE BUFFER statement when defining
    * a buffer for a temp-table or work-table.  The DMO interface and global
    * setting of the master buffer are used for the new buffer.  This buffer
    * will have the same multiplex ID as the master buffer, once it is
    * assigned to either buffer.
    *
    * @param   <T>
    *          Type of constructed instance.
    * @param   template
    *          An instance of <code>Temporary</code> which is associated with
    *          a temporary buffer that serves as a template for the new
    *          buffer.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer is defined or <code>-1</code> for the current scope.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer, but which initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Temporary template,
                                                String variable,
                                                int blockDepth)
   {
      return define(template, variable, null, blockDepth);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer.
    * This is used as an analog to the DEFINE BUFFER statement when defining
    * a buffer for a temp-table or work-table.  The DMO interface and global
    * setting of the master buffer are used for the new buffer.  This buffer
    * will have the same multiplex ID as the master buffer, once it is
    * assigned to either buffer.
    *
    * @param   <T> 
    *          Type of constructed instance.
    * @param   template
    *          An instance of <code>Temporary</code> which is associated with
    *          a temporary buffer that serves as a template for the new
    *          buffer.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   legacyName
    *          Legacy name of the buffer.
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer is defined or <code>-1</code> for the current scope.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer, but which initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   @SuppressWarnings("unchecked")
   public static <T extends Temporary> T define(Temporary template, 
                                                String variable,
                                                String legacyName,
                                                int blockDepth)
   {
      TemporaryBuffer m = (TemporaryBuffer) get(template);
      TemporaryBuffer masterBuf = (m.master == null ? m : m.master);
      Class<?> dmoBufIface = masterBuf.getDMOBufInterface();
      TemporaryBuffer buffer = makeBuffer(dmoBufIface,
                                          variable,
                                          masterBuf.global,
                                          masterBuf.undoable,
                                          blockDepth);
      buffer.master = masterBuf;
      buffer.multiplexID = masterBuf.multiplexID;
      
      Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains = context.get().domains;
      Map<TemporaryBuffer, Object> domain = domains.get(masterBuf);
      domain.put(buffer, null);
      
      return (T) createProxy(dmoBufIface, buffer, legacyName);
   }
   
   /**
    * Define a new buffer which is based upon a &quot;master&quot; buffer
    * stored in the {@link com.goldencode.p2j.util.SharedVariableManager
    * SharedVariableManager}.  This is used as an analog to the DEFINE SHARED
    * TEMP-TABLE statement when defining a buffer for a shared temp-table or
    * work-table which has been established elsewhere.  The DMO interface,
    * global, and undoability settings of the master buffer are used for the
    * new buffer.  This buffer will have the same multiplex ID as the master
    * buffer, which should have been assigned by an enclosing scope by the
    * time this method is invoked.
    *
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager and shared variable
    *          manager.
    *
    * @return  An object which implements the DMO interface of the master
    *          buffer stored in the shared variable manager, but which
    *          initially has no backing data record.
    *
    * @throws  IllegalArgumentException
    *          if <code>master</code>'s DMO interface is not assignable from
    *          <code>Temporary</code>.
    */
   public static Temporary useShared(String variable)
   {
      Temporary master = SharedVariableManager.lookupTempTable(variable);
      
      return define(master, variable, ((Buffer) master).name().getValue());
   }
   
   /**
    * Associate two temp tables with one another, such that all relevant
    * records can be copied between them at prescribed times.
    *
    * @param   src
    *          <code>TableParameter</code> containing reference to the source
    *          temporary buffer or result set.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs flushing the source buffer in
    *          preparation for an input copy or while copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   public static void associate(final TableParameter src,
                                final Temporary dstDMO,
                                boolean input,
                                boolean output)
   {
      associate(src, dstDMO, input, output, false);
   }
   
   /**
    * Associate two temp tables with one another, such that all relevant
    * records can be copied between them at prescribed times.
    *
    * @param   src
    *          <code>TableParameter</code> containing reference to the source
    *          temporary buffer or result set.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    * @param   appendInput
    *          <code>true</code> to add new records without affecting
    *          existing records;  <code>false</code> to remove all existing
    *          records before the copy takes place.  This option affects the
    *          <code>input</code> copy only, not the <code>output</code> copy.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs flushing the source buffer in
    *          preparation for an input copy or while copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   public static void associate(final TableParameter src,
                                final Temporary dstDMO,
                                boolean input,
                                boolean output,
                                boolean appendInput)
   {
      if (input)
      {
         if (!src.isRemoteParameter())
         {
            Temporary srcDMO = src.getTable();
            
            TemporaryBuffer srcBuf = (TemporaryBuffer) ((BufferReference) srcDMO).buffer();
            Persistence p = srcBuf.getPersistence();
            try
            {
               boolean shouldCommit = srcBuf.isTransient() &&
                                      p.beginTransaction();
               srcBuf.flush();
               if (shouldCommit)
               {
                  p.commit();
               }
            }
            catch (ValidationException | PersistenceException exc)
            {
               ErrorManager.recordOrThrowError(exc);
            }
            
            boolean res = copyAllRows(srcDMO, dstDMO, appendInput, CopyTableMode.INPUT_PARAM_MODE);
            if (!res)
               BlockManager.returnError();
         }
         else
         {
            insertAllRows(src.getResultSet(),
                          dstDMO,
                          appendInput,
                          CopyTableMode.INPUT_PARAM_MODE,
                          src.getParameterIndex());
         }
      }
      
      if (output)
      {
         // Set the destination DMO if output copy is required.  This is
         // actually the source DMO passed to this method, since the direction
         // of the copy is reversed on output.
         TemporaryBuffer dstBuf = (TemporaryBuffer) ((BufferReference) dstDMO).buffer();
         dstBuf.outputParameter = src;
         
         OutputTableCopier copier = new OutputTableCopier(dstBuf);
         TransactionManager.registerTopLevelFinalizable(copier, false);
      }
   }
   
   /**
    * Creates a dynamic temporary table for the given temp table handle. Intended to handle
    * conversion of DEFINE PARAMETER TABLE-HANDLE.
    *
    * @param   src
    *          <code>TableParameter</code> containing reference to the source table.
    * @param   target
    *          Handle for storing the target table.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    */
   public static void createDynamicTable(TableParameter src,
                                         handle target,
                                         boolean input,
                                         boolean output)
   {
      createDynamicTable(src, target, input, output, false);
   }
   
   /**
    * Creates a dynamic temporary table for the given temp table handle. Intended to handle
    * conversion of DEFINE PARAMETER TABLE-HANDLE.
    *
    * @param   src
    *          <code>TableParameter</code> containing reference to the source table.
    * @param   target
    *          Handle for storing the target table.
    * @param   input
    *          if <code>true</code>, records are copied from source table to
    *          destination table immediately upon invocation of this method,
    *          otherwise they are not.
    * @param   output
    *          if <code>true</code>, records are copied from destination
    *          table back to source table when the current transaction or
    *          subtransaction is committed;  otherwise they are not.
    * @param   append
    *          <code>true</code> to add new records without affecting
    *          existing records;  <code>false</code> to remove all existing
    *          records before the copy takes place.  This option affects the
    *          <code>input</code> copy only, not the <code>output</code> copy.
    */
   public static void createDynamicTable(TableParameter src,
                                         handle target,
                                         boolean input,
                                         boolean output,
                                         boolean append)
   {
      if (input)
      {
         handle srcHandle = src.getTableHandle();
         if (srcHandle._isValid())
         {
            TempTable dstTable = new TempTableBuilder();
            logical res = dstTable.copyTempTable(srcHandle, append, false, false, "");
            if (!res.booleanValue())
            {
               BlockManager.returnError();
            }
            target.assign(dstTable);
         }
         else
         {
            TempTableBuilder builder = new TempTableBuilder();
            builder.createFromTableWrapper(src.getResultSet(), src.getParameterIndex());
            target.assign(builder);
         }
      }
      
      if (output)
      {
         OutputTableHandleCopier copier = new OutputTableHandleCopier(src, target);
         TransactionManager.registerTopLevelFinalizable(copier, false);
         BufferManager.get().registerOutputTableHandleCopier(copier);
      }
   }
   
   /**
    * Get default buffer for a temporary table.
    *
    * @param  tempTableHandle
    *         Handle to a temporary table.
    *
    * @return see above.
    */
   public static Temporary getDefaultBuffer(handle tempTableHandle)
   {
      if (!tempTableHandle._isValid())
      {
         return null;
      }
      
      TempTable tempTable = (TempTable) tempTableHandle.getResource();
      return (Temporary) tempTable.defaultBufferHandle().getResource();
   }
   
   /**
    * Get an iterator on the names of all temp tables which currently exist
    * in the current context.
    * 
    * @return  Iterator on active table names.
    */
   public static Iterator<String> activeTables()
   {
      return context.get().activeTables.iterator();
   }
   
   /**
    * Report the number of temp tables which currently exist in the current
    * context.
    *
    * @return  Number of temp tables which have been created for the current
    *          context and not yet dropped.
    */
   public static int getTableCount()
   {
      return context.get().getTableCount();
   }
   
   /**
    * Populate the temp table with records created using the specified result set.
    * 
    * @param    srcResultSet
    *           The source result set.
    * @param    dstDMO
    *           DMO instance refering the destination DMO buffer.
    * @param    append
    *           <code>true</code> to append the data, <code>false</code> to remove all records
    *           first.
    * @param    errorMode
    *           Table copy mode, affects error handling.
    * @param    tableParameterIndex
    *           1-based index of the corresponding table parameter (in declaration of the
    *           function). Affects error handling.
    *
    * @return   <code>true</code> if copy was successful. <code>false</code> if an error occurs
    *           during copying records and error was recorder rather than thrown.
    *
    * @throws   ErrorConditionException
    *           if an error occurs during copying records.
    */
   public static boolean insertAllRows(TableWrapper srcResultSet,
                                       Temporary dstDMO,
                                       boolean append,
                                       CopyTableMode errorMode,
                                       int tableParameterIndex)
   {
      TemporaryBuffer dstBuf = (TemporaryBuffer) ((BufferReference) dstDMO).buffer();
      dstBuf.initialize();
      TempTable dstTempTable = dstBuf.getParentTable();
      
      // get map of destination setter and sizer methods, keyed by property name.
      Map<String, Method> dstSetters = dstBuf.getSetterMap();
      Map<String, Method> dstSizers = dstBuf.getSizerMap();
      Map<String, Method> dstGetters = dstBuf.getGetterMap();
      
      List<PropertyDefinition> srcPropList = srcResultSet.getProperties();
      int srcPropListSize                  = srcPropList.size();
      String[] srcProperties               = new String[srcPropListSize];
      for (int i = 0; i < srcPropListSize; i++)
      {
         srcProperties[i] = srcPropList.get(i).getName();
      }
      
      String[] dstProperties =
            TempTableBuilder.getOrderedPropertyNames(dstTempTable.getDMOClass(), dstTempTable);
      
      if (srcProperties.length != dstProperties.length)
      {
         if (errorMode == CopyTableMode.INPUT_PARAM_MODE)
         {
            String errorText = String.format(ERROR_TABLE_MISMATCH_TEMPLATE,
                                             tableParameterIndex,
                                             dstTempTable.name().toStringMessage(),
                                             dstProperties.length,
                                             srcProperties.length);
            throw new InvalidParameterConditionException(
                                                new NumberedException(errorText, 8030, false));
         }
         else if (errorMode == CopyTableMode.OUTPUT_TABLE_PARAM_MODE ||
                  errorMode == CopyTableMode.OUTPUT_TABLE_HANDLE_PARAM_MODE)
         {
            String errorText;
            if (srcResultSet.isTableHandle())
            {
               errorText = String.format(ERROR_TABLE_MISMATCH_TEMPLATE,
                                         tableParameterIndex,
                                         dstTempTable.name().toStringMessage(),
                                         dstProperties.length,
                                         srcProperties.length);
            }
            else
            {
               errorText = String.format(ERROR_TABLE_MISMATCH_TEMPLATE,
                                         tableParameterIndex,
                                         srcResultSet.getTableName(),
                                         srcProperties.length,
                                         dstProperties.length);
            }
            throw new InvalidParameterConditionException(
                                                new NumberedException(errorText, 8030, false));
         }
         else
         {
            throw new IllegalStateException("Wrong table copy error mode: " + errorMode);
         }
      }
      
      for (int i = 0; i < srcProperties.length; i++)
      {
         PropertyDefinition property = srcPropList.get(i);
         String dstPropName = dstProperties[i];
         
         // compare data types
         Method dstGetter = dstGetters.get(dstPropName);
         ParmType srcType = ParmType.fromClass(property.getType());
         ParmType dstType = ParmType.fromClass(dstGetter.getReturnType());
         
         if (srcType == null || dstType == null || !srcType.equals(dstType))
         {
            if (errorMode == CopyTableMode.INPUT_PARAM_MODE)
            {
               throw new InvalidParameterConditionException(
                                 new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
            }
            else if (errorMode == CopyTableMode.OUTPUT_TABLE_PARAM_MODE ||
                     errorMode == CopyTableMode.OUTPUT_TABLE_HANDLE_PARAM_MODE)
            {
               if (!srcResultSet.isTableHandle())
               {
                  throw new InvalidParameterConditionException(
                        new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
               }
               else
               {
                  String errorText = String.format(ERROR_FIELD_TYPE_MISMATCH_TEMPLATE,
                                                   dstTempTable.name().toStringMessage(),
                                                   i + 1);
                  ErrorManager.displayError(8028, errorText, false);
                  throw new InvalidParameterConditionException(
                        new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
               }
            }
            else
            {
               throw new IllegalStateException("Wrong table copy error mode: " + errorMode);
            }
         }
         
         // compare extent sizes
         Method dstSizer = dstSizers.get(dstPropName);
         long srcExtentSize = property.getExtent();
         if (srcExtentSize < 1)
            srcExtentSize = 1;
         
         long dstExtentSize =
               TempTableBuilder.getExtentSize(dstBuf.getDMOImplementationClass(), dstSizer);
         
         if (srcExtentSize != dstExtentSize)
         {
            if (errorMode == CopyTableMode.INPUT_PARAM_MODE)
            {
               throw new InvalidParameterConditionException(
                     new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
            }
            else if (errorMode == CopyTableMode.OUTPUT_TABLE_PARAM_MODE ||
                     errorMode == CopyTableMode.OUTPUT_TABLE_HANDLE_PARAM_MODE)
            {
               if (!srcResultSet.isTableHandle())
               {
                  throw new InvalidParameterConditionException(
                        new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
               }
               else
               {
                  String errorText = String.format(ERROR_EXTENT_MISMATCH_TEMPLATE,
                                                   dstTempTable.name().toStringMessage(),
                                                   i + 1);
                  ErrorManager.displayError(8029, errorText, false);
                  throw new InvalidParameterConditionException(
                        new NumberedException(ERROR_MISMATCHED_FIELDS, 12324, false));
               }
            }
            else
            {
               throw new IllegalStateException("Wrong table copy error mode: " + errorMode);
            }
         }
      }
      
      Persistence persistence = dstBuf.getPersistence();
      boolean savedAny = false;
      int commit = 0;   // 0 = do nothing, 1 = commit, -1 = rollback
      try
      {
         commit = (persistence.beginTransaction() ? 1 : 0);
         
         // if not in input-output mode, first remove all records
         if (!append)
         {
            dstBuf.removeRecords(null, null, true);
         }
         
         Iterator<Object[]> rowIter = srcResultSet.rowIterator();
         while (rowIter.hasNext())
         {
            Object[] row = rowIter.next();
            
            Persistable dstRec = dstBuf.instantiateDMO();
            Serializable id = dstBuf.nextPrimaryKey();
            
            for (int i = 0; i < dstProperties.length; i++)
            {
               Method setter = dstSetters.get(dstProperties[i]);
               Method sizer = dstSizers.get(dstProperties[i]);
               if (sizer == null)
               {
                  Object data = row[i];
                  setter.invoke(dstRec, data);
               }
               else
               {
                  // indexed setter invocation
                  BaseDataType[] data = (BaseDataType[]) row[i];
                  int size = ((int64) sizer.invoke(dstRec)).intValue();
                  for (int idx = 0; idx < size; idx++)
                  {
                     setter.invoke(dstRec, idx, data[idx]);
                  }
               }
            }
            
            persistence.save(dstRec, id);
            savedAny = true;
         }
      }
      catch (PersistenceException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         ErrorManager.recordOrThrowError(exc);
      }
      catch (IllegalAccessException | InstantiationException exc)
      {
         commit *= -1;
         
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InvocationTargetException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         
         throw new RuntimeException("Error creating record", exc);
      }
      finally
      {
         if (savedAny)
         {
            ChangeBroker.get().forcePendingFlush(dstBuf);
         }
         
         try
         {
            switch (commit)
            {
               case 1:
                  persistence.commit();
                  break;
               case -1:
                  persistence.rollback();
                  break;
               case 0:
                  break;
            }
         }
         catch (PersistenceException exc)
         {
            DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
            ErrorManager.recordOrThrowError(exc);
         }
      }
      
      return true;
   }
   
   /**
    * Open scope for the dynamic buffers which back the specified DMO instances at the given block
    * depth.
    *
    * @param blockDepth
    *        Zero-based depth of the block (starting from the outermost scope) at which the
    *        buffer scopes are opened.
    * @param dmos
    *        Variable length list of DMO instances returned by previous calls to {@link #define}.
    */
   public static void openScopeAt(int blockDepth, DataModelObject... dmos)
   {
      for (DataModelObject dmo : dmos)
      {
         get(dmo).openScopeAt(blockDepth);
      }
   }
   
   /**
    * Read all rows from the table targeted by the specified buffer and return each row as an
    * array, each row's values in the ordered specified by the given property list. 
    * 
    * @param    srcDMO
    *           The source DMO used to read the data.
    * @param    props
    *           The properties needed for this DMO.
    */
   static List<Object[]> readAllRows(Temporary srcDMO, List<PropertyDefinition> props)
   {
      // Get source and destination buffers, both of which must be
      // TemporaryBuffer instances.
      TemporaryBuffer srcBuf = (TemporaryBuffer) ((BufferReference) srcDMO).buffer();
      srcBuf.initialize();
      
      Map<String, Method> getters = srcBuf.getGetterMap();
      Persistence persistence = srcBuf.getPersistence();
      List<Object[]> data = new ArrayList<>();
      
      try
      {
         HQLExpression buf = new HQLExpression("from ");
         buf.append(srcBuf.getDMOImplementationName());
         // srcBuf is a TemporaryBuffer, so it isMultiplexed()
         buf.append(" where ");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ", true);
         // force the records to be ordered by their primary index (or by ID if none defined)
         String orderBy = srcBuf.getPrimaryOrderByClause();
         buf.append(orderBy);
         
         String hql = buf.toFinalExpression();
         Object[] parms = new Object[] { srcBuf.getMultiplexID() };
         ScrollableResults results = persistence.scroll(hql,
                                                        parms,
                                                        COPY_ALL_ROWS_QUERY_PARM_TYPES,
                                                        0,
                                                        0,
                                                        ScrollMode.FORWARD_ONLY);
         
         BufferManager bufMgr = BufferManager.get();
         
         while (results.next())
         {
            // Copy each source record into a new destination DMO instance.
            Persistable srcRec = (Persistable) results.get(0);
            
            int idx = 0;
            Object[] row = new Object[props.size()];
            for (PropertyDefinition property : props)
            {
               Method getter = getters.get(property.getName());
               if (property.isExtent())
               {
                  int extentSize = property.getExtent();
                  BaseDataType[] extentArray = new BaseDataType[extentSize];
                  
                  for (int i = 0; i < extentSize; i++)
                  {
                     extentArray[i] = (BaseDataType) getter.invoke(srcRec, i);
                  }
                  
                  row[idx++] = extentArray;
               }
               else
               {
                  row[idx++] = getter.invoke(srcRec);
               }
            }
            
            data.add(row);
            bufMgr.evictDMOIfUnused(persistence, srcRec);
         }
      }
      catch (PersistenceException exc)
      {
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         ErrorManager.recordOrThrowError(exc);
      }
      catch (IllegalArgumentException | IllegalAccessException | InvocationTargetException exc)
      {
         throw new RuntimeException("Error reading DMO property", exc);
      }
      
      return data;
   }
   
   /**
    * Create a dynamic buffer for the temporary table. For creation of default buffers for dynamic
    * tables use {@link #createDefaultDynamicBufferForTempTable(TempTableBuilder)}.
    *
    * @param  dmoIface
    *         DMO interface associated with the table.
    * @param  bufferName
    *         Target buffer name (4GL form).
    * @param  master
    *         Master buffer (default buffer of the target table).
    * @param  blockDepth
    *         Zero-based depth of the block (starting from the outermost scope) at which the
    *         scope of the create buffer is opened.
    *
    * @return DMO proxy of the created buffer.
    */
   static Temporary createDynamicBufferForTempTable(Class dmoIface,
                                                    String bufferName,
                                                    RecordBuffer master,
                                                    int blockDepth)
   {
      return createDynamicBufferForTempTable(dmoIface,
                                             bufferName,
                                             master,
                                             master.getParentTable(),
                                             blockDepth);
   }
   
   /**
    * Create the default dynamic buffer for the dynamic temporary table.
    *
    * @param  tempTable
    *         Temp-table object associated with the target table.
    *
    * @return DMO proxy of the created buffer.
    */
   static Temporary createDefaultDynamicBufferForTempTable(TempTableBuilder tempTable)
   {
      return createDynamicBufferForTempTable(tempTable.getDMOInterface(),
                                             tempTable.name().getValue(),
                                             null,
                                             tempTable,
                                             0);
   }

   /**
    * Create a dynamic buffer for the temporary table.
    *
    * @param  dmoIface
    *         DMO interface associated with the table.
    * @param  bufferName
    *         Target buffer name (4GL form).
    * @param  master
    *         Master buffer (default buffer of the target table). Can be <code>null</code> if the
    *         default buffer for the table is created.
    * @param  tempTable
    *         Temp-table object associated with the target table.
    * @param  blockDepth
    *         Zero-based depth of the block (starting from the outermost scope) at which the
    *         scope of the create buffer is opened.
    *
    * @return DMO proxy of the created buffer.
    */
   private static Temporary createDynamicBufferForTempTable(Class dmoIface,
                                                            String bufferName,
                                                            RecordBuffer master,
                                                            TempTable tempTable,
                                                            int blockDepth)
   {
      Class bufClass = dmoIface.getDeclaredClasses()[0];
      boolean undoable = tempTable.canUndo().booleanValue();
      final BufferManager bufferManager = BufferManager.get();
      String uniqueBufferName = bufferManager.generateUniqueBufferName(bufferName);
      Temporary temporary;
      if (master != null)
      {
         temporary = TemporaryBuffer.define((Temporary) master.getDMOProxy(),
                                            uniqueBufferName,
                                            bufferName,
                                            blockDepth);
      }
      else
      {
         temporary = TemporaryBuffer.define(bufClass,
                                            uniqueBufferName,
                                            bufferName,
                                            blockDepth == 0,
                                            undoable,
                                            blockDepth);
      }
      TemporaryBuffer temporaryBuffer = ((TemporaryBuffer) RecordBuffer.get(temporary));
      temporaryBuffer.setDynamic();
      temporaryBuffer.setTempTableRef(tempTable);
      ((BufferImpl) temporary).interlink(); // HandleChain implementation
      TemporaryBuffer.openScopeAt(blockDepth, temporary);
      
      if (tempTable._dynamic())
      {
         TempTableBuilder builder = (TempTableBuilder) tempTable;
         DynamicTablesHelper helper = DynamicTablesHelper.getInstance();
         helper.registerBufferForDynamicTable(builder, temporaryBuffer);
      }
      
      return temporary;
   }
   
   /**
    * Factory method to construct a new temporary buffer.
    *
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          subinterface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    *
    * @return  A new instance of this class.
    */
   private static TemporaryBuffer makeBuffer(Class<?> dmoBufIface,
                                             String variable,
                                             boolean global,
                                             boolean undoable)
   {
      return makeBuffer(dmoBufIface, variable, global, undoable, -1);
   }
   
   /**
    * Factory method to construct a new temporary buffer.
    *
    * @param   dmoBufIface
    *          Interface which defines DMO's public API.  This must be a
    *          subinterface of {@link Temporary}.
    * @param   variable
    *          Name of the variable with which the record buffer will be
    *          associated in the transaction manager.
    * @param   global
    *          <code>true</code> if the buffer survives until end of context
    *          life;  <code>false</code> if it expires upon leaving the top
    *          level procedure scope.
    * @param   undoable
    *          <code>true</code> if changes made to records managed by this
    *          buffer can be undone;  <code>false</code> if such changes are
    *          committed immediately, such that they cannot be undone.
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer is defined or <code>-1</code> for the current scope.
    *
    * @return  A new instance of this class.
    */
   private static TemporaryBuffer makeBuffer(Class<?> dmoBufIface,
                                             String variable,
                                             boolean global,
                                             boolean undoable,
                                             int blockDepth)
   {
      if (!DatabaseManager.isTempTable(dmoBufIface))
      {
         throw new IllegalArgumentException("DMO interface must extend Temporary");
      }
      
      return new TemporaryBuffer(dmoBufIface, variable, global, undoable, blockDepth);
   }
   
   /**
    * Get a context-local <code>Commitable</code> which needs to have its
    * methods invoked on commit/rollback.
    * 
    * @return  The <code>Commitable</code> object for the current context.
    */
   static Commitable getContextCommitable()
   {
      return context.get();
   }
   
   /**
    * Copy all records from one temp table to another, optionally appending
    * to records already present in the destination table.
    * <p>
    * TODO:  it seems this method can be optimized to use a DMO's deepCopy()
    * method instead of reflection when the source and destination DMO types
    * are the same.
    *
    * @param   srcDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the source temporary buffer.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on
    *          the destination temporary buffer.
    * @param   append
    *          <code>true</code> to add new records without affecting
    *          existing records;  <code>false</code> to remove all existing
    *          records before the copy takes place.
    * @param   errorMode
    *          Table copy mode, affects error handling.
    *
    * @return  <code>true</code> if copy was successful. <code>false</code> if validation error
    *          has occurred or the table is appended to itself.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the
    *          underlying DMOs to get or set data values.
    */
   @SuppressWarnings("unchecked")
   static boolean copyAllRows(Temporary srcDMO,
                              Temporary dstDMO,
                              boolean append,
                              CopyTableMode errorMode)
   {
      return copyAllRows(srcDMO, dstDMO, append, errorMode, false, false, false);
   }
   
   /**
    * Copy all records from one temp table to another, optionally appending* to records already
    * present in the destination table.
    * <p>
    * TODO:  it seems this method can be optimized to use a DMO's deepCopy() method instead of
    * reflection when the source and destination DMO types are the same.
    *
    * @param   srcDMO
    *          DMO instance returned by a previous call to {@link #define} on the source
    *          temporary buffer.
    * @param   dstDMO
    *          DMO instance returned by a previous call to {@link #define} on the
    *          destination temporary buffer.
    * @param   append
    *          <code>true</code> to add new records without affecting existing records;
    *          <code>false</code> to remove all existing records before the copy takes place.
    * @param   errorMode
    *          Table copy mode, affects error handling.
    * @param   skipUniqueConflicts
    *          Corresponds append-mode in COPY-TEMP-TABLE. If there is a unique index on the
    *          target temp-table and a row with a duplicate key is found, the row is not replaced.
    *          This parameter is ignored if <code>replacePrimaryRecords</code> is
    *          <code>true</code>.
    * @param   replacePrimaryRecords
    *          TODO implement support
    *          Corresponds replace-mode in COPY-TEMP-TABLE. When <code>true</code> records in the
    *          target temp-table are replaced with corresponding records from the source
    *          temp-table. The target temp-table must have a unique primary index that be can used
    *          to find the corresponding record. When the corresponding record is found in the
    *          target temp-table, the target record is replaced with the source record.
    * @param   looseCopy
    *          Corresponds loose-copy-mode in COPY-TEMP-TABLE. Only the fields with the same name
    *          that appear in both the source and target temp-table are copied. Other fields are
    *          ignored.
    *
    * @return  <code>true</code> if copy was successful. <code>false</code> if validation error
    *          has occurred or the table is appended to itself.
    *
    * @throws  ErrorConditionException
    *          if a recoverable error occurs copying records.
    * @throws  RuntimeException
    *          if an unrecoverable error occurs invoking methods on the underlying DMOs to get or
    *          set data values.
    */
   @SuppressWarnings("unchecked")
   static boolean copyAllRows(Temporary srcDMO,
                              Temporary dstDMO,
                              boolean append,
                              CopyTableMode errorMode,
                              boolean skipUniqueConflicts,
                              boolean replacePrimaryRecords,
                              boolean looseCopy)
   {
      // Get source and destination buffers, both of which must be
      // TemporaryBuffer instances.
      TemporaryBuffer srcBuf = (TemporaryBuffer) ((BufferReference) srcDMO).buffer();
      TemporaryBuffer dstBuf = (TemporaryBuffer) ((BufferReference) dstDMO).buffer();
      
      srcBuf.initialize();
      dstBuf.initialize();
      
      try
      {
         srcBuf.flush();
         dstBuf.flush();
      }
      catch (ValidationException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      
      TempTable srcTempTable = srcBuf.getParentTable();
      TempTable dstTempTable = dstBuf.getParentTable();
      
      String srcBufName = ((HandleChain) srcDMO).name().getValue();
      String dstBufName = ((HandleChain) dstDMO).name().getValue();
      
      if (srcTempTable.equals(dstTempTable))
      {
         if (errorMode == CopyTableMode.COPY_TEMP_TABLE_MODE)
         {
            String msg = "Source and target are the same table in temp-table copy " +
                         "operation for " + dstBufName;
            ErrorManager.displayError(12301, msg, false);
            return false;
         }
         
         if (append)
         {
            String msg = srcTempTable.name().getValue() + " cannot be appended to itself";
            ErrorManager.recordOrShowError(5362, msg, false, false);
            return false;
         }
         
         return true;
      }
      
      // Get maps of source getter methods and destination setter methods,
      // each keyed by property name.
      Map<String, Method> srcGetters = srcBuf.getGetterMap();
      Map<String, Method> srcSizers = srcBuf.getSizerMap();
      Map<String, Method> setters = dstBuf.getSetterMap();
      Map<String, Method> dstGetters = dstBuf.getGetterMap();
      Map<String, Method> dstSizers = dstBuf.getSizerMap();
      
      // PROGRESS does not care if the field names are different; what it
      // cares about are only the field type and position - if a source field
      // has the same position and type as the destination field, then they
      // are a match. If the source and destination fields at a certain
      // position do not match, then PROGRESS will throw an error and abend
      // the application.
      
      // if the backing tables are different, check if the fields match
      Map<DatumAccess, DatumAccess> propsMap = null;
      
      if (!srcBuf.getTable().equals(dstBuf.getTable()))
      {
         // we need to modify the setter properties, so instantiate a new map
         setters = new HashMap<>(dstBuf.getSetterMap());
         propsMap = new HashMap<>();
         
         String[] srcProperties =
               TempTableBuilder.getOrderedPropertyNames(srcTempTable.getDMOClass(), srcTempTable);
         String[] dstProperties =
               TempTableBuilder.getOrderedPropertyNames(dstTempTable.getDMOClass(), dstTempTable);
         
         if (!looseCopy && (srcProperties.length != dstProperties.length))
         {
            handleTablesDoNotMatch(errorMode, srcBufName, dstBufName);
            return false;
         }
         
         String[] srcPropertiesLegacy = null;
         Map<String, Integer> dstPropertiesLegacy = null;
         if (looseCopy)
         {
            srcPropertiesLegacy = new String[srcProperties.length];
            for (int i = 0; i < srcProperties.length; i++)
            {
               srcPropertiesLegacy[i] =
                  TempTableBuilder.getLegacyFieldName(srcTempTable.getDMOClass(),
                                                      srcProperties[i],
                                                      srcTempTable);
            }
            
            dstPropertiesLegacy = new HashMap<>();
            for (int i = 0; i < dstProperties.length; i++)
            {
               String legacyName = TempTableBuilder.getLegacyFieldName(dstTempTable.getDMOClass(),
                                                                       dstProperties[i],
                                                                       dstTempTable);
               dstPropertiesLegacy.put(legacyName, i);
            }
         }
         
         for (int i = 0; i < srcProperties.length; i++)
         {
            String dstPropName;
            if (looseCopy)
            {
               Integer idx = dstPropertiesLegacy.get(srcPropertiesLegacy[i]);
               if (idx == null)
               {
                  continue;
               }
               dstPropName = dstProperties[idx];
            }
            else
            {
               dstPropName = dstProperties[i];
            }
            
            // compare data types
            Method srcGetter = srcGetters.get(srcProperties[i]);
            Method srcSizer = srcSizers.get(srcProperties[i]);
            Method dstGetter = dstGetters.get(dstPropName);
            Method dstSizer = dstSizers.get(dstPropName);
            ParmType srcType = ParmType.fromClass(srcGetter.getReturnType());
            ParmType dstType = ParmType.fromClass(dstGetter.getReturnType());
            
            if (srcType == null ||
                dstType == null ||
                !srcType.equals(dstType))
            {
               if (looseCopy)
               {
                  continue;
               }
               else
               {
                  handleTablesDoNotMatch(errorMode, srcBufName, dstBufName);
                  return false;
               }
            }
            
            // compare extent sizes
            long srcExtentSize =
                  TempTableBuilder.getExtentSize(srcBuf.getDMOImplementationClass(), srcSizer);
            long dstExtentSize =
                  TempTableBuilder.getExtentSize(dstBuf.getDMOImplementationClass(), dstSizer);
            
            if (srcExtentSize != dstExtentSize)
            {
               if (looseCopy)
               {
                  continue;
               }
               else
               {
                  handleTablesDoNotMatch(errorMode, srcBufName, dstBufName);
                  return false;
               }
            }
            
            DatumAccess srcAcc = new DatumAccess(srcProperties[i], null);
            DatumAccess dstAcc = new DatumAccess(dstPropName, null);
            propsMap.put(srcAcc, dstAcc);
         }
      }
      
      // Establish the set of properties being copied. This set will contain ALL source getter
      // properties, as the setters and getters maps were synced.
      if (propsMap == null)
      {
         Map<String, String> srcFieldMap = srcBuf.getLegacyFieldNameMap().getLegacyField2Name();
         Map<String, String> dstFieldMap = dstBuf.getLegacyFieldNameMap().getLegacyField2Name();
         
         propsMap = RecordBuffer.createSimplePropsMap(srcFieldMap.keySet(),
                                                      srcGetters,
                                                      dstGetters,
                                                      srcFieldMap,
                                                      dstFieldMap,
                                                      OperationType.COPY);
      }
      
      Persistence persistence = srcBuf.getPersistence();
      boolean savedAny = false;
      int commit = 0;   // 0 = do nothing, 1 = commit, -1 = rollback
      boolean validationError = false;
      try
      {
         commit = (persistence.beginTransaction() ? 1 : 0);
         
         // if not in append mode, first remove records (if any) from destination table.
         if (!append)
         {
            dstBuf.removeRecords(null, null, true);
         }
         
         // get the list of source records to be copied.  These will be all records in srcBuf's
         // backing temp table which have srcDMO's multiplex ID.
         HQLExpression buf = new HQLExpression("from ");
         buf.append(srcBuf.getDMOImplementationName());
         // srcBuf is a TemporaryBuffer, so it isMultiplexed()
         buf.append(" where ");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ", true);
         
         // force the records to be ordered by their ID
         // TODO in copy-temp-table sorting order affects which record remains in constraints
         // violation case. The correct sorting is by primary index if it presents or by
         // the index with the lexicographically first name.
         buf.append(" order by id");
         
         String hql = buf.toFinalExpression();
         Object[] parms = new Object[] { srcBuf.getMultiplexID() };
         ScrollableResults results =
            persistence.scroll(hql,
                               parms,
                               COPY_ALL_ROWS_QUERY_PARM_TYPES,
                               0,
                               0,
                               ScrollMode.FORWARD_ONLY);
         
         BufferManager bufMgr = BufferManager.get();
         
         while (results.next())
         {
            // Copy each source record into a new destination DMO instance.
            Persistable srcRec = (Persistable) results.get(0);
            Persistable dstRec = dstBuf.instantiateDMO();
            Serializable id = dstBuf.nextPrimaryKey();
            copy(srcRec, (DataModelObject) dstRec, propsMap, srcGetters, setters, srcSizers);
            bufMgr.evictDMOIfUnused(persistence, srcRec);
            if (LOG.isLoggable(Level.FINEST))
            {
               String msg =
                  "Copy temp table record from MPID " +
                  srcBuf.getMultiplexID() + " to MPID " +
                  dstBuf.getMultiplexID() + "...";
               LOG.log(Level.FINEST, msg);
               LOG.log(Level.FINEST, srcBuf.describe() + srcBuf.toString(srcRec));
               LOG.log(Level.FINEST, dstBuf.describe() + dstBuf.toString(dstRec));
            }
            
            // validate a record against the target table
            // TODO also validate against already copied records
            try
            {
               dstRec.setId((Long) id);
               dstBuf.validate(dstRec, false);
            }
            catch (ValidationException e)
            {
               if (skipUniqueConflicts)
               {
                  continue;
               }
               else
               {
                  ErrorManager.recordOrShowError(e.getNumber(), e.getMessage(), false);
                  commit *= -1;
                  validationError = true;
                  break;
               }
            }
            
            persistence.save(dstRec, id);
            savedAny = true;
         }
      }
      catch (PersistenceException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         ErrorManager.recordOrThrowError(exc);
      }
      catch (IllegalAccessException | InstantiationException exc)
      {
         commit *= -1;
         
         throw new RuntimeException("Error creating record", exc);
      }
      catch (InvocationTargetException exc)
      {
         commit *= -1;
         DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
         
         throw new RuntimeException("Error creating record", exc);
      }
      catch (StopConditionException exc)
      {
         // thrown by Persistence when a JDBC-level error was encountered; Hibernate session will
         // have been closed, don't try to commit or rollback below
         commit = 0;
         
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Error copying all temp table records", exc);
         }
         
         throw exc;
      }
      finally
      {
         if (savedAny)
         {
            ChangeBroker.get().forcePendingFlush(dstBuf);
         }
         
         try
         {
            switch (commit)
            {
               case 1:
                  persistence.commit();
                  break;
               case -1:
                  persistence.rollback();
                  break;
               case 0:
                  break;
            }
         }
         catch (PersistenceException exc)
         {
            DBUtils.handleException(DatabaseManager.TEMP_TABLE_DB, exc);
            ErrorManager.recordOrThrowError(exc);
         }
      }
      
      return !validationError;
   }
   
   /**
    * If <code>copyTempTableErrorMode</code> is <code>true</code> then dispaly "The caller's
    * temp-table parameter {srcBufName} does not match to the target temp-table {dstBufName}"
    * error message and abend. Else display "COPY-TEMP-TABLE requires {srcBufName} and
    * {dstBufName} columns to match exactly unless the fourth loose-mode parameter is passed as
    * TRUE" error message.
    *
    * @param errorMode
    *        Table copy mode, affects error handling.
    * @param srcBufName
    *        Source buffer name.
    * @param dstBufName
    *        Destination buffer name.
    */
   static void handleTablesDoNotMatch(CopyTableMode errorMode,
                                      String srcBufName,
                                      String dstBufName)
   {
      NumberedException e;
      switch (errorMode)
      {
         case COPY_TEMP_TABLE_MODE:
            e = new NumberedException("COPY-TEMP-TABLE requires " + srcBufName + " and " +
                  dstBufName + " columns to match exactly unless the fourth loose-mode " +
                  "parameter is passed as TRUE", 12303, false);
            ErrorManager.displayError(e.getNumber(), e.getMessage(), e.isPrefix());
            throw new InvalidParameterConditionException(e);
            
         case INPUT_PARAM_MODE:
            e = new NumberedException("The caller's temp-table parameter " + srcBufName +
                        " does not match to the target temp-table " + dstBufName, 5363, false);
            ErrorManager.displayError(e.getNumber(), e.getMessage(), e.isPrefix());
            throw new InvalidParameterConditionException(e);
            
         case OUTPUT_TABLE_PARAM_MODE:
            e = new NumberedException("The caller's temp-table parameter " + dstBufName +
                        " does not match to the target temp-table " + srcBufName, 5363, false);
            ErrorManager.displayError(e.getNumber(), e.getMessage(), e.isPrefix());
            throw new InvalidParameterConditionException(e);
            
         case OUTPUT_TABLE_HANDLE_PARAM_MODE:
            ErrorManager.displayError(5363,
                  "The caller's temp-table parameter " + dstBufName +
                        " does not match to the target temp-table " + srcBufName, false);
            e = new NumberedException(
                  "Try switching TEMP-TABLE method ADD-FIELDS-FROM for CREATE-LIKE", 9030, false);
            ErrorManager.displayError(e.getNumber(), e.getMessage(), e.isPrefix());
            throw new InvalidParameterConditionException(e);
      }
   }
   
   /**
    * Restores this buffer's state to that which it had at the beginning of
    * the current block.  In particular, this rolls back all the reversible
    * items for the current scope and removes these entries from the rollback
    * dictionaries managed by the <code>BufferManager</code>.
    * <p>
    * Auto-commit buffers are never rolled back.
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction rollback;  <code>false</code> if it represents a
    *          sub-transaction rollback.
    */
   public void rollback(boolean transaction)
   {
      if (!isUndoable())
      {
         // in case of the no-undo tables, we need to re-apply all the changes 
         // recorded during this transaction; so, the empty flag needs to be 
         // reset
         // TODO: maybe a restoration mechanism would be better ?
         Context local = context.get();
         local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
      }
      
      super.rollback(transaction);
   }
   
   /**
    * Commit the changes made within a subtransaction.  This entails:
    * <ul>
    *   <li>copying all reversible, record-level actions associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    *   <li>copying all undoable, property-level changes associated with the
    *       current scope to the next higher scope, so rollback capability
    *       for these actions is not lost when the scope is popped;
    * </ul>
    * <p>
    * Nothing is done if this is a full transaction commit, because no
    * rollback will be possible at the parent scope once the database-level
    * transaction is committed.
    * <p>
    * If this is a full transaction and there are pending bulk actions, force
    * those actions to be executed (i.e. force the delete).
    *
    * @param   transaction
    *          <code>true</code> if this commit represents the master
    *          transaction commit;  <code>false</code> if it represents a
    *          sub-transaction commit.
    */
   public void commit(boolean transaction)
   {
      List<PersistenceException> errors = new ArrayList<>();
      
      if (transaction && isUndoable())
      {
         // flush the bulk deleted records which had their multiplex ID
         // changed; this is performed only for undoable tables.
         Set<Reversible> r0 = bufferManager.getReversibles(this, 0, false);
         if (r0 != null && !r0.isEmpty())
         {
            ReversibleBulkDelete rev = null;
            try 
            {
               Iterator<Reversible> iter = r0.iterator();
               while (iter.hasNext())
               {
                  Reversible r = iter.next();
                  
                  if (r instanceof ReversibleBulkDelete)
                  {
                     rev = (ReversibleBulkDelete) r;
                     rev.forceDelete();
                     iter.remove();
                  }
               }
            }
            catch (PersistenceException exc)
            {
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.log(Level.SEVERE,
                          describe() + ":  error rolling back undoable bulk delete " + rev,
                          exc);
               }
               
               errors.add(exc);
            }
         }
      }
      super.commit(transaction);
      
      // Report or eat errors.
      for (PersistenceException exc : errors)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Respond to a record change event.  Override the parent's implementation
    * to handle bulk delete events.  If our current record no longer exists in
    * the database, we must release it.
    *
    * @param   event
    *          Event which describes the DMO state change.
    *
    * @throws  PersistenceException
    *          if an error occurs executing a query to detect whether current
    *          record still exists.
    */
   public void stateChanged(RecordChangeEvent event)
   throws PersistenceException
   {
      if (event.isInsert())
      {
         Context local = context.get();
         local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
         
         super.stateChanged(event);
      }
      else if (event.isBulkDelete())
      {
         if (!isActive())
         {
            return;
         }
         
         Persistable dmo = getCurrentRecord();
         if (dmo == null)
         {
            // No release necessary.
            return;
         }
         
         // We can avoid issuing a query if the event represents a full bulk
         // delete and the deleted multiplex ID matches our own, because we
         // know our record must have been deleted.
         if (event.isFullBulkDelete())
         {
            Integer srcMPID = event.getSource().getMultiplexID();
            if (srcMPID.equals(multiplexID))
            {
               release();
               
               return;
            }
         }
         
         // Query whether the buffer's current record still exists.
         Serializable id = dmo.getId();
         Persistence persistence = getPersistence();
         if (bulkDeleteEventHQL == null)
         {
            String entity = getEntityName();
            HQLExpression buf = new HQLExpression();
            buf.append("select tt.id from ");
            buf.append(entity);
            buf.append(" tt where tt.id = ", true);
            bulkDeleteEventHQL = buf.toFinalExpression();
         }
         List list = persistence.list(bulkDeleteEventHQL,
                                      new Object[] { id },
                                      BULK_DELETE_EVENT_QUERY_PARM_TYPES,
                                      1,
                                      0,
                                      true);
         if (list == null)
         {
            release();
         }
      }
      else
      {
         super.stateChanged(event);
      }
   }
   
   /**
    * Override the superclass method to always return LockType.NONE. Locking
    * is a no-op for temp table records, since each user has exclusive access
    * to their temp table instance.
    *
    * @param  id
    *         Not used.
    *
    * @return LockType.NONE
    */
   public LockType getPinnedLockType(Serializable id)
   {
      return LockType.NONE;
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    *
    * @param   id
    *          Not used.
    * @param   pinnedLockType
    *          Not used.
    */
   public void setPinnedLockType(Serializable id, LockType pinnedLockType)
   {
   }
   
   /**
    * Returns TABLE-HANDLE for the buffer.
    *
    * @return TABLE-HANDLE for the buffer.
    */
   public handle tableHandle()
   {
      handle h = new handle();
      if (tempTableRef != null)
      {
         h.assign(tempTableRef, true);
      }
      
      return h;
   }
   
   /**
    * Set reference to the parent temp table.
    *
    * @param tempTableRef
    *        Parent temp table.
    */
   void setTempTableRef(TempTable tempTableRef)
   {
      this.tempTableRef = tempTableRef;
   }
   
   /**
    * Close multiplex scope and optionally drop the temporary table.
    *
    * @param dropTable
    *        <code>true</code> if the table should be dropped in addition to scope closing.
    */
   void closeMultiplexScope(boolean dropTable)
   {
      boolean multplexClosed = isMultiplexScopeClosed();
      Context local = context.get();
      Class<?> dmoIface = getDMOInterface();
      boolean tableDropped = !local.openTables.containsKey(dmoIface);
      
      if (!multplexClosed || (dropTable && !tableDropped))
      {
         try
         {
            Persistence persistence = getPersistence();
            boolean shouldCommit = persistence.beginTransaction();
            
            if (!multplexClosed)
            {
               doCloseMultiplexScope();
            }
            
            if (dropTable && !tableDropped)
            {
               local.openTables.remove(dmoIface); // prevent from trying to drop the table twice
                                                  // in dropAllTables
               local.doDropTable(dmoIface);
               
               TableMapper.removeTemporaryTable(tempTableResource());
            }
            
            if (shouldCommit)
            {
               persistence.commit();
            }
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
      }
   }
   
   /**
    * Copy output data for the linked table parameter, if any.
    */
   void performOutputCopy()
   {
      if (outputParameter != null)
      {
         Temporary tempDMO = (Temporary) getDMOProxy();
         if (!outputParameter.isRemoteParameter())
         {
            boolean res = copyAllRows(tempDMO,
                                      outputParameter.getTable(),
                                      outputParameter.isAppend(),
                                      CopyTableMode.OUTPUT_TABLE_PARAM_MODE);
            if (!res)
               deferCopyError();
         }
         else
         {
            try
            {
               TempTableResultSet resultSet = new TempTableResultSet(tempDMO,
                                                                     false,
                                                                     true,
                                                                     outputParameter.isAppend());
               TableWrapper tableWrapper = new TableWrapper(resultSet);
               tableWrapper.setTableName(getParentTable().name().toStringMessage());
               outputParameter.setResultSet(tableWrapper);
            }
            catch (ClassNotFoundException e)
            {
               deferCopyError();
            }
         }

         outputParameter = null;
      }
   }

   /**
    * Initialize the core data on which this buffer relies.  This should be
    * called only once per buffer instance, when the buffer's scope is first
    * opened.
    * 
    * @return  <code>true</code> if the buffer was initialized with this call;
    *          <code>false</code> if the buffer had been initialized with a
    *          previous call to this method.
    *
    * @throws  IllegalArgumentException
    *          if no implementation class is found for the given database and
    *          interface.
    * @throws  RuntimeException
    *          if the backing DMO implementation class does not implement a
    *          constructor which accepts a multiplex ID.
    */
   protected boolean initialize()
   {
      if (!super.initialize())
      {
         return false;
      }
      
      Class<?> dmoCls = getDMOImplementationClass();
      try
      {
         dmoCtor = dmoCls.getDeclaredConstructor(Integer.class);
         Field fld = dmoCls.getDeclaredField(MULTIPLEX_FIELD_NAME);
         if (!fld.getType().equals(Integer.class))
         {
            String msg = "Invalid temporary table DMO:  " +
                         dmoCls.getName() +
                         "." +
                         MULTIPLEX_FIELD_NAME +
                         " field must be of type Integer";
            
            throw new RuntimeException(msg);
         }
      }
      catch (NoSuchMethodException exc)
      {
         String msg = "Invalid temporary table DMO:  " +
                      dmoCls.getName() +
                      " does not implement a multiplex constructor";
            
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, describe() + ":  " + msg, exc);
         }
         
         throw new RuntimeException(msg, exc);
      }
      catch (NoSuchFieldException exc)
      {
         String msg = "Invalid temporary table DMO:  " +
                      dmoCls.getName() +
                      " does not declare a " +
                      MULTIPLEX_FIELD_NAME +
                      " field";
            
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, describe() + ":  " + msg, exc);
         }
         
         throw new RuntimeException(msg, exc);
      }
      
      return true;
   }
   
   /**
    * Get the logical database name associated with this buffer.  This will
    * always be the standard name for the temp table database as defined in
    * {@link DatabaseManager#TEMP_TABLE_DB}, regardless of the value of the
    * <code>ldbOrAlias</code> parameter.
    *
    * @param   ldbOrAlias
    *          A logical database name or database alias.
    *
    * @return  Standard, temp table database name.
    */
   protected String getLDBName(String ldbOrAlias)
   {
      return DatabaseManager.TEMP_TABLE_DB.getName();
   }
   
   /**
    * Get the database information object associated with this buffer.  This
    * will always be the standard temp table database as defined in
    * {@link DatabaseManager#TEMP_TABLE_DB}, regardless of the value of the
    * <code>ldbName</code> parameter.
    *
    * @param   ldbName
    *          A logical database name.
    *
    * @return  Standard, temp table database.
    */
   protected Database getPDB(String ldbName)
   {
      return DatabaseManager.TEMP_TABLE_DB;
   }
   
   /**
    * Register this buffer with the connection manager.
    * <p>
    * This implementation overrides the parent method to do nothing, since
    * accesses to the temp table database are outside the scope of the
    * connection manager's responsibilities.
    */
   protected void openConnectionManagerScope()
   {
   }
   
   /**
    * Deregister this buffer with the connection manager.
    * <p>
    * This implementation overrides the parent method to do nothing, since
    * accesses to the temp table database are outside the scope of the
    * connection manager's responsibilities.
    */
   protected void closeConnectionManagerScope()
   {
   }
   
   /**
    * Create a new instance of the DMO implementation class associated with
    * this buffer, and assign this buffer's multiplex ID to it.  This method
    * must not be invoked before a buffer scope has been opened on this
    * buffer.
    *
    * @return  A transient DMO instance.
    *
    * @throws  InstantiationException
    *          if the class cannot be instantiated.
    * @throws  IllegalAccessException
    *          if the DMO class' target constructor cannot be accessed.
    * @throws  InvocationTargetException
    *          if the underlying constructor throws an exception.
    */
   protected Persistable instantiateDMO()
   throws InstantiationException,
          IllegalAccessException,
          InvocationTargetException
   {
      if (multiplexID == null)
      {
         throw new IllegalStateException(
            "Cannot create a new record before buffer scope is opened");
      }
      
      return (Persistable) dmoCtor.newInstance(multiplexID);
   }
   
   /**
    * Convenience method to indicate whether this buffer is backed by a
    * temporary table.
    * 
    * @return  <code>true</code>.
    */
   protected boolean isTemporary()
   {
      return true;
   }
   
   /**
    * Indicate whether the table backing this buffer is known to contain no
    * rows.
    * <p>
    * This implementation involves only the virtual table backing the current
    * buffer (i.e., whose multiplex ID matches that of this buffer), not the
    * entire physical table.
    * 
    * @return  <code>true</code> if the table has no rows;  <code>false</code>
    *          if the table might have rows.
    */
   protected boolean isTableDefinitelyEmpty()
   {
      Context local = context.get();
      Boolean empty = local.emptyTableFlags.get(multiplexID);
      
      return (empty != null && empty);
   }
   
   /**
    * Indicate whether changes made to records managed by this buffer can be
    * undone during a rollback.
    *
    * @return  <code>true</code> if record changes can be undone, else
    *          <code>false</code>.
    */
   protected boolean isUndoable()
   {
      return undoable;
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    *
    * @throws  PersistenceException
    *          never.
    */
   protected void upgradeLock()
   throws PersistenceException
   {
   }
   
   /**
    * Override the parent's implementation to do nothing.
    */
   protected void pushPinnedLockScope()
   {
   }
   
   /**
    * Override the parent's implementation to do nothing.
    * 
    * @param   transaction
    *          <code>true</code> if we are at a full transaction boundary,
    *          else <code>false</code>.
    * @param   undo
    *          <code>true</code> if we are processing an undo on the current
    *          scope, else <code>false</code>.
    */
   protected void popPinnedLockScope(boolean transaction, boolean undo)
   {
   }
   
   /**
    * Get the multiplex ID of this temporary buffer.  This will be
    * <code>null</code> if this method is invoked before a buffer scope is
    * opened on this buffer for the first time.  Otherwise, it will be a
    * unique identifier within the current context.
    *
    * @return  Multiplex ID for this temporary buffer.
    */
   protected Integer getMultiplexID()
   {
      return multiplexID;
   }
   
   /**
    * Indicate whether the backing table for this buffer is multiplexed;  that
    * is, whether the backing table represents multiple, virtual tables by
    * using a multiplex ID to differentiate table segments.
    * <p>
    * Temp tables always are multiplexed.
    *  
    * @return  <code>true</code>.
    */
   protected boolean isMultiplexed()
   {
      return true;
   }
   
   /**
    * Indicate whether this buffer implementation allows bulk actions.
    *  
    * @return  <code>true</code>.
    */
   protected boolean allowsBulkActions()
   {
      return true;
   }
   
   /**
    * Retrieve the master record buffer for this buffer.
    * <p>
    * This only has meaning for shared temp tables.  For a temp table defined
    * as NEW SHARED, this buffer is returned.  For a temp table defined as
    * SHARED, the domain master buffer is returned.
    * 
    * @return  The master buffer.
    */
   protected RecordBuffer getMasterBuffer()
   {
      return (master != null ? master : this);
   }
   
   /**
    * Open a new buffer scope for this record buffer.  Open a new multiplex
    * scope and assign a multiplex ID the first time we are invoked for a
    * buffer instance.
    */
   protected void openScope()
   {
      super.openScope();
      
      // Only need to process if we are opening the outermost scope for this buffer.  Nested
      // scopes need not apply.
      if (getOpenScopeCount() == 1)
      {
         // A temporary buffer can be initialized as soon as its first scope is opened, because
         // the _temp database is permanently connected.
         initialize();
         
         // If this is the first time this table is being used, create it in the database.
         if (openMultiplexScope())
         {
            context.get().createTable(getDMOInterface(), getTable());
         }
         
         // map legacy info (this is tolerant to being called more than once)
         TableMapper.mapTemporaryTable(tempTableResource());
         
         // Register for temp table cleanup.
         Multiplexer multiplexer = new Multiplexer();
         if (global)
         {
            // A global temp table buffer is cleaned up when the context ends.
            TransactionManager.registerFinalizable(multiplexer, true);
            
            if (isDynamic())
            {
               dynamicMultiplexer = multiplexer;
            }
         }
         else
         {
            // A non-global temp table buffer is cleaned up when the top
            // level method scope ends (this corresponds with the external
            // procedure in Progress).
            TransactionManager.registerTopLevelFinalizable(multiplexer, true);
         }
         
         // register for no-undo validation if outside an application transaction
         if (!isUndoable() && !TransactionManager.isTransaction())
         {
            TransactionManager.registerFinalizable(new NoUndoValidator(), false);
         }
      }
   }
   
   /**
    * Open a new buffer scope for this record buffer at the given block depth. Open a new
    * multiplex scope and assign a multiplex ID the first time we are invoked for a buffer
    * instance.
    *
    * @param   blockDepth
    *          Zero-based depth of the block (starting from the outermost scope) at which the
    *          buffer scope is opened.
    */
   protected void openScopeAt(int blockDepth)
   {
      super.openScopeAt(blockDepth);
      
      initialize();
      
      // if this is the first time this table is being used, create it in the database
      if (openMultiplexScope())
      {
         context.get().createTable(getDMOInterface(), getTable());
      }
      
      // map legacy info (this is tolerant to being called more than once)
      TableMapper.mapTemporaryTable(tempTableResource());
      
      // register for temp table cleanup
      Multiplexer multiplexer = new Multiplexer();
      TransactionManager.registerFinalizableAt(blockDepth, multiplexer);
      if (isDynamic() && blockDepth == 0)
      {
         dynamicMultiplexer = multiplexer;
      }
      
      // register for no-undo validation if outside an application transaction
      // TODO: doing this global scope creates a memory leak; does it even make sense to register
      // these at a scope other than the current scope?
      if (blockDepth > 0 && !isUndoable() && !TransactionManager.isTransaction())
      {
         TransactionManager.registerFinalizableAt(blockDepth, new NoUndoValidator());
      }
   }
   
   /**
    * Deregister a dynamic multiplexer object from the transaction manager's global finalizables
    * when this buffer goes out of scope.
    */
   @Override
   protected void resourceDeleted()
   {
      if (dynamicMultiplexer != null)
      {
         TransactionManager.deregisterGlobalFinalizable(dynamicMultiplexer);
      }
   }
   
   /**
    * Initialize the validation helper object.
    * 
    * @return  Validation helper object.
    */
   protected ValidationHelper initValidationHelper()
   {
      return (new TempValidationHelper());
   }
   
   /**
    * Override the superclass method to do nothing.  Locking is a no-op for
    * temp table records, since each user has exclusive access to their temp
    * table instance.
    */
   protected void initPinnedLockTypes()
   {
   }
   
   /**
    * Perform a bulk delete of all records in the table associated with this buffer.
    * 
    * @throws  ErrorConditionException
    *          if an error occurs performing the bulk delete operation.
    */
   protected void deleteAll()
   {
      try
      {
         removeRecords(null, null, false);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
   }
   
   /**
    * Get the order by clause determined by this buffer's backing table's primary index, if any.
    * If no primary index is defined for the backing table, the order by clause will specify an
    * ascending sort by primary key.
    * 
    * @return  Order by clause for temp-table's primary index.
    */
   private String getPrimaryOrderByClause()
   {
      if (primaryOrderBy != null)
      {
         return primaryOrderBy;
      }
      
      String entity = getEntityName();
      P2JIndex index = null;
      
      try
      {
         index = getIndexHelper().getPrimaryIndex(entity);
      }
      catch (PersistenceException exc)
      {
         // no primary index found, use default
         primaryOrderBy = " order by id asc";
         
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, "Setting default order by clause ' order by id asc'", exc);
         }
      }
      
      if (primaryOrderBy == null)
      {
         StringBuilder buf = new StringBuilder(" order by ");
         int count = 0;
         Iterator<P2JIndexComponent> iter = index.components();
         while (iter.hasNext())
         {
            if (count > 0)
            {
               buf.append(", ");
            }
            count++;
            P2JIndexComponent comp = iter.next();
            String name = comp.getName();
            String direction = comp.isDescending() ? "desc" : "asc";
            buf.append(name);
            buf.append(' ');
            buf.append(direction);
         }
         
         primaryOrderBy = buf.toString();
      }
      
      return primaryOrderBy;
   }
   
   /**
    * Create and set the current bulk delete reversible action.
    * This variant produces a <code>ReversibleBulkDelete</code> instance.
    * <p>
    * If this block is not undoable, this method is a no-op.
    * 
    * @param   mpid
    *          The multiplex ID used to mark those records as deleted.
    * @param   hasWhere
    *          Flag indicating if this bulk delete had a where clause or not.
    */
   private void setReversibleBulkDelete(int mpid, boolean hasWhere)
   {
      // If we are in a transaction, remember this record change so it can be
      // undone at rollback if necessary.
      if (TransactionManager.isTransaction())
      {
         ReversibleBulkDelete item = new ReversibleBulkDelete(mpid, hasWhere);
         
         bufferManager.addReversible(this, null, item, false);
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, describe() + ":  adding reversible [" + item + "]");
         }
      }
   }
   
   /**
    * Open a new multiplex scope for this buffer.  The return value indicates
    * whether it is the first time a multiplex scope has been opened for this
    * buffer's DMO interface in this context.  If so, the caller will need to
    * create the backing temp table in the database.  This method permanently
    * assigns the multiplex ID to this buffer instance.
    *
    * @return  <code>true</code> if this is the first multiplex scope opened
    *          for the underlying DMO type;  <code>false</code> if multiplex
    *          scopes are already open in this context for this DMO type.
    */
   private boolean openMultiplexScope()
   {
      if (!DynamicTablesHelper.isDynamicTableBuffer(this))
      {
         // buffers can be opened in an arbitrary order, so the first buffer opened for a table
         // registers static temp-table object and sets tempTableRef in the table default buffer
         if (master == null)
         {
            if (tempTableRef == null)
            {
               tempTableRef = bufferManager.registerStaticTempTable((Temporary) getDMOProxy(),
                                                                    global);
            }
         }
         else
         {
            if (master.tempTableRef != null)
            {
               tempTableRef = master.tempTableRef;
            }
            else
            {
               tempTableRef = bufferManager.registerStaticTempTable((Temporary) getDMOProxy(),
                                                                    global);
               master.tempTableRef = tempTableRef;
            }
         }
      }
      
      Class<?> dmoIface = getDMOInterface();
      Context local = context.get();
      Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
      
      // Get the map of multiplex IDs which are in use for this DMO type
      // within our context.  The value for each multiplex ID key is the
      // reference count of buffers using that ID.
      Map<Integer, Integer> inUse = map.get(dmoIface);
      
      if (inUse == null)
      {
         // This is the first use of this DMO type for this context;  create a new in-use map.
         inUse = new HashMap<>();
         map.put(dmoIface, inUse);
      }
      
      int refCount = 0;
      
      if (multiplexID == null)
      {
         // Generate a unique multiplex ID.
         Integer id = local.getNextMPID(inUse);
         
         // Assign the unique ID as the permanent multiplex ID of every
         // buffer in this buffer's domain.
         RecordBuffer key = getMasterBuffer();
         Map<TemporaryBuffer, Object> domain = local.domains.get(key);
         for (TemporaryBuffer next : domain.keySet())
         {
            next.multiplexID = id;
         }
         
         // Start off the reference count.
         refCount = 1;
      }
      else
      {
         // Increment the existing reference count for this multiplex ID.
         Integer rcObj = inUse.get(multiplexID);
         refCount = (rcObj == null ? 1 : rcObj + 1);
      }
      
      boolean createTable = inUse.isEmpty();
      
      inUse.put(multiplexID, refCount);
      
      // If this is the first use of this multiplex ID, the virtual table must be empty.
      if (refCount == 1)
      {
         local.emptyTableFlags.put(multiplexID, Boolean.TRUE);
      }
      
      return createTable;
   }
   
   /**
    * Perform a bulk delete, restricted by the given where clause and query
    * substitution parameters.
    * <p>
    * The targeted records are not actually removed from the buffer; instead,
    * they are assigned another multiplex ID, which marks them as deleted.
    * <p>
    * Note:  primary keys surrendered via bulk deletes are not currently recycled.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.
    * @param   args
    *          Query substitution parameters required by the where clause.
    * 
    * @throws  PersistenceException
    *          if an error occurs performing the bulk delete operation.
    */
   protected void delete(String where, Object[] args)
   throws PersistenceException
   {
      removeRecords(where, args, false);
   }
   
   /**
    * Delete the current record from the database.  Set the current record to
    * null if the deletion is successful.  Reclaim the record's primary key
    * for later re-use.
    *
    * @throws  PersistenceException
    *          if no record currently is loaded in the buffer, or if there is
    *          an error deleting the current record from the database.
    */
   protected void delete()
   throws PersistenceException
   {
      boolean autoCommit = !TransactionManager.isTransaction();
      
      Persistence persistence = null;
      Serializable id = null;
      if (autoCommit)
      {
         persistence = getPersistence();
         persistence.beginTransaction();
         id = getCurrentRecord().getId();
      }
      
      super.delete();
      
      if (autoCommit)
      {
         Context local = context.get();
         local.reclaimKey(getDMOInterface(), id);
         persistence.commit();
         local.commit(true);
      }
   }
   
   /**
    * Reclaim the given list of primary keys for re-use.
    * 
    * @param   keys
    *          A list of primary keys which are available for re-use.  Must
    *          not be <code>null</code>.
    */
   @Override
   protected void reclaimKeys(List<Serializable> keys)
   {
      Context local = context.get();
      Class<?> dmoIface = getDMOInterface();
      for (Serializable key : keys)
      {
         local.reclaimKey(dmoIface, key);
      }
   }
   
   /**
    * Returns the primary key for a new DMO associated with this buffer.
    *
    * @return  The primary key for a new DMO associated with this buffer.
    *
    * @throws  PersistenceException
    *          if there was an error while retrieving the next primary key.
    */
   @Override
   protected Serializable nextPrimaryKey()
   throws PersistenceException
   {
      Context local = context.get();
      
      return local.nextPrimaryKey(getDMOInterface());
   }
   
   /**
    * Close the current multiplex scope for the underlying DMO type in the
    * current context.  Remove all records associated with this buffer's
    * multiplex ID and adjust the multiplex ID in use map.  If this is the
    * last scope to be closed for this DMO type, request the context to drop
    * the backing temp table from the database when the database session is
    * closed.
    * 
    * @throws  PersistenceException
    *          if any error occurs within a listener while processing a
    *          change event.
    */
   private void doCloseMultiplexScope()
   throws PersistenceException
   {
      // Do we need to copy our current records before cleaning up?
      performOutputCopy();
      
      Class<?> dmoIface = getDMOInterface();
      Context local = context.get();
      Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
      
      // Get the map of multiplex IDs which are in use for this DMO type within our context.
      Map<Integer, Integer> inUse = map.get(dmoIface);
      
      // Access our reference count from this map.
      int refCount = inUse.get(multiplexID);
      
      // Decrement the reference count.  If 0, remove the key, else store the new value back in
      // the map.
      if (--refCount == 0)
      {
         inUse.remove(multiplexID);
      }
      else
      {
         inUse.put(multiplexID, refCount);
      }
      
      // If the in-use map is now empty, the backing table will need to be dropped.  Also remove
      // the inUse map from the master map and remove this buffer's domain if it is the master
      // buffer.
      if (inUse.isEmpty())
      {
         map.remove(dmoIface);
         
         try
         {
            removeRecords(null, null, true);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         finally
         {
            local.emptyTableFlags.remove(multiplexID);
            
            // Remove any maps of reversible/undoables pinned by this buffer.
            // The virtual table associated with these is gone.
            bufferManager.removeAllReversibles(this);
            
            // Drop the table.
            local.dropTable(tempTableResource(), getTable());
         }
      }
      // Otherwise, if the current multiplexID is no longer in use, remove
      // all records for this multiplex ID from the table, but leave the
      // table intact for the other buffers still using it.
      else if (refCount == 0)
      {
         try
         {
            removeRecords(null, null, true);
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
         finally
         {
            local.emptyTableFlags.remove(multiplexID);
            
            // Remove any maps of reversible/undoables pinned by this buffer.
            // The virtual table associated with these is gone.
            bufferManager.removeAllReversibles(this);
         }
      }
   }
   
   /**
    * Determine if the multiplex scope for the underlying DMO type in the current context is
    * closed. See {@link #closeMultiplexScope(boolean)} for more information.
    *
    * @return <code>true</code> if multiplex scope is closed.
    */
   private boolean isMultiplexScopeClosed()
   {
      Context local = context.get();
      Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
      
      // Get the map of multiplex IDs which are in use for this DMO type within our context.
      Map<Integer, Integer> inUse = map.get(getDMOInterface());
      
      if (inUse == null)
      {
         return true;
      }
      
      // Access our reference count from this map.
      Integer rc = inUse.get(multiplexID);
      
      return rc == null;
   }
   
   /*
   private void count()
   {
      Persistence persistence = getPersistence();
      StringBuilder buf = new StringBuilder("select count(*) from ");
      buf.append(getDMOImplementationName());
      String hql = buf.toString();
      try
      {
         List list = persistence.list(hql, null, null, 0, 0, true);
         Number count = (Number) list.get(0);
         LOG.info(getTable() + " contains " + count + " records.");
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(-1,
                                         "Error counting temp table records",
                                         exc);
      }
   }
   */
   
   /**
    * Remove or change the multiplex ID for all record rows from the backing 
    * table which meet the specified criteria.  Because temp table instances
    * are multiplexed across nested buffer scopes and across multiple buffers
    * which use the same backing DMO type, the backing table cannot be dropped
    * until after the last buffer scope referencing the backing table in the
    * current context has been closed.  This method is invoked for all buffer
    * scope closing events.  This is also the worker method for the {@link
    * #deleteAll()} and {@link #delete(String, Object[])} methods.
    * 
    * @param   where
    *          An HQL where clause snippet which defines the restriction
    *          criteria to apply to the delete.  All references to properties
    *          in a DMO must be unqualified.  <code>null</code> to delete all
    *          records whose multiplex IDs match that of this buffer.
    * @param   args
    *          Query substitution parameters required by the where clause.
    *          <code>null</code> if there are no substitutions or no where clause.
    * @param   forceDelete
    *          Force the records to be deleted from the table, not just
    *          marking them with a negative multiplex ID.
    * 
    * @throws  PersistenceException
    *          if an error occurs performing the bulk delete operation.
    */
   private void removeRecords(String where, Object[] args, boolean forceDelete)
   throws PersistenceException
   {
      // after executing this method, the buffer is empty regardless the where predicate matches
      // the currently stored record
      release();
      
      // if we are not within a transaction, or this temp-table is not undoable
      forceDelete = forceDelete                          ||
                    !TransactionManager.isTransaction()  ||
                    !isUndoable();
      
      boolean hasWhere = (where != null);
      Persistence persistence = getPersistence();
      
      StringBuilder buf = new StringBuilder();
      Integer delMPID = 0;
      
      // Create portion of where clause which restricts by multiplex ID.
      buf.append(MULTIPLEX_FIELD_NAME);
      buf.append(" = ");
      buf.append(multiplexID);
      String mpidWhere = buf.toString();
      
      HQLExpression query = new HQLExpression();
      
      if (forceDelete) 
      {
         query.append("delete from ");
      }
      else
      {
         query.append("update ");
         Context local = context.get();
         Map<Class<?>, Map<Integer, Integer>> map = local.inUseMPIDs;
         Map<Integer, Integer> inUse = map.get(getDMOInterface());
         
         delMPID = local.getNextMPID(inUse);
      }
      
      String dmoImplName = getDMOImplementationName();
      query.append(dmoImplName);
      
      if (!forceDelete)
      {
         // do not delete, just change the multiplex ID
         query.append(" set ");
         query.append(MULTIPLEX_FIELD_NAME);
         query.append(" = ");
         query.append(Integer.toString(delMPID));
      }
      
      query.append(" where ");
      
      // Only prepend a test for multiplex ID if temp table is multiplexed or
      // if we have a where clause.  Otherwise, it is more efficient to omit
      // this test.
      if (hasWhere)
      {
         query.append("(");
      }
      query.append(mpidWhere);
      if (hasWhere)
      {
         query.append(") and (");
      }
      
      Type[] types = null;
      if (hasWhere)
      {
         query.append(HQLPreprocessor.get(where, this, args, false).getHQL());
         query.append(")");
         if (args != null)
         {
            types = DBUtils.makeTypeArray(args);
         }
      }
      String hql = query.toFinalExpression();
      
      int rowCount = persistence.deleteOrUpdate(hql, args, types);
      if (rowCount > 0)
      {
         changeBroker.forcePendingFlush(this);
         
         if (!forceDelete)
         {
            // if the records had their multiplex ID changed, register a
            // ReversibleBulkDelete for this where clause and arguments
            setReversibleBulkDelete(delMPID, hasWhere);
         }
         
         if (!isUndoable() && TransactionManager.isTransaction())
         {
            // force the records to be deleted, on db tx rollback
            ReversibleBulkDelete rev = new ReversibleBulkDelete(where, args, types, multiplexID);
            rev.setForceDelete(true);
            bufferManager.addReversible(this, null, rev, false);
         }
         
         boolean full = !hasWhere;
         
         if (!full)
         {
            // Execute a simple query to determine if there is any record in
            // this virtual table.
            buf.setLength(0);
            buf.append("select tt.id from ");
            buf.append(dmoImplName);
            buf.append(" tt where ");
            buf.append(mpidWhere);
            hql = buf.toString();
            
            List<Object> list = persistence.list(hql, null, null, 1, 0, true);
            full = (list == null);
         }
         
         if (full)
         {
            Context local = context.get();
            local.emptyTableFlags.put(multiplexID, Boolean.TRUE);
         }
         
         // Notify all interested parties that a bulk delete occurred.
         // Listening buffers may need to release deleted records.
         RecordChangeEvent.Type eventType =
            (full
             ? RecordChangeEvent.Type.FULL_BULK_DELETE
             : RecordChangeEvent.Type.RESTRICTED_BULK_DELETE);
         changeBroker.stateChanged(eventType, this, null, null, null);
      }
   }
   
   /**
    * Get the {@link TempTable} resource associated with this temporary buffer.
    * 
    * @return   See above.
    */
   private TempTable tempTableResource()
   {
      return tempTableRef;
   }
   
   /**
    * Concrete implementation of <code>Reversible</code> which on rollback
    * reverses a bulk record delete action by setting the multiplex ID for 
    * the targeted DMO's back to its original value.
    */
   private final class ReversibleBulkDelete
   extends AbstractBulkReversible
   {      
      /** The multiplex ID used to mark the records as deleted. */
      private final int delMPID;
      
      /** 
       * If not-null, it represents the record which we want to rollback or
       * delete, only if it satisfies the where clause.
       */
      private final Serializable key;
      
      /** 
        * The where clause representing the records to be deleted. 
        * May be null.
        */
      private String where;
      
      /** The arguments for the where clause. May be null.*/
      private Object[] args;
      
      /** The associated type for each argument. May be null.*/
      private Type[] types;
      
      /** 
        * When <code>true</code>, this reversible will force the deletion of
        * the records, instead of updating the multiplex ID.
        */
      private boolean forceDelete = false;
      
      /**
       * Flag indicating if this bulk delete had a where clause or not.
       */
      private boolean hasWhere = false;
      
      /**
       * Constructor.
       *
       * @param   delMPID
       *          The multiplex ID value used to mark those records as deleted
       * @param   hasWhere
       *          Flag indicating if this bulk delete had a where clause.
       */
      ReversibleBulkDelete(int delMPID, boolean hasWhere)
      {
         this(delMPID, null);
         this.hasWhere = hasWhere;
      }
      
      /**
       * This constructor will limit the bulk delete to the where clause, if 
       * if specified. The c'tor will be used for re-roll forward actions, 
       * after a database tx rollback, for no-undo temp-tables. In this case,
       * the <code>delMPID</code> will specify the actual multiplex ID for
       * the records which should be deleted.
       *
       * @param   where  
       *          The conditions which targets the deleted records. May be 
       *          null.
       * @param   args
       *          The needed arguments for the <code>where</code> clause. May 
       *          be null.
       * @param   types
       *          The associated type for each argument. May be null.
       * @param   delMPID
       *          The multiplex ID value used to mark those records as deleted 
       */
      ReversibleBulkDelete(String where, Object[] args, Type[] types, int delMPID)
      {
         this(delMPID, null);
         this.where = where;
         this.args = args;
         this.types = types;
      }
      
      /**
       * Constructor to rollback a record which may deleted during a bulk
       * action.
       *
       * @param   delMPID
       *          the multiplex ID value used to mark those records as deleted
       * @param   key
       *          The key for the record which we want to delete or rollback.
       */
      ReversibleBulkDelete(int delMPID, Serializable key)
      {
         this.delMPID = delMPID;
         this.key = key;
      }
      
      /**
       * Get a string representation of the internal state of this object,
       * primarily for debugging purposes.
       *
       * @return  String representing this object's state.
       */
      public String toString()
      {
         return "Bulk deletion" + " of " + getDMOName() + " [MPID " + delMPID + "]";
      }
      
      /**
       * This method checks if this bulk reversible has deleted the record 
       * withthe given key.
       * 
       * @param   id
       *          Primary key of the record to be checked.
       *          
       * @return  <code>true</code> if the record was deleted by this bulk
       *          reversible.
       */
      protected boolean containsRecord(Serializable id)
      {
         if (!hasWhere)
         {
            // no where clause, the bulk delete affects the entire virtual
            // table
            return true;
         }
         
         Persistence persistence = getPersistence();
         if (bulkDeleteCheckHQL == null)
         {
            String entity = getEntityName();
            HQLExpression buf = new HQLExpression();
            buf.append("select tt.id from ");
            buf.append(entity);
            buf.append(" tt where tt.id = ", true);
            buf.append(" and tt.");
            buf.append(MULTIPLEX_FIELD_NAME);
            buf.append(" = ", true);
            bulkDeleteCheckHQL = buf.toFinalExpression();
         }
         
         try
         {
            List list = persistence.list(bulkDeleteCheckHQL,
                                         new Object[] { id, delMPID },
                                         BULK_DELETE_CHECK_QUERY_PARM_TYPES,
                                         1,
                                         0,
                                         true);
            
            return (list != null);
         }
         catch (PersistenceException exc)
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Error checking for record related to bulk delete", exc);
            }
         }
         
         return false;
      }
      
      /**
       * Indicate whether this <code>Reversible</code> implementation stores its
       * DMO locally, as opposed to storing an identifier and having to retrieve
       * the DMO from the backing database, or not making a DMO available at all.
       * 
       * @return  <code>true</code> if the DMO is stored locally, else
       *          <code>false</code>.
       */
      protected boolean storesDMO()
      {
         return false;
      }
      
      /**
       * Get the unqualified interface name of the DMO being managed by this
       * reversible.
       * 
       * @return  DMO interface name.
       */
      protected String getDMOName()
      {
         return TemporaryBuffer.this.getDMOName();
      }
      
      /**
       * Set the <code>forceDelete</code> flag: when <code>true</code>, the 
       * records which satisfy the <code>where</code> clause (if any) will be deleted.
       * 
       * @param   forceDelete
       *          The new forceDelete flag as explained above.
       */
      void setForceDelete(boolean forceDelete)
      {
         this.forceDelete = forceDelete;
      }
      
      /** 
       * No-op.
       * 
       * @return  Nothing.
       *
       * @throws  UnsupportedOperationException  
       *          No DMO is required for a bulk delete. 
       */
      protected Serializable getId()
      {
         throw new UnsupportedOperationException(
            "ReversibleBulkDelete does not require a DMO for persistable " + getDMOName());
      }
      
      /**
       * Duplicate this reversible action. 
       * 
       * @param   key
       *          The key to set for the new reversible action.
       */
      Reversible duplicate(Serializable key)
      {
         ReversibleBulkDelete rev = new ReversibleBulkDelete(this.delMPID, key);
         rev.forceDelete = this.forceDelete;
         // copy the where clause only if this reversible will force the
         // delete and not just restore the multiplex ID. in the case of a
         // forced delete, the multiplex ID wasn't changed, so the where
         // clause is needed to ensure that the record with the specified key
         // can be deleted by bulk delete.
         if (forceDelete)
         {
            rev.where = this.where;
            rev.types = this.types;
            rev.args  = this.args;
         }
         
         return rev;
      }
      
      /** 
       * No-op.
       * 
       * @return  Nothing.
       * 
       * @throws  UnsupportedOperationException  
       *          always;  no DMO is required for a bulk delete. 
       */
      protected Persistable getDMO()
      {
         throw new UnsupportedOperationException(
            "ReversibleBulkDelete does not require a DMO for persistable " + getDMOName());
      }
      
      /**
       * Set the underlying DMO for this reversible action.  This
       * implementation is a no-op.
       * 
       * @param   dmo
       *          The underlying DMO.
       */
      protected void setDMO(Persistable dmo)
      {
      }
      
      /**
       * Reverse the bulk deletion of the records which meet the specified 
       * criteria, by setting back their multiplex ID to a positive value.
       * <p>
       * If <code>forceDelete</code> is set to <code>true</code>, this method
       * will actually force the deletion of the records which satisfy the
       * <code>where</code> clause (if any) and have the specified multiplex
       * ID.
       *
       * @param   rollForward
       *          If <code>true</code>, this method is being invoked in order
       *          to re-roll forward a non-undoable change, after a
       *          database-level rollback.  This indicates the modification is
       *          made within a new transaction invoked after the
       *          database-level rollback.
       *          If <code>false</code>, this method is being invoked in order
       *          to roll back an undoable change within the context of an
       *          application-level, full transaction or sub-transaction
       *          rollback.
       * @param   transaction
       *          <code>true</code> if this rollback occurs at a full
       *          transaction boundary, else <code>false</code>.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when updating the records.
       */
      protected void rollback(boolean rollForward, boolean transaction)
      throws PersistenceException
      {
         if (forceDelete)
         {
            forceDelete();
            
            return;
         }
         
         StringBuilder buf = new StringBuilder("update ");
         buf.append(getDMOImplementationName());
         buf.append(" set ");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ");
         buf.append(multiplexID.intValue());
         
         executeQuery(buf);
      }
      
      /**
       * Execute an update or delete statement.
       * 
       * @param   buf
       *          The string used to construct the statement.
       *          
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when updating or deleting the records.
       */
      private void executeQuery(StringBuilder buf)
      throws PersistenceException
      {
         Persistence persistence = getPersistence();
         String hql = null;
         
         buf.append(" where (");
         buf.append(MULTIPLEX_FIELD_NAME);
         buf.append(" = ");
         buf.append(delMPID);
         buf.append(")");
         
         if (key != null)
         {
            buf.append(" and (id = ");
            buf.append(key);
            buf.append(")");
         }
         
         if (where != null)
         {
            buf.append(" and (");
            buf.append(where);
            buf.append(")");
         }
         
         hql = buf.toString();
         
         if (persistence.deleteOrUpdate(hql, args, types) > 0)
         {
            changeBroker.forcePendingFlush(TemporaryBuffer.this);
            
            // we need to set the empty table flag back to OFF, as this bulk
            // delete was rolled back and table is no longer empty
            Context local = context.get();
            local.emptyTableFlags.put(multiplexID, Boolean.FALSE);
         }         
      }
      
      /**
       * Remove the marked records from the table.
       *
       * @throws  PersistenceException
       *          if an error occurs at the database or in the persistence
       *          layer when deleting the records.
       */
      private void forceDelete() 
      throws PersistenceException
      {
         StringBuilder buf = new StringBuilder("delete from ");
         buf.append(getDMOImplementationName());
         executeQuery(buf);
      }
   }
   
   /**
    * Helper which tracks the multiplex IDs currently in use by DMO interface
    * in the current client context, as well as the reference count of temp
    * tables created in the database using the context's current connection,
    * and domains of master/slave buffers.
    * <p>
    * This object also tracks tables which have been created during the
    * current database session, and drops them when the session is being
    * closed.  This is done as an optimization to prevent potential thrashing
    *in the event a class which creates a temporary buffer and then allows it
    * to go out of scope is invoked repeatedly in a loop.
    */
   private static class Context
   implements SessionListener,
              Commitable
   {
      /** Map of DMO interfaces to counts of multiplex IDs currently in use */
      private final Map<Class<?>, Map<Integer, Integer>> inUseMPIDs = new HashMap<>();
      
      /** Map of open tables to most recent primary key ID generated */
      private final Map<Class<?>, Long> openTables = new HashMap<>();
      
      /** Primary keys which will be recycled after the next commit */
      private final Map<Class<?>, List<Serializable>> pendingReclaimedKeys = new HashMap<>();
      
      /** Recycled primary keys available for re-use (lowest first) */
      private final Map<Class<?>, SortedSet<Serializable>> reclaimedKeys = new HashMap<>();
      
      /** Names of active temp tables in the order they were created */
      private final Set<String> activeTables = new LinkedHashSet<>();
      
      /** Map of flags indicating whether each virtual table is empty */
      private final Map<Integer, Boolean> emptyTableFlags = new HashMap<>();
      
      /** Singleton map of a unique key to session ID string */
      private final Map<String, String> sweepMap;
      
      /** Persistence service object for temp tables */
      private Persistence persistence;
      
      /** Map of master buffers to weak hash maps with slave buffer keys */
      private Map<TemporaryBuffer, Map<TemporaryBuffer, Object>> domains = new WeakHashMap<>();
      
      /** Next value to try when assigning a multiplex ID */
      private int nextMPID = 1;
      
      /**
       * Default constructor.
       */
      Context()
      {
         SecurityManager secMgr = SecurityManager.getInstance();
         String sid = String.valueOf(secMgr.getSessionId());
         sweepMap = Collections.singletonMap(TempTableHelper.UNIQUE_SUFFIX_KEY, sid);
      }
      
      /**
       * Invoked when a Hibernate session is about to close, but is still
       * available for cleanup processing.  This implementation drops all
       * temporary tables which have been created during the current database
       * session if the session is not being closed due to an error.
       * Otherwise, it just sets the temp table reference count to zero, so
       * that the connection provider is allowed to release the connection.
       *
       * @param   event
       *          Session event type.
       *
       * @return  <code>false</code> indicating that this session listener
       *          should remain registered after this method returns.
       *
       * @throws  PersistenceException
       *          if any error occurs dropping temp tables from the closing
       *          session.
       */
      public boolean sessionEvent(SessionListener.Event event)
      throws PersistenceException
      {
         switch (event)
         {
            case SESSION_CLOSING_NORMALLY:
               dropAllTables();
               break;
               
            case SESSION_CLOSING_WITH_ERROR:
               
               // On error, we want TempTableConnectionProvider to release the
               // broken connection, so make sure we are not pinning it open.
               activeTables.clear();
               
               // Clear remainder of state.
               Iterator<Class<?>> iter = openTables.keySet().iterator();
               while (iter.hasNext())
               {
                  Class<?> next = iter.next();
                  inUseMPIDs.remove(next);
                  iter.remove();
               }
               
               break;
               
            default:
               return false;
         }
         
         // Only invoked in SESSION_CLOSING_* cases, due to return in default
         // case in switch above.
         pendingReclaimedKeys.clear();
         reclaimedKeys.clear();
         
         return false;
      }
      
      /**
       * Invoked when this session listener is removed from the list of
       * registered session listeners.
       * <p>
       * This implementation does nothing.
       */
      public void deregisteredSessionListener()
      {
      }
      
      /**
       * On a full transaction commit, reclaim all of the primary keys which
       * have been surrendered by record deletes and clear the set of pending,
       * reclaimable keys.
       * 
       * @param   transaction
       *          <code>true</code> if called on a full transaction boundary;
       *          <code>false</code> if called on a subtransaction boundary.
       */
      public void commit(boolean transaction)
      {
         if (!transaction)
         {
            return;
         }
         
         for (Map.Entry<Class<?>, List<Serializable>> next : pendingReclaimedKeys.entrySet())
         {
            Class<?> dmoIface = next.getKey();
            List<Serializable> newIDs = next.getValue();
            SortedSet<Serializable> allIDs = reclaimedKeys.get(dmoIface);
            if (allIDs == null)
            {
               allIDs = new TreeSet<>();
               reclaimedKeys.put(dmoIface, allIDs);
            }
            allIDs.addAll(newIDs);
         }
         
         pendingReclaimedKeys.clear();
      }
      
      /**
       * On a full transaction rollback,clear the set of pending, reclaimable
       * keys.
       * 
       * @param   transaction
       *          <code>true</code> if called on a full transaction boundary;
       *          <code>false</code> if called on a subtransaction boundary.
       */
      public void rollback(boolean transaction)
      {
         if (transaction)
         {
            pendingReclaimedKeys.clear();
         }
      }
      
      /**
       * No-op.
       * 
       * @see com.goldencode.p2j.util.Commitable#rollbackPending()
       */
      public void rollbackPending()
      {
      }
      
      /**
       * No-op
       * 
       * @see com.goldencode.p2j.util.Commitable#validate(boolean)
       */
      public void validate(boolean transaction)
      {
      }
      
      /**
       * Initialize this object by storing the persistence service object and
       * registering as a session listener.  This method is tolerant of being
       * invoked multiple times, but only does its initialization work the
       * first time it is invoked for the current <code>Context</code>
       * instance.
       */
      void init()
      {
         if (persistence != null)
         {
            return;
         }
         
         persistence = PersistenceFactory.getInstance(DatabaseManager.TEMP_TABLE_DB);
         persistence.registerSessionListener(this, SessionListener.Scope.NONE);
      }
      
      /**
       * Report the number of temp tables which currently are in use in the
       * current context for the database associated with this buffer.
       *
       * @return  Number of temp tables associated with
       *          <code>persistence</code>.
       */
      int getTableCount()
      {
         return activeTables.size();
      }
      
      /**
       * Returns a context-local multiplex ID value.
       * 
       * @param   inUse
       *          Map of multiplex IDs currently in use in this context, to
       *          reference count of buffers using them.
       * 
       * @return  The next available multiplex ID.
       */
      Integer getNextMPID(Map<Integer, Integer> inUse)
      {
         Integer id = null;
         do
         {
            id = nextMPID++;
         }
         while (inUse.containsKey(id));
         
         return id;
      }
      
      /**
       * Create a temp table in the database for the given DMO interface, if
       * it does not already exist.  This will also create all indexes defined
       * for the temp table.
       * 
       * @param   dmoIface
       *          DMO interface for which a backing temp table is needed.
       * @param   table
       *          Name of temp table to be created.
       */
      void createTable(Class<?> dmoIface, String table)
      {
         init();
         
         activeTables.add(table);
         if (!openTables.containsKey(dmoIface))
         {
            openTables.put(dmoIface, 0L);
            doCreateTable(dmoIface);
         }
      }
      
      /**
       * Decrement the reference count of active temp tables in use within
       * the current context.  This method does not actually drop a table,
       * since all table drops are deferred until the database session
       * actually closes.  This prevents potential thrashing in the event
       * a class which creates a temporary buffer and then allows it to go
       * out of scope is invoked repeatedly in a loop.
       * 
       * @param   tt
       *          The temp-table resource.
       * @param   table
       *          Name of temp table to be dropped.
       */
      void dropTable(TempTable tt, String table)
      {
         activeTables.remove(table);
         
         TableMapper.removeTemporaryTable(tt);
      }
      
      /**
       * Returns the next available primary key for the temp table associated
       * with the given DMO interface.
       *
       * @param   dmoIface
       *          Data model object interface.
       *
       * @return  The primary key for a new DMO associated with this buffer.
       * 
       * @throws  PersistenceException
       *          if <code>dmoIface</code> is unknown.
       */
      Serializable nextPrimaryKey(Class<?> dmoIface)
      throws PersistenceException
      {
         Long id = openTables.get(dmoIface);
         if (id == null)
         {
            throw new PersistenceException(
               "No temp table registered for DMO " + dmoIface.getName());
         }
         
         // If we have a recycled key, use it first.  Otherwise, allocate a
         // new key.
         SortedSet<Serializable> keys = reclaimedKeys.get(dmoIface);
         if (keys != null)
         {
            id = (Long) keys.first();
            keys.remove(id);
            if (keys.isEmpty())
            {
               reclaimedKeys.remove(dmoIface);
            }
         }
         else
         {
            id++;
            openTables.put(dmoIface, id);
         }
         
         return id;
      }
      
      /**
       * Prepare a single, primary key for reclamation.  On the next, full
       * transaction commit, this key will be made available for re-use.
       * Until then, it will be stored in a set of keys which are pending
       * reclamation.  This set will be cleared if the current transaction is
       * rolled back.
       * 
       * @param   dmoIface
       *          DMO interface for temporary table records.
       * @param   id
       *          Primary key to be recycled.
       */
      void reclaimKey(Class<?> dmoIface, Serializable id)
      {
         List<Serializable> keys = pendingReclaimedKeys.get(dmoIface);
         if (keys == null)
         {
            keys = new ArrayList<>();
            pendingReclaimedKeys.put(dmoIface, keys);
         }
         keys.add(id);
      }
      
      /**
       * Drop all temporary tables which have been created during the current
       * database session.  This typically occurs when the session is being
       * closed.
       *
       * @throws  PersistenceException
       *          if a database error occurs dropping any table or creating
       *          the transaction which contains these actions.
       */
      private void dropAllTables()
      throws PersistenceException
      {
         try
         {
            Iterator<Class<?>> iter = openTables.keySet().iterator();
            if (!iter.hasNext())
            {
               return;
            }
            
            // we should only try to drop tables if there is an open transaction; the transaction
            // may already have been rolled back, in which case the tables already are lost; if
            // we try to execute SQL to drop tables and indexes in this case, we will cause JDBC
            // errors
            if (persistence.isTransactionOpen())
            {
               while (iter.hasNext())
               {
                  Class<?> dmoIface = iter.next();
                  doDropTable(dmoIface);
                  iter.remove();
               }
            }
         }
         finally
         {
            TableMapper.removeDynamicTempTables();
            openTables.clear();
         }
      }
      
      /**
       * Issue the SQL statement(s) against the backing database necessary to
       * create the backing temp table needed for the given DMO interface, as
       * well as all indexes defined for that table.  Because temp table
       * instances are multiplexed across nested buffer scopes and across
       * multiple buffers which use the same backing DMO type, this method is
       * invoked only the first time a backing table is actually needed for a
       * particular DMO interface type in the current context.  Thereafter,
       * this method is not invoked again, unless the backing table has first
       * been dropped.
       *
       * @param   dmoIface
       *          DMO interface for which a backing temp table is needed.
       *
       * @throws  ErrorConditionException
       *          if an error occurs creating the table and silent error mode
       *          is disabled.
       *
       * @see     #doDropTable
       */
      private void doCreateTable(Class<?> dmoIface)
      {
         String sql = null;
         
         try
         {
            boolean shouldCommit = persistence.beginTransaction();
            Database database = persistence.getDatabase();
            
            // Create table(s).
            Iterator<String> tableIter = TempTableHelper.sqlTempTableCreate(database, dmoIface);
            while (tableIter.hasNext())
            {
               sql = tableIter.next();
               persistence.executeSQL(sql);
            }
            
            // Create indexes (if any).
            Iterator<String> indexIter =
               TempTableHelper.sqlTempTableIndexCreate(database, dmoIface);
            while (indexIter.hasNext())
            {
               sql = indexIter.next();
               sql = StringHelper.sweep(sql, sweepMap);
               persistence.executeSQL(sql);
            }
            
            if (shouldCommit)
            {
               persistence.commit();
            }
         }
         catch (PersistenceException exc)
         {
            openTables.remove(dmoIface);
            
            ErrorManager.recordOrThrowError(
               -1,
               "Error creating temp table:  " + getTableName(dmoIface) +
               " [" + sql + "]",
               exc);
         }
      }
      
      /**
       * Issue the SQL statement(s) against the backing database necessary to
       * drop the backing temp table for the given DMO interface.  Because
       * temp table instances are multiplexed across nested buffer scopes and
       * across multiple buffers which use the same backing DMO type, this
       * method is invoked only when the current database session is being
       * closed, after the last buffer scope referencing the backing table in
       * the current context has been closed.
       *
       * @param   dmoIface
       *          DMO interface for which the backing temp table should be
       *          dropped.
       *
       * @throws  PersistenceException
       *          if an error occurs dropping the table and silent error mode
       *          is disabled.
       *
       * @see     #doCreateTable
       */
      private void doDropTable(Class<?> dmoIface)
      throws PersistenceException
      {
         try
         {
            if (LOG.isLoggable(Level.FINER))
            {
               String msg = Utils.describeContext() + "Dropping temp-table for " + dmoIface;
               LOG.log(Level.FINER, msg, new Throwable());
            }
            
            Database database = persistence.getDatabase();
            
            // drop indexes (if any)
            Iterator<String> indexIter =
               TempTableHelper.sqlTempTableIndexDrop(database, dmoIface);
            while (indexIter.hasNext())
            {
               String sql = indexIter.next();
               sql = StringHelper.sweep(sql, sweepMap);
               persistence.executeSQL(sql);
            }
            
            // drop table
            Iterator<String> tableIter = TempTableHelper.sqlTempTableDrop(database, dmoIface);
            while (tableIter.hasNext())
            {
               String sql = tableIter.next();
               persistence.executeSQL(sql);
            }
         }
         catch (PersistenceException exc)
         {
            throw new PersistenceException(
               "Error dropping temp table:  " + getTableName(dmoIface),
               exc);
         }
      }
      
      /**
       * Get the backing temp table name associated with the given DMO
       * interface, primarily for debugging and error reporting purposes.
       *
       * @param   dmoIface
       *          DMO interface for which a table name is needed.
       *
       * @return  Name of the temp table associated with the given interface.
       */
      private String getTableName(Class<?> dmoIface)
      {
         Database tempDB = DatabaseManager.TEMP_TABLE_DB;
         String schema = DatabaseManager.getSchema(tempDB);
         Class<?> cls = DMOIndex.getImplementingClass(schema, dmoIface);
         
         return DatabaseManager.getBackingTableName(tempDB, cls);
      }
   }
   
   /**
    * An extended <code>ValidationHelper</code> which overrides methods which
    * must behave differently when validating temp tables.
    */
   protected final class TempValidationHelper
   extends ValidationHelper
   {
      /**
       * Override the parent's implementation to provide an empty list when
       * non-unique, dirty indexes are requested.  These are used only for
       * dirty share processing, which is not used for temp tables.
       * 
       * @param   unique
       *          If <code>true</code>, only unique indexes are checked.  If
       *          <code>false</code>, no indexes are checked.
       * @param   fullMatch
       *          If <code>true</code> every property in an index must be
       *          dirty for that index to be considered dirty.  If
       *          <code>false</code>, any dirty property in an index is enough
       *          to consider that index dirty.
       * 
       * @return  A list of names of indexes considered dirty, in ascending
       *          alphabetical order.
       */
      @Override
      protected List<String> listDirtyIndexes(boolean unique, boolean fullMatch)
      {
         if (unique)
         {
            return super.listDirtyIndexes(unique, fullMatch);
         }
         
         return Collections.emptyList();
      }
   }
   
   /**
    * An object registered with the <code>TransactionManager</code> to be
    * notified when its enclosing buffer's scope is closing.  This drives
    * cleanup of the backing temp table, which can take one of two forms:
    * <ul>
    *   <li>if no other buffers are using the backing table any longer, the
    *       table is dropped (along with all records);
    *   <li>otherwise, only those records associated with this buffer
    *       instance (as defined by this buffer's multiplex ID) are removed
    *       from the table, but the table itself is left intact.
    * </ul>
    */
   private class Multiplexer
   implements Finalizable
   {
      /**
       * No-op.
       */
      @Override
      public void iterate()
      {
      }
      
      /**
       * No-op.
       */
      @Override
      public void retry()
      {
      }
      
      /**
       * Cleans up the backing temp table by either dropping it (if no other
       * buffers are using it any longer), or by removing this buffer's
       * records from it.  If we are not already within a database-level
       * transaction, begin a transaction before making the changes, and
       * commit it afterward.
       */
      @Override
      public void finished()
      {
         // Only allow cleanup if the outermost scope opened on this buffer
         // has been closed.
         if (getOpenScopeCount() > 0)
         {
            return;
         }
         
         closeMultiplexScope(false);
      }
      
      /**
       * Provides a notification that the external program in which the object is registered is
       * being deleted and the object's reference may be lost after this method is called.
       * <p>
       * This implementation does nothing.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
   }
   
   /**
    * An object of this class is registered at all blocks within a no-undo buffer's scope, except
    * internal procedures (TODO: further testing is needed to verify that this is correct). Upon
    * each iteration, retry, and exit (finished) event of one of those blocks, it performs a
    * non-destructive validation (i.e., validation which reports any error, but does not alter
    * control flow).
    */
   private class NoUndoValidator
   implements Finalizable
   {
      /**
       * The block is exiting; perform a non-destructive validation of the record in the buffer.
       */
      @Override
      public void finished()
      {
         nonDestructiveValidate();
      }
      
      /**
       * The block is iterating; perform a non-destructive validation of the record in the buffer.
       */
      @Override
      public void iterate()
      {
         nonDestructiveValidate();
      }
      
      /**
       * The block is retrying; perform a non-destructive validation of the record in the buffer.
       */
      @Override
      public void retry()
      {
         nonDestructiveValidate();
      }
      
      /**
       * No-op.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
      
      /**
       * Validate the current record in the buffer (if any), but do not change the control flow
       * of the block logic in the event of an error; just report the error if not in silent
       * error mode.
       */
      private void nonDestructiveValidate()
      {
         try
         {
            validate(false);
         }
         catch (ErrorConditionException exc)
         {
            // TODO: we eat this because we can't allow the exception to unravel control flow,
            // but we need to report the error if not in silent error mode; requires some changes
            // to the TransactionManager or BlockManager infrastructure
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Error validating NO-UNDO temp-table record", exc);
               if (LOG.isLoggable(Level.FINER))
               {
                  LOG.log(Level.FINER, TemporaryBuffer.this.toString());
               }
            }
         }
      }
   }
   
   /**
    * Table copy mode.
    */
   public enum CopyTableMode
   {
      /** Table is copied using COPY-TEMP-TABLE. */
      COPY_TEMP_TABLE_MODE,
      
      /** Table is passed as an INPUT TABLE[-HANDLE] parameter. */
      INPUT_PARAM_MODE,
      
      /** Table is passed as an OUTPUT TABLE parameter. */
      OUTPUT_TABLE_PARAM_MODE,
      
      /** Table is passed as an OUTPUT TABLE-HANDLE parameter. */
      OUTPUT_TABLE_HANDLE_PARAM_MODE
   }
}
