/*
** Module   : TransparentWrapper.java
** Abstract : simple container that implements WrappedResource for any object
**
** Copyright (c) 2007-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 GES 20070221   @32199 Created initial version which represents a
**                           simple container that implements
**                           WrappedResource for any object.
** 002 CA  20090227   @41382 Added WrappedResources.isUnknown method; this
**                           implementation returns false, if referent is
**                           not an instance of WrappedResource; else,
**                           it returns referent.isUnknown() value.
** 003 CA  20130112          Made the referent field protected. Fixed equals.
** 004 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 005 CA  20150502          The resource ID needs to be kept at the resource, so it can be 
**                           accessed after the resource is deleted.
*/

package com.goldencode.p2j.util;

/**
 * Simple container that stores an object instance and implements the
 * {@link WrappedResource} interface. Since {@link #hashCode} and 
 * {@link #equals} are redirected to the contained reference, this is as
 * suitable for use in a collection as the contained reference.
 */
public class TransparentWrapper
implements WrappedResource
{
   /** The object being contained. */
   protected Object referent = null;
   
   /**
    * The resource's ID.  Remains <code>null</code> until the ID is requested via 
    * {@link handle#resourceId}
    */
   private Long id = null;
   
   /**
    * Construct an instance given the object reference to contain.
    *
    * @param    referent
    *           The object to contain.
    */
   public TransparentWrapper(Object referent)
   {
      this.referent = referent;
   }
   
   /**
    * Get this resource's ID, if is already set.
    * 
    * @return   The resource's ID or <code>null</code> if not set.
    */
   @Override
   public Long id()
   {
      return id;
   }
   
   /**
    * Set this resource's ID.
    *  
    * @param    id
    *           The resource's ID.
    */
   @Override
   public void id(long id)
   {
      this.id = id;
   }

   /**
    * Returns the referent of this instance as an <code>Object</code>.
    *
    * @return   The value as an <code>Object</code> or <code>null</code> if
    *           there is no contained object.
    */
   public Object get()
   {
      return referent;
   }   
   
   /**
    * Set the referent of this instance.
    *
    * @param    referent
    *           The value to be represented in this instance.
    */
   public void set(Object referent)
   {
      this.referent = referent;
   }
   
   /**
    * Reports if this object is valid for use.  If the contained object  
    * reference itself implements {@link WrappedResource}, then this
    * call will be redirected to the contained object.  Otherwise, if
    * the contained object is <code>null</code>, then it is considered
    * invalid and not <code>null</code> is valid.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      if (referent != null && referent instanceof WrappedResource)
      {
         // pass through the validity check
         WrappedResource res = (WrappedResource) referent; 
         return res.valid();
      }
      
      // otherwise non-null means valid
      return (referent != null); 
   }
   
   /**
    * Reports if this object is unknown.<p>
    * If {@link #referent} is not an instance of {@link WrappedResource}, it
    * returns false. Else, it returns {@link WrappedResource#unknown()} 
    * result.
    *
    * @return   <code>true</code> if object is unknown.
    */
   public boolean unknown()
   {
      if (referent != null && referent instanceof WrappedResource)
      {
         WrappedResource res = (WrappedResource) referent; 
         return res.unknown();
      }

      return false;
   }

   /**
    * Hash code implementation which is redirected to that of the 
    * contained object so long as that object is not <code>null</code>.
    *
    * @return   Hash code value for this object instance.
    */
   public int hashCode()
   {  
      return referent == null ? ((Object) this).hashCode() 
                              : referent.hashCode();
   }
   
   /**
    * Determines if this instance and the passed-in instance are equivalent.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   <code>true</code> if the objects are the same.
    */
   public boolean equals(Object obj)   
   {  
      if (obj != null && obj instanceof TransparentWrapper)
      {
         obj = ((TransparentWrapper) obj).referent;
      }

      return (referent == null) ? ((Object) this).equals(obj)
                                : referent.equals(obj);
   }
}
